###############################################################################
#                                                                             #
#                        HB/CF 18.09.98 written                               #
#                                                                             #
###############################################################################
###############################################################################
#                                                                             #
#                              Description                                    #
#                                                                             #
###############################################################################
###############################################################################
#
# AbelianGroup := rec( Relations,
#                      Generators,
#                      DiscreteLog,
#                      DiscreteExp,
#                      Basis,
#                      Information,
#                      Type := "AbelianGroup",
#                      Subgroups,
#                      Supergroups,
#                      Isomorphic,
#                      Composition           
#                    );
#
# Entrys with * MUST be specified!!!!!
#
# Relations   :*Matrix in Z^m\times n
# Generators  : n somethings in a List allowing multiplication
# DiscreteLog : function: Input something
#                         Output element in Z^n
# DiscreteExp : function: Input element in Z^n
#                         Output something
# Basis       : List of basis elements (if known)
# Subgroups   : List of AbelianGroupHom's  with g1 = g
# Supergroups : List of AbelianGroupHom's
# Isomorphic  : List of AbelianGroupHom's
# Factorgroups: List of AbelianGroupHom's
# Composition : Compositiontype of Generators "*" or "+" 
# Conductor   : List of finite and infinite places 
#               (example: [<6>, [1]], only set if Information.RCG set)
#
# Information : rec of .... representing useful information
#                           e.g. IsDirectProduct true/false
#                                Order           int
#                                Orders          list of int
#                                Exponent        int
#                                Smith           hom          Hom zu SNF
#                                SmithFrom       hom          
#                                Quotients       list of hom  QuotientenGruppen
#                                QuotientsFrom   list of hom          
#                                IsSmith         boolean      true falls Relationen-
#                                                             Matrix in SNF
#                                Name            string       
#                                DiscLog                      some information
#                                                           used by DiscreteLog
#                                ExpIntModul                  some information
#                                                           used by DiscreteExp
#                                RCG             rec         of m0, minf, [sub]
#
###############################################################################
#
# AbelianGroupElt := rec( expvec,
#                         elt,
#                         Group,
#                         operations,
#                         Information,
#                         Type := "AbelianGroupElt"
#                       );
# expvec :* Matrix in Z^\n
# elt    : something
# Group  :*AbelianGroup
# operations:* at least containing *
# Information: as above
#
###############################################################################
#
# AbelianGroupHom := rec( g1,
#                         g2,
#                         Mat12,
#                         Mat21,
#                         Information,
#                         Type := "AbelianGroupHom"
#                       );
# g1, g2:*Group
# Mat12 :*Matrix in Z^n1\times n2 : Hom from g1\to g2
# Mat21 : Inverse map
# Information: rec containing useful information e. g.
#                         Image        grp      Bild des Hom
#                         Kernel       grp      Kern des Hom 
#                         Hom12IsProc  boolean  wahr, falls Hom nicht durch
#                                               eine Matrix, sondern durch
#                                               Prozedur gegeben ist
#
#                         Hermite      rec
# Hermite := rec of       Mat       mat     HNF von ... ben"otigt f"ur 
#                                           AbelianGroupEltMove
#                         Trans     mat     zugh. TrafoMartix
#                                           ben"otigt f"ur AbelianGroupEltMove
#                         Rang      int     Rang von Mat
#
###############################################################################
#
# Produkt, Schnitt, Vereinigung testen (move)
#
# IsSub testen (_AbelianGroupCreateHomGroupToSub)
#
# _AbelianGroupExtension  testen 
#
#
# Da war noch ein Auftrag von Claus: homs durch kash-Fkt erzeugen ???
#
#
###############################################################################
#
# AbelianGroupPrintLevel := 0;     print few
#                        := 1;     print more
#                        := 2;     print all  
#
###############################################################################
#
# elt := elt1 + elt2;
# elt := elt1 - elt2;
# elt := int * elt;
# elt := elt^int;
#  
# hom := hom1 * hom2;
# hom := hom1 + hom2;
# hom := hom1 - hom2;
# hom := hom^int;
#
# elt2 := hom * elt1;
#
# group := group1 * group2;
# group := group^int;
#
# q := elt1 * elt2;    if elt2 in g and elt1 in dual group of g.
#                      q is a rational. see AbelianDualGoup
#
######################################################################
#
# elt := AbelianGroupEltCreate(group, list_of_ints);
#          list_of_ints is the exponentvector
#
# elt := AbelianGroupDiscreteLog(group, elt);
#          from object to corresponding abstract groupelt
#
# elt := AbelianGroupDiscreteExp(elt);
#          from abstract groupelt to corresponding object 
#
# elt := AbelianGroupEltReduce(elt); 
#          works only if SNF known
#
# elt := AbelianGroupEltMove(elt, group); 
#
######################################################################
#    
# group := AbelianGroupCreate(mat); or
#          AbelianGroupCreate(mat. "*"); or
#          AbelianGroupCreate(mat. "+");
#            `*' or `+' are the composition of the group
#
# group := AbelianGroupSmithCreate(group, [with_gens]); or
#            SmithNormalForm of group
#            if with_gens = true then create generators, too.  
#
# group := AbelianQuotientGroup(group1, group2);
#            compute quotientgroup group1/group2
#
# group := AbelianGroupHomImage(hom, [with_gens]);
#            compute image.
#            if with_gens = true then create generators, too.
#
# group := AbelianGroupHomKernel(hom, [with_gens]);
#            compute kernel.
#            if with_gens = true then create generators, too.
#
# group := AbelianGroupIntersect(group1, group2);
#            if path = true then install direct embeddings from
#            group1 and group2 in group. see AbelianGroupIsSub
#
# group := AbelianGroupUnite(group1, group2, [path]);
#            if path = true then install direct embeddings from
#            group1 and group2 in group. see AbelianGroupIsSub
#
# group := AbelianSubGroup([group], list_of_elts, [with_gens]); or
#          AbelianSubGroup(group, mat, [with_gens]);
#            if with_gens = true then creat generators, too.
#            the rows of mat represent a system of 
#            generators of the subgroup
#
# group := AbelianHomGroup(group1, group2);
#            group consisting of homs from group1 to group2
#
# group := AbelianDualGroup(group);
#            dual group: consisting of homs from group to C (after
#            use of AbelianGroupDiscreteExp) 
#            (here not C but Q: q -> e^(i*2*pi*q))
#
# group := AbelianGroupCanonicalQuotient(hom);
#            return quotientgroup group/kernel(hom)
#            computes image(hom);
#            and installs homs and the isomorphism
#
# group := _AbelianGroupExtension(group1, group2);
#            sub = group1 and quot = group2.
#            Computes a group g with subgroup sub and g/sub = quot.
#            This function needs the DiscreteLog of sub. 
#            Generators of sub and quot has to be in a common supergroup.
#            application: sub = H/S, quot = I/H, g = I/S 
#                         g/sub = (I/S) / (H/S) = I/H = quot
#            I : ideals, H : principal ideals, S : ray
#           
# group := RayClassGroupToAbelianGroup(ideal, [list_of_places]);
#            creates the (abstract) RayClassGroup with generators
#            (ZZ : the ideal is an integer)
#
# group := RayResidueRingToAbelianGroup(ideal, [list_of_places]);
#            creates the (abstract) multiplicative Group of the
#            RayResidueRing with generators
#            (ZZ : the ideal is an integer)
#
# group := AbelianFixPointGroup(hom);
#            or AbelianFixPointGroup(list of homs);
#            hom has to be endomorphism.
#            Computes the group consisting of elements than are
#            (pointwise) fix under each given endomorphism
#
######################################################################
#
# bool := _AbelianGroupIsHom(hom); 
#            returns true iff function is group homomorphism
# list := AbelianGroupIsAut(hom, [boolean]);
#           returns the list [true, inverse_hom] iff hom is automorphism,
#           the list [false] otherwise.
#           if false is added as second parameter, the check if hom 
#           is endomorphism is obmitted
#
# bool := AbelianGroupEqual(g1, g2);
#           Checks if g1 and g2 are contained in each other
#           
# int  := AbelianGroupEltOrder(elt); 
#
# int  := AbelianGroupNumberGenerators(group);
#           returns the number of generators of group in the current
#           representation
#
# int  := AbelianGroupMinNumberGenerators(group);
#           returns the minimal number of generators of group
#           = number of generators of the SNF of group
#
# int  := AbelianGroupIndex(group1, group2)
#           returns the the groupindex [group1 : group2]
#
# int  := AbelianGroupExponent(group)
#           returns the exxponent of group
#
# list_of_elts := AbelianGroupBasis(group, [boolean]);
#                   returns a list of basiselts of the group.    
#                   if true is added, DiscreteExp is applied    
#
# list_of_elts := AbelianGroupGenerators(group, [boolean]);
#                   returns a list of generators of the group.
#                   if true is added, DiscreteExp is applied    
#
# hom  := AbelianGroupHomCreate(group1, group2, mat, [check]); or
#         AbelianGroupHomCreate(group1, group2, mat, [mat_inv]); or
#         AbelianGroupHomCreate(group1, group2, function); or
#           if check = true then check if created object is hom
#           mat_inv is the mat of inverse hom
#
# bool := AbelianGroupIsSub(group1, group2);
#           Returns true if g1 is subgroup of g2 and false ohterwise. 
#           In the first case dircet homomorphisms are installed.
#           This is useful if you often move between g1 and g2.
#
# list_of_homs := _AbelianGroupFindHom(group1, group2, [list_of_strings]);
#                   returns a list of homs from group1 to group2.
#                   list_of_strings has to be a sublist of 
#                   ["Subgroups", "Supergroups", "Quotients",
#                    "QuotientsFrom", "Smith", "SmithFrom"];
#
# hom := _AbelianGroupHomMove(g1, g2);
#          Creates the homomorphism that moves from g1 to g2, if possible
# 
# void := AbelianGroupName(g, name_string);
#           The name of the group is added to the relation matrix by Print
#
###############################################################################
###############################################################################

###############################################################################
#                                                                             #
#                              FUNCTIONS                                      #
#                                                                             #
###############################################################################

AbelianGroupEltCreate             := Ignore;
_AbelianGroupEltPower             := Ignore;
AbelianGroupHomCreate             := Ignore;
AbelianGroupCreate                := Ignore;
AbelianGroupHomImage              := Ignore;
AbelianGroupEltMove               := Ignore;
_AbelianGroupFindSupergroups      := Ignore;
_AbelianGroupFindCommomSupergroup := Ignore;
_AbelianGroupFindHom              := Ignore;
AbelianGroupIsSub                 := Ignore;
AbelianSubGroup                   := Ignore;
AbelianGoupHomCreate              := Ignore;
AbelianGroupSmithCreate           := Ignore;
_AbelianGroupCreateHomGroupToSub  := Ignore;
AbelianGroupDiscreteExp           := Ignore;
_AbelianGroupStandardDiscreteExp  := Ignore;
_AbelianGroupEltDirectToTensor    := Ignore;
AbelianGroupMultiHomCreate        := Ignore;
_AbelianGroupHomFromFunc          := Ignore;
AbelianGroupOrder                 := Ignore;
AbelianQuotientGroup              := Ignore;
_AbelianGroupHomMove              := Ignore;
_AbelianGroupSpecialKernel        := Ignore;
AbelianGroupEqual                 := Ignore;
AbelianGroupIsAut                 := Ignore;
_IntFastPower                     := Ignore;
AbelianGroupEltReduce             := Ignore;
AbelianGroupGenerators            := Ignore;
AbelianGroupEltOrder              := Ignore;
AbelianGroupBasis                 := Ignore;
AbelianGroupCyclicFactors         := Ignore;
AbelianGroupDirectProduct         := Ignore;
_AbelianGroupEqual                := Ignore;
_AbelianGroupApplyHomChain        := Ignore;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupPrintLevel
#CALLING   AbelianGroupPrintLevel := a;
#SHORTDOC  Sets printlevel.
#LONGDOC   Sets printlevel for printing groups:
#          \begin{itemize}
#          \item AbelianGroupPrintLevel := 0;   : few information (standard)
#          \item AbelianGroupPrintLevel := 1;   : more information
#          \item AbelianGroupPrintLevel := 2;   : all information
#          \end{itemize}
#ARGUS     a @ integer
#
#XREF      
#KASH      few information (standard): @
#          g := AbelianGroupCreate(Mat(Z, [[4,2], [8,2]]));
#          AbelianGroupSmithCreate(g);;
#          h := AbelianGroupHomCreate(g, g, Mat(Z, [[0,1],[1,0]]), true);
#          @ more information: @
#          AbelianGroupPrintLevel := 1;;
#          g;
#          h := AbelianGroupHomCreate(g, g, Mat(Z, [[0,1],[1,0]]), true);
#          AbelianGroupPrintLevel := 0;;
#          
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupPrintLevel := 0;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupNumberGenerators
#CALLING   a := AbelianGroupNumberGenerators(g);
#SHORTDOC  Returns the number of generators.
#LONGDOC   Returns the number of generators of group $g$ 
#          in the current representation.
#          
#ARGUS     g @ group
#          a @ integer
#
#XREF      AbelianGroupGenerators
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          a := AbelianGroupNumberGenerators(g);          
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupNumberGenerators := function(arg)
local g;

  if Length(arg) <> 1 then
    Error("use AbelianGroupNumberGenerators(group)\n");
  fi;
  g := arg[1];  

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupNumberGenerators(group)\n");
  fi;
  return MatCols(g.Relations);
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupMinNumberGenerators
#CALLING   a := AbelianGroupMinNumberGenerators(g); 
#SHORTDOC  Returns the minimal number of generators of group $g$.
#LONGDOC   
#          
#ARGUS     g @ group
#          a @ integer
#
#XREF      AbelianGroupBasis
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          a := AbelianGroupMinNumberGenerators(g);  
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupMinNumberGenerators := function(arg)
local g;
  
  if Length(arg) <> 1 then 
    Error("use AbelianGroupMinNumberGenerators(group)\n");
  fi;
  g := arg[1];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupMinNumberGenerators(group)\n");
  fi;
  return MatCols(AbelianGroupSmithCreate(g).Relations);
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupExponent
#CALLING   a := AbelianGroupExponent(g);
#SHORTDOC  Returns the exponent of a group.
#LONGDOC   Returns the exponent $a$ of group $g$.
#          
#ARGUS     g @ group
#          a @ integer
#
#XREF      AbelianGroupOrder, AbelianGroupEltOrder
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,0], [0,4]]));
#          a := AbelianGroupExponent(g);  
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupExponent := function(arg)
local g;

  if Length(arg) <> 1 then
    Error("use AbelianGroupExponent(group)\n");
  fi;
  g := arg[1];  

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupExponent(group)\n");
  fi;
  AbelianGroupSmithCreate(g);
  return g.Information.Exponent;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupBasis
#CALLING   L := AbelianGroupBasis(g [, exp]); 
#SHORTDOC  Returns a basis of a group.
#LONGDOC   Returns a list of basis elements of group $g$. If the second 
#          (optional) parameter $exp$ is 'true' then the discrete exp is
#          applied.
#          
#ARGUS     g   @ group
#          L   @ list @ list of basis elements of $g$
#          exp @ boolean
#           
#XREF      AbelianGroupMinNumberGenerators
#KASH      compute the Basis of a ray class group: @ 
#          O := OrderMaximal(Z, 2, 10);
#          g := RayClassGroupToAbelianGroup(3*O, [1, 2]);
#          l := AbelianGroupBasis(g);
#          l := AbelianGroupBasis(g, true);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupBasis := function(arg)
local g, s, exp, j,l;
  
  if Length(arg) >2 or Length(arg) < 1 then
    Error("use AbelianGroupBasis(group [,exp])\n");
  fi;

  g := arg[1];
  if Length(arg) > 1 then
    exp := arg[2];
  else
    exp := false;
  fi;

  if exp and IsBound(g.Basis) then
    return g.Basis;
  fi;

  s := AbelianGroupSmithCreate(g);

  l := MatToRowList(MatId(Z, AbelianGroupNumberGenerators(s)));
  l := List(List(l, x->AbelianGroupEltCreate(s, x)), x->AbelianGroupEltMove(x, g));
   for j in l do j.Group.Supergroups:=g.Supergroups; od;  

  if exp then
    if (AbelianGroupIsDirectProduct(g) and IsBound(g.Generators)) then
      g.Basis := g.Generators;
      l := g.Basis;
    elif not(IsBool(AbelianGroupDiscreteExp(l[1]))) then
      Apply(l, x-> AbelianGroupDiscreteExp(x));
      g.Basis := l;
    fi;
  fi;

  return l;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupGenerators
#CALLING   L := AbelianGroupGenerators(g [, exp]); 
#SHORTDOC  Returns generators of a group.
#LONGDOC   Returns a list of generators of group $g$. If the second 
#          (optional) parameter $exp$ is 'true' then the discrete exp is
#          applied.
#          
#ARGUS     g   @ group
#          L   @ list @ list of generators of $g$
#          exp @ boolean
#XREF      AbelianGroupNumberGenerators, AbelianGroupBasis
#KASH      @ 
#          O := OrderMaximal(Z, 2, 10);
#          g := RayClassGroupToAbelianGroup(3*O, [1, 2]);
#          l := AbelianGroupGenerators(g);
#          l := AbelianGroupGenerators(g, true);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupGenerators := function(arg)
local g, exp, l;

  if Length(arg) >2 or Length(arg) < 1 then
    Error("use AbelianGroupGenerators(group [,exp])\n");
  fi;

  g := arg[1];
  if Length(arg) > 1 then
    exp := arg[2];
  else
    exp := false;
  fi;

  if exp and IsBound(g.Generators) then
    return g.Generators;
  fi;

  l := MatToRowList(MatId(Z, AbelianGroupNumberGenerators(g)));
  Apply(l, x->AbelianGroupEltCreate(g, x));
  if not IsBound(g.Information.Orders) then
    g.Information.Orders := List(l, x-> AbelianGroupEltOrder(x));
  fi;

  if exp then
    Apply(l, x-> AbelianGroupDiscreteExp(x));
    g.Generators := l;
  fi;

  return l;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupCyclicFactors
#CALLING   L := AbelianGroupCyclicFactors(g);
#SHORTDOC  Returns cyclic factors and orders. 
#LONGDOC   Returns a list of lists of cyclic factors of the group
#          and orders of the cyclic factors. 
#          
#ARGUS     g @ group
#          L @ list @ list of cyclic factors of $g$
#
#XREF      AbelianGroupSmithCreate
#KASH      compute the cyclic factors of a ray class group: @ 
#          O := OrderMaximal(Z, 2, 10);
#          g := RayClassGroupToAbelianGroup(3*O, [1, 2]);
#          L := AbelianGroupCyclicFactors(g);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupCyclicFactors := function(arg)
local i, basis_exp, orders, list, g;

  if Length(arg) <>1 then
    Error("use AbelianGroupCyclicFactors(group)\n");
  fi;
  g := arg[1];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupCyclicFactors(group)\n");
  fi;

  basis_exp := AbelianGroupBasis(g, true);
  orders := AbelianGroupSmithCreate(g).Information.Orders;

  list := [];
  for i in [1..Length(orders)] do
    list[i] := [];
    list[i][1] := basis_exp[i];
    list[i][2] := orders[i];
  od;

  return list;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupIndex
#CALLING   a := AbelianGroupIndex(g, s);
#SHORTDOC  Returns the index of a subgroup in a group.
#LONGDOC   Returns the index $a$ of $s$ in $g$. 
#          $s$ has to be a subgroup of $g$.
#          
#ARGUS     g @ group
#          s @ group @ subgroup of $g$
#          a @ integer
#
#XREF      AbelianSubGroup, AbelianGroupOrder
#KASH      @ 
#          g := AbelianGroupCreate([[0]]);
#          bas := AbelianGroupBasis(g); 
#          s := AbelianSubGroup(g, 2*bas[1]); 
#          a := AbelianGroupIndex(g, s);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupIndex := function(arg)
local i, m, b, g1, g2, g, s;

  if Length(arg) <> 2 then
    Error("use AbelianGroupIndex(group, group)\n");
  fi;
  g := arg[1];
  s := arg[2];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup"
          and IsRec(s) and IsBound(s.Type) and s.Type = "AbelianGroup") then
    Error("use AbelianGroupIndex(group, group)\n");
  fi;

  if not(AbelianGroupIsSub(s, g)) then
    Error("second group is not subgroup of first group");
  fi;

  if (not(AbelianGroupOrder(s)) = 0) then
    return AbelianGroupOrder(g)/AbelianGroupOrder(s);
  fi;
  m := [];
  if (AbelianGroupNumberGenerators(g) = AbelianGroupNumberGenerators(s)) then  
    g1 := g;
    g2 := s;

    b := AbelianGroupGenerators(g2);
    for i in [1..Length(b)] do
      m[i] := MatToRowList(AbelianGroupEltMove(b[i], g).expvec)[1];
    od;
    IsMat(m);
    return Abs(MatDet(m));
  elif (IsBound(g.Information.Smith) and AbelianGroupMinNumberGenerators(g) = AbelianGroupNumberGenerators(s)) then
    g1 := g.Information.Smith.g2;
    g2 := s;

    b := AbelianGroupGenerators(g2);
    for i in [1..Length(b)] do
      m[i] := MatToRowList(AbelianGroupEltMove(b[i], g).expvec)[1];
    od;
    IsMat(m);
    return Abs(MatDet(m));
  elif (IsBound(s.Information.Smith) and AbelianGroupMinNumberGenerators(s) = AbelianGroupNumberGenerators(g)) then
    g1 := g;
    g2 := s.Information.Smith.g2;

    b := AbelianGroupGenerators(g2);
    for i in [1..Length(b)] do
      m[i] := MatToRowList(AbelianGroupEltMove(b[i], g).expvec)[1];
    od;
    IsMat(m);
    return Abs(MatDet(m));
  elif (IsBound(g.Information.Smith) and IsBound(s.Information.Smith) and AbelianGroupMinNumberGenerators(g) = AbelianGroupMinNumberGenerators(s)) then
    g1 := g.Information.Smith.g2;
    g2 := s.Information.Smith.g2;
    
    b := AbelianGroupGenerators(g2);
    for i in [1..Length(b)] do
      m[i] := MatToRowList(AbelianGroupEltMove(b[i], g).expvec)[1];
    od;
    IsMat(m);
    return Abs(MatDet(m));
  else
    return AbelianGroupOrder(AbelianQuotientGroup(g,s));
  fi;
end;

_IsZeroVec := function(v)
  local i, zero;

  IsMat(v);
  if MatRows(v) > 1 then return false; fi;
  i := 1;
  zero := true;
  while zero and i<=MatCols(v) do
    zero := (v[1][i] = 0);
    i := i+1;  
  od;
  return zero;
end;

_AbelianGroupHomPrint := function(h)
  if h.Type = "AbelianGroupHom" then
    if IsBound(h.Mat12) then
    Print("HomMatrix=","\n",h.Mat12, "from ","\n", h.g1, " to ","\n", h.g2, "\n");
    else
    Print("Hom=from ", h.g1, " to ", h.g2, "\n");
    fi;
  elif h.Type = "AbelianGroupMultiHom" then
    Print("MultiHom ", List(h.Tensor.Information.TensorInfo, x->MatCols(x.g1.Relations))," from ", h.g1, " to ", h.g2, "\n");
  fi;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupOrder
#CALLING   a := AbelianGroupOrder(g);
#SHORTDOC  Returns the order of group $g$.
#LONGDOC   
#          
#ARGUS     g @ group
#          a @ integer
#
#XREF      AbelianGroupExponent
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          a := AbelianGroupOrder(g);  
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupOrder := function(arg)
local g;
  
  if Length(arg) <> 1 then
    Error("use AbelianGroupOrder(group)\n");
  fi;
  g := arg[1];

  if not IsRecType(g, "AbelianGroup") then
    Error("use AbelianGroupOrder(group)\n");
  fi;

  if not IsBound(g.Information.Order) then
    AbelianGroupSmithCreate(g, false);  # CF: without generators in this case
  fi;

  return g.Information.Order;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupDiscreteLog
#CALLING   a := AbelianGroupDiscreteLog(g, b);
#SHORTDOC  Returns the representation of an object in a group. 
#LONGDOC   Returns the abstract representation $a$ (exponent vector, corresponding to
#          the generators of $g$) of a concrete $b$ in the abstract group $g$.
#          If $g$ represents for example a class group of a number field, then a concrete
#          Ideal is represented by {\tt AbelianGroupDiscreteLog(g,a)} as
#           an abstract element of this group.
#          
#ARGUS     g @ group
#          a @ group element @ representation of $b$ in the abstract group $g$
#          b @ object
#
#XREF      AbelianGroupDiscreteExp
#KASH      @ 
#          O := OrderMaximal(x^2-10);
#          g := RayResidueRingToAbelianGroup(27*O, [2]);
#          a := AbelianGroupEltCreate(g, [1, 2, 3, 0, 0]);
#          b := AbelianGroupDiscreteExp(a);
#          a := AbelianGroupDiscreteLog(g, b);
#      		@
#          g := RayClassGroupToAbelianGroup(1*O);
#          I := Ideal(2,Elt(O,[0,1]));
#          b := AbelianGroupDiscreteLog(g,I);         
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupDiscreteLog := function(arg)
local i, j, l, weiter, i, s, g, elt;

  if Length(arg) <> 2 then
    Error("use AbelianGroupDiscreteLog(group, elt)");
  fi;
  g   := arg[1];
  elt := arg[2];  

  if not IsRecType(g, "AbelianGroup") then
    Error("use AbelianGroupDiscreteLog(group, elt)");
  fi;

  if not IsBound(g.DiscreteLog) then
    l := _AbelianGroupFindSupergroups(g);
    weiter := true; 
    i := 1;
    while (i <= Length(l) and weiter) do
      if (IsBound(l[i].DiscreteLog)) then
        s := l[i];
        weiter := false;
      fi;
      i := i+1; 
    od;
    if weiter then
      if AbelianGroupPrintLevel >0 then
        Print("no DiscreteLog found\n");
      fi;  
      return false;    
    fi;
    l := _AbelianGroupFindHom(s, g);
    l := Product(l);
    g.DiscreteLog := function(i, j)
      return l*s.DiscreteLog(s, j);
    end;
    return g.DiscreteLog(g, elt);
  else
    return g.DiscreteLog(g, elt);
  fi;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupDiscreteExp
#CALLING   b := AbelianGroupDiscreteExp(a);
#SHORTDOC  Returns the object corresponding to a representation in a group. 
#
#LONGDOC   Returns the concrete object $b$ corresponding to the abstact 
#          representation of $a$ in $g$. For example, if the abstract group $g$ 
#          represents a concrete class group of a number field, then
#          {\tt AbelianGroupDiscreteExp(a)} calculates an Ideal {\tt b},
#          that is representing a class of ideals in the class group. 
#          If $g$ is the dual group of a given group, $a \in g$, then 
#          $b:=${\tt AbelianGroupDiscreteExp(a)} is the homomorphism
#          represented by $a$. In this Form it can directly be used to 
#          evaluate element of the original group, to witch $g$ is dual.  
#                  
#ARGUS     g @ group
#          a @ group element @ representation of $b$ in the abstract group $g$
#          b @ object
#
#XREF      AbelianGroupDiscreteLog
#KASH      @ 
#          O := OrderMaximal(x^2-2*x-5);
#          g := RayResidueRingToAbelianGroup(27*O, [2]);
#          a := AbelianGroupEltCreate(g, [1, 2, 3, 0, 0]);
#          b := AbelianGroupDiscreteExp(a);
#          a := AbelianGroupDiscreteLog(g, b);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupDiscreteExp := function (arg)
local g, l, i, weiter, s, elt;

  if Length(arg) <> 1 then
    Error("use AbelianGroupDiscreteExp(elt)");
  fi;
  elt := arg[1];
  g := elt.Group;

  if not IsBound(g.DiscreteExp) then
    l := _AbelianGroupFindSupergroups(g);
    weiter := true; 
    i := 1;
    while (i <= Length(l) and weiter) do
      if (IsBound(l[i].DiscreteExp)) then
        s := l[i];
        weiter := false;
      fi;
      i := i+1; 
    od;
    if weiter then
      if AbelianGroupPrintLevel > 0 then
        Print("no DiscreteExp found\n");
      fi;
      return false;    
    fi;
    elt := AbelianGroupEltMove(elt, s);
    return s.DiscreteExp(elt);
  else
    return g.DiscreteExp(elt);
  fi;
end;

_AbelianGroupStandardDiscreteExp := function (elt)
  local i, vec, a, g, l, s, weiter, m;

  elt := AbelianGroupEltReduce(elt);
  g := elt.Group;
  if not IsBound(g.Generators) then
    l := _AbelianGroupFindSupergroups(g);
    weiter := true; 
    i := 1;
    if i <= Length(l) and weiter then
      if IsBound(l[i].Generators) then
        s := l[i];
        weiter := false;
      fi;
      i := i+1; 
    fi;
    if weiter then
      if AbelianGroupPrintLevel >0 then
        Print("no generators found\n");
      fi;
      return false;    
    fi;
    elt := AbelianGroupEltMove(elt, s);
  fi;
  vec := elt.expvec;
  IsMat(vec);
  if elt.Group.Composition = "*" then
    if IsBound(elt.Group.Information) and 
       IsBound(elt.Group.Information.ExpIntModul) then
       m := elt.Group.Information.ExpIntModul;

       a := _IntFastPower(elt.Group.Generators[1], elt.expvec[1][1], m);
       for i in [2..MatCols(vec)] do
         a := a * _IntFastPower(elt.Group.Generators[i], elt.expvec[1][i], m);
       od;
       a := a mod m;
    else
      a := elt.Group.Generators[1] ^ elt.expvec[1][1];
      for i in [2..MatCols(vec)] do
        a := a * (elt.Group.Generators[i] ^ elt.expvec[1][i]);
      od;
    fi;
  elif elt.Group.Composition = "+" then
    a := elt.expvec[1][1] * elt.Group.Generators[1];
    for i in [2..MatCols(vec)] do
      a := a + elt.expvec[1][i] * elt.Group.Generators[i];
    od;
  else
    Error("not supported");
  fi;

  return a;
end;

_IntFastPower := function(a, b, m)
local i, list, c, power;

  if b<0 then
    a := 1/a mod m;
    b := -b;
  fi;  

  list := [];
  i := 1;  
  while b>0 do
    list[i] := b mod 2;
    b := (b-list[i])/2; 
    i := i+1;
  od;
  
  c := 1;
  power := a;
  for i in [1..Length(list)] do
    c := c*power^list[i] mod m;
    power := power*power mod m;
  od;
  return c;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupEltReduce
#CALLING   elt2 := AbelianGroupEltReduce(elt1 [, positive]);
#SHORTDOC  Returns the reduced group element.
#LONGDOC   Returns the reduced group element (exponent vector) $elt2$ of 
#          $elt1$.
#          If the second (optional) parameter is 'true' then all entries of
#          $elt2$ are not negative. Only in this case it is sure that a 
#          good reduction is performed.
#          
#ARGUS     elt1 @ group element
#          elt2 @ group element @ reduced representation
#          positive @ boolean
#
#XREF      
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          elt1 := AbelianGroupEltCreate(g, [-100, -217]);
#          elt2 := AbelianGroupEltReduce(elt1);
#          AbelianGroupSmithCreate(g);
#          elt2 := AbelianGroupEltReduce(elt1);
#          elt2 := AbelianGroupEltReduce(elt1, true);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupEltReduce := function(arg)
  local a, i, g, l, b, pos;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use AbelianGroupEltReduce(Elt [, positive])");
  fi;

  pos := false; # do not change this

  if Length(arg) = 2 then
    if IsBool(arg[2]) then
      pos := arg[2];
    else
      Error("use AbelianGroupEltReduce(Elt [, positive])");
    fi;
  fi;

  a := arg[1];
  g := a.Group;
  l := g.Relations;

  if pos then
    AbelianGroupGenerators(g); 
  fi;

  if IsBound(g.Information.Smith) then     # true if pos = true
    b := g.Information.Smith * a;
    a := AbelianGroupEltCreate(
          g, b.expvec * g.Information.Smith.Mat21);
  elif IsBound(g.Information.IsDirectProduct) and
     g.Information.IsDirectProduct then
    for i in [1..MatCols(a.expvec)] do
      if l[i][i]<>0 then
        a.expvec[1][i] := a.expvec[1][i] mod l[i][i];
      fi;
    od;
  fi;

  if IsBound(g.Information.Orders) then    # true if pos = true
    for i in [1..MatCols(a.expvec)] do
      if g.Information.Orders[i]<>0 then
        a.expvec[1][i] := a.expvec[1][i] mod g.Information.Orders[i];
      fi;
    od;
  fi;

  return a;
end;

_AbelianGroupEltAdd := function(a, b)
  local c;
  if a.Type<>"AbelianGroupElt" or b.Type<>"AbelianGroupElt" or
    a.Group<>b.Group then
    Error("can only add elements of the same Type");
  fi;
  c := rec(expvec := a.expvec+b.expvec,
           operations := a.operations,
           Type := a.Type,
           Group := a.Group
          );

  if IsBound(a.elt) and IsBound(b.elt) then
    c.elt := a.elt + b.elt;
  fi;

  return AbelianGroupEltReduce(c);
end;

_AbelianGroupEltSubtract := function(a, b)
  local c;
  if a.Type<>"AbelianGroupElt" or b.Type<>"AbelianGroupElt" or
    a.Group<>b.Group then
    Error("can only subtract elements of the same Type");
  fi;
  c := rec(expvec := a.expvec-b.expvec,
           operations := a.operations,
           Type := a.Type,
           Group := a.Group
          );

  if IsBound(a.elt) and IsBound(b.elt) then
    c.elt := a.elt - b.elt;
  fi;

  return AbelianGroupEltReduce(c);
end;

_AbelianGroupEltMoveToSub := function(elt, hom)
  local i, group, sub, H, T, newvec, oldvec, k, ok, H, T, n, M, w;

  group   := hom.g1;
  sub     := hom.g2;
  H       := hom.Information.Hermite.Mat;
  T       := hom.Information.Hermite.Trans;
  n       := hom.Information.Hermite.Rang;

  oldvec := elt.expvec;
  newvec := [];
  for i in [1..MatRows(T)] do
    Add(newvec, 0);
  od;
  IsMat(newvec);
  ok := true;
  i:= 1;
  k:= 0;
  while (i<=n) and ok do
    while (H[i][i+k] = 0) do
      if not(oldvec[1][i+k] = 0) then
        ok := false;
      fi;
      k := k+1;
    od;
    w := oldvec[1][i+k] / H[i][i+k];

    if IsInt(w) then
      newvec[1][i] := w;
      oldvec := oldvec - newvec[1][i]*H[i];  
    else
      ok := false;
    fi;   
    i := i+1;
  od;
  if ok and _IsZeroVec(oldvec) then
    
#    if not (IsMat(H) or newvec * H = elt.expvec) then  
#      Error("bug in _AbelianGroupEltMoveToSub\n");     
#    fi;

    return AbelianGroupEltCreate(sub, newvec * T); # H = T*B  
  else
    return false;
  fi;
end;

_AbelianGroupHomMult := function(a, b)
  local g, c, expvec, hom, b_new;
   
  if IsInt(a) and
    b.Type = "AbelianGroupElt" then
    return _AbelianGroupEltPower(b, a);
  elif IsInt(a) and
    b.Type = "AbelianGroupHom" then
    return AbelianGroupHomCreate(b.g1, b.g2, a*b.Mat12);
  elif IsInt(a) and  
    b.Type = "AbelianGroupMultiHom" then
    return AbelianGroupMultiHomCreate(b.g1, b.g2, b.Tensor, a*b.HomTensor.Mat12);    
  elif a.Type = "AbelianGroupHom" and
    b.Type = "AbelianGroupHom" and
    IsBound(a.Mat12) and IsBound(b.Mat12) then 
    if (b.g2 = a.g1) then
      if IsBound(a.Mat21) and IsBound(b.Mat21) then
        return AbelianGroupHomCreate(b.g1, a.g2, b.Mat12 * 
               a.Mat12, a.Mat21 * b.Mat21);
      else 
        return AbelianGroupHomCreate(b.g1, a.g2, b.Mat12 * a.Mat12);
      fi;
    else 
      Error("homomorphisms don't match");
    fi;

# Hom * MultiHom, MultiHom * Hom, MultiHom * MultiHom  ist nicht
# implementiert, aber z.B.:  multi_hom * (hom * elt) 

  elif a.Type = "AbelianGroupHom" and
       b.Type = "AbelianGroupElt" then

    if not a.g1 = b.Group then
      Error("not defined\n");
    fi;

    IsMat(b.expvec);
    if IsBound(a.Information) and IsBound(a.Information.Hom12IsProc)
                              and a.Information.Hom12IsProc = true then
      c := _AbelianGroupEltMoveToSub(b, a);
    else
      expvec := b.expvec * a.Mat12;
      c := AbelianGroupEltCreate(a.g2, expvec);
    fi;
    if c = false then #!@!@!@!@!
      if AbelianGroupPrintLevel > 0 then
        Print("element not contained in subgroup\n");
      fi;
    else 
      c := AbelianGroupEltReduce(c);
    fi;    
    return c;
  elif a.Type = "AbelianGroupMultiHom" and
       b.Type = "AbelianGroupElt" then
    if not a.g1 = b.Group then
      Error("not defined\n");
    fi;
    b_new := _AbelianGroupEltDirectToTensor(b, a.Tensor); # from g1 to a.Tensor
    return a.HomTensor * b_new;                          # from a.Tensor to g2
  elif a.Type = "AbelianGroupElt" and 
       b.Type = "AbelianGroupElt" and
       IsBound(a.Group.Information) and                
       IsBound(a.Group.Information.DualFrom) and
       a.Group.Information.DualFrom = b.Group then
    hom := AbelianGroupDiscreteExp(a);
    c := AbelianGroupEltReduce(hom*b);
    return AbelianGroupDiscreteExp(c);  
  elif a.Type = "AbelianGroupElt" and 
       IsBound(a.Group.Information) and                
       IsBound(a.Group.Information.DualFrom) then
    g := a.Group.Information.DualFrom;
    b_new := AbelianGroupDiscreteLog(b, g);
    hom := AbelianGroupDiscreteExp(a);
    c := AbelianGroupEltReduce(hom*b_new);
    return AbelianGroupDiscreteExp(c);  
  fi;

  Error("not defined\n");
end;

_AbelianGroupHomPower := function(a, b)
  local m, g_from, g_to, list, inverse;

  if (IsInt(b) and  a.Type = "AbelianGroupHom") then
    if b<0 then
      if not IsBound (a.Mat21) then
        list := AbelianGroupIsAut(a, false);
        if list[1] = true then
          a.Mat21 := list[2].Mat12;
        else 
          if AbelianGroupPrintLevel > 0 then
            Print("homomorphism not invertible\n");
          fi;
          return false;
        fi;
      fi;
      inverse := true;
      g_from  := a.g2;
      g_to    := a.g1;
      m       := a.Mat21;
      b       := -b;
    else
      inverse := false;
      g_from  := a.g1;
      g_to    := a.g2;
      m       := a.Mat12;
    fi;

    if (b=1) then
      if inverse then
        return AbelianGroupHomCreate(g_from, g_to, (m), (a.Mat12));
      else
        return AbelianGroupHomCreate(g_from, g_to, (m));
      fi;
    elif (b<>1 and g_from = g_to) then
      if inverse then
        return AbelianGroupHomCreate(g_from, g_to, (m)^b, (a.Mat12)^b);
      else
        return AbelianGroupHomCreate(g_from, g_to, (m)^b);
      fi;
    else
      if AbelianGroupPrintLevel > 0 then
        Print("homomorphism is no endomorphism\n");
      fi;  
      return false;
    fi;
  else
    Error("use hom^int");
  fi;
end;

_AbelianGroupHomAdd := function(a, b)
  local c;

  if a.Type = "AbelianGroupHom" and b.Type = "AbelianGroupHom" and
     a.g1 = b.g1 and a.g2 = b.g2 then
    c := AbelianGroupHomCreate(a.g1, a.g2, b.Mat12 + a.Mat12);
    return c;
  elif a.Type = "AbelianGroupMultiHom" and b.Type = "AbelianGroupMultiHom" and
     a.g1 = b.g1 and a.g2 = b.g2 and a.Tensor = b.Tensor then
    c := AbelianGroupMultiHomCreate(a.g1, a.g2, a.Tensor, b.HomTensor.Mat12 + a.HomTensor.Mat12);
    return c;
  else 
    Error("not defined\n");
  fi;
end;

_AbelianGroupHomSubtract := function(a, b)

  if a.Type = "AbelianGroupHom" and b.Type = "AbelianGroupHom" and
     a.g1 = b.g1 and a.g2 = b.g2 then
    return AbelianGroupHomCreate(a.g1, a.g2, b.Mat12 - a.Mat12);
  elif a.Type = "AbelianGroupMultiHom" and b.Type = "AbelianGroupMultiHom" and
     a.g1 = b.g1 and a.g2 = b.g2 and a.Tensor = b.Tensor then
    return AbelianGroupMultiHomCreate(a.g1, a.g2, a.Tensor, 
           b.HomTensor.Mat12 - a.HomTensor.Mat12);
  else 
    Error("not defined\n");
  fi;
end;


_AbelianGroupEltPower := function(a, n)
  local c;
  if a.Type<>"AbelianGroupElt" or not IsInt(n) then
    Error("can only elements with ints");
  fi;
  c := rec(expvec := n*a.expvec,
           operations := a.operations,
           Type := a.Type,
           Group := a.Group
          );

  if IsBound(a.elt) then
    c.elt := n*a.elt;
  fi;

  return AbelianGroupEltReduce(c);
end;

_AbelianGroupEltPrint := function(a)
  if IsBound(a.elt) then
    Print(a.elt);
  elif IsBound(a.expvec) then
    Print(a.expvec);
  else
    Print([],"\n");
  fi;
end;

_AbelianGroupPrint := function(g)
local i, m, printlevel_old;

  printlevel_old := AbelianGroupPrintLevel;
  AbelianGroupPrintLevel := 0;

  if IsBound(g.Information) and IsBound(g.Information.Name) then
    Print(g.Information.Name,"\n");
  fi;
  Print("Group with Relations:\n", g.Relations);
  
  if not IsBound(AbelianGroupPrintLevel) or 
     not IsInt(AbelianGroupPrintLevel) then
    AbelianGroupPrintLevel := 0;
  fi;
  
  if printlevel_old >= 1 then
    if IsBound(g.Information) and IsBound(g.Information.IsSmith) and 
       g.Information.IsSmith = true then
      Print("\nis in SNF  ");
    fi;
    if IsBound(g.Information) and IsBound(g.Information.Exponent) then
      Print("\nexponent   : ", g.Information.Exponent);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.Order) then
      Print("\norder      : ", g.Information.Order);
    fi;
    if IsBound(g.Generators) then
      Print("\ngenerators : ", g.Generators);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.Orders) then
      Print("\norder of generators : ", g.Information.Orders);
    fi;
    if IsBound(g.Basis) then
      Print("\nbasis      : ", g.Basis);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.Smith) and 
       not IsBound(g.Information.IsSmith) then
      Print("\nhas SNF    : ", g.Information.Smith.g2);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.SmithFrom) then
      Print("\nis SNF from: ", g.Information.SmithFrom.g2);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.RRR) then
      m := [g.Information.RRR.m0, g.Information.RRR.minf];
      Print("\nis mult. group of ray residue ring of ray: ", m);
    fi;
    if IsBound(g.Information) and IsBound(g.Information.HomGroup) then
      Print("\nis  group of homs from ",g.Information.HomGroup[1],
            " to ",g.Information.HomGroup[2]);
    fi;

    if IsBound(g.Information) and IsBound(g.Information.RCG) then
      m := [g.Information.RCG.m0, g.Information.RCG.minf];
      if not IsBound(g.Information.RCG.Sub) then
        Print("\nis ray class group of ray: ", m);
        if IsBound(g.Conductor) then 
          Print("\nconductor  : ", g.Conductor);
        fi;
      else
        Print("\ngroup is quotient of ray class group of ray:", m);     
        if IsBound(g.Conductor) then 
          Print("\nconductor  : ", g.Conductor);
        fi;
      fi;
    fi;
  fi;
  if printlevel_old >= 2 then
    if IsBound(g.Information) and IsBound(g.Information.Quotients) then
      Print("\nquotients  :");
      for i in [1..Length(g.Information.Quotients)] do
        Print("\n", g.Information.Quotients[i].g2);
      od;    
    fi;
    if IsBound(g.Information) and IsBound(g.Information.QuotientsFrom) then
      Print("\nquotient from:");
      for i in [1..Length(g.Information.QuotientsFrom)] do
        Print("\n", g.Information.QuotientsFrom[i].g2);
      od;    
    fi;
    if IsBound(g.Subgroups) then
      Print("\nsubgroups  :");
      for i in [1..Length(g.Subgroups)] do
        Print("\n", g.Subgroups[i].g2);
      od;    
    fi;
    if IsBound(g.Supergroups) then
      Print("\nsupergroups:");
      for i in [1..Length(g.Supergroups)] do
        Print("\n", g.Supergroups[i].g2);
      od;    
    fi;
    if IsBound(g.Isomorphic) then
      Print("\nisomorphic groups:");
      for i in [1..Length(g.Isomorphic)] do
        Print("\n", g.Isomorphic[i].g2);
      od;    
    fi;
    if IsBound(g.Composition) then
      Print("\ncomposition: ", g.Composition);
    fi;    
  fi;
  AbelianGroupPrintLevel := printlevel_old;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupName
#CALLING   AbelianGroupName(g, s);
#SHORTDOC  Gives a name to a group.
#LONGDOC   Gives the name of the string $s$ to group $g$.
#          The name of $g$ is shown everytime when $g$ is printed.
#          
#ARGUS     g @ group
#          s @ string @ name of $g$
#
#XREF      AbelianGroupPrintLevel
#KASH      @ 
#          g1 := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          g2 := AbelianGroupCreate(Mat(Z, [[2,1], [1,3]]));
#          AbelianGroupName(g1, "first group");
#          AbelianGroupName(g2, "second group");
#          g1;
#          g2;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupName := function(arg)
local g, str;
  
  if Length(arg) <> 2 then
    Error("use AbelianGroupName(group, string)");
  fi;
  g   := arg[1];
  str := arg[2];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup"
          and IsString(str)) then
    Error("use AbelianGroupName(group, string)");
  fi;
  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  g.Information.Name := str;
end;

_AbelianGroupEltEqual := function(a, b)
  local c, g;
# Achtung: Wenn die SNF noch nicht ausgerechnet wurde,
# dann kann nicht reduziert werden und gleiche Gruppenelemente
# mit verschiedenen Exponentenvektoren werden als verchieden
# angesehen !!!!
# Abhilfe: AbelianGroupSmithCreate


  if IsBool(a) or IsBool(b) then return false; fi;

  if (IsBound(a.Group) and not IsBound(b.Group))
      or (IsBound(a.expvec) and not IsBound(b.expvec)) then
    return false;
  fi;
  if (IsBound(b.Group) and not IsBound(a.Group))
      or (IsBound(b.expvec) and not IsBound(a.expvec)) then
    return false;
  fi;

  if IsBound(a.Group) and IsBound(b.Group) then
    if not (a.Group = b.Group) then 
      return false;
    fi;
  fi;

  g := a.Group;
  if not 
    (IsBound(g.Information.IsDirectProduct) and g.Information.IsDirectProduct)
    then
    AbelianGroupSmithCreate(a.Group, false);
  fi;

  if IsBound(a.expvec) and IsBound(b.expvec) then
    if not (a.expvec = b.expvec) then
      c := AbelianGroupEltReduce(a-b);      
      if  not _IsZeroVec(c.expvec) then
        return false;
      fi;
    fi;
  fi;

  return true;
end;

AbelianGroupIsDirectProduct := function(arg)
local g;

  if Length(arg) <> 1 then
    Error("use AbelianGroupIsDirectProduct(group)\n");
  fi;
  g := arg[1];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupIsDirectProduct(group)\n");
  fi;

  if IsBound(g.Information) and
     ((IsBound(g.Information.IsDirectProduct) and g.Information.IsDirectProduct)
       or
      (IsBound(g.Information.IsSmith) and g.Information.IsSmith)) then
    return true;
  else
    return false;
  fi;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupIntersect
#CALLING   g := AbelianGroupIntersect(g1, g2);
#SHORTDOC  Returns the intersection of two groups.
#LONGDOC   Returns the intersection $g$ of the groups $g1, g2$. $g$ is a group.
#          
#ARGUS     g, g1, g2 @ group
#
#XREF      AbelianGroupUnite
#KASH      @ 
#          g0 := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          g1 := AbelianSubGroup(g0, [1,0,0]);
#          g2 := AbelianSubGroup(g0, [[0,1,0], [0,0,5]]);
#          g := AbelianGroupIntersect(g1, g2);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupIntersect := function(arg)
  local i, g, g1, g2, inter, smith_inter, list, A, B, G, K, path;

  if Length(arg) <> 2 then 
    Error("use AbelianGroupIntersect(group, group)\n");
  fi;

  g1 := arg[1];
  g2 := arg[2];

  if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
          and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
    Error("use AbelianGroupIntersect(group, group)\n");
  fi;

  if Length(arg) > 2 then 
    path := arg[3];
  else
    path := false;
  fi;

  if g1 = g2 then return g1; fi;

  g := _AbelianGroupFindCommomSupergroup(g1, g2);
  if g = false then
    Error("no common supergroup found");
  fi;

  A := AbelianGroupGenerators(g1);
  B := AbelianGroupGenerators(g2);
  Apply(A, x->AbelianGroupEltMove(x, g).expvec);
  Apply(B, x->AbelianGroupEltMove(x, g).expvec);
  G := MatToRowList(g.Relations);
  A := Concatenation(A, G);

  IsMat(A);
  IsMat(B);

  K := _AbelianGroupSpecialKernel(A, B);

  list := MatToRowList(K*B);
  Apply(list, x-> AbelianGroupEltCreate(g, x));

  inter := AbelianSubGroup(g, list);
  smith_inter := AbelianGroupSmithCreate(inter);

  if path then
    AbelianGroupIsSub(smith_inter, g1);
    AbelianGroupIsSub(smith_inter, g2);
  fi;

  return smith_inter;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupUnite
#CALLING   g := AbelianGroupUnite(g1, g2);
#SHORTDOC  Returns the smallest group containing two given groups.
#
#LONGDOC   Returns the group $g$ generated by the elements of the groups 
#          $g1, g2$. $g$ is a group.
#          
#ARGUS     g, g1, g2 @ group
#
#XREF      AbelianGroupIntersect
#KASH      @ 
#          g0 := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          g1 := AbelianSubGroup(g0, [1,0,0]);
#          g2 := AbelianSubGroup(g0, [[0,1,0], [0,0,5]]);
#          g := AbelianGroupUnite(g1, g2);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupUnite := function(arg)
  local i, g, g1, g2, path, hom1, hom2, m1, m2, united, smith_united, 
        hom1_chain, hom2_chain;

  if Length(arg) <> 2 then
    Error("use AbelianGroupUnite(group, group)\n");
  fi;

  g1 := arg[1];
  g2 := arg[2];

  if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
          and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
    Error("use AbelianGroupUnite(group, group)\n");
  fi;

  if Length(arg) > 2 then 
    path := arg[3];
  else
    path := false;
  fi;

  if g1 = g2 then return g1; fi;

  g := _AbelianGroupFindCommomSupergroup(g1, g2);
  if g = false then
    Error("no common supergroup found");
  fi;

  hom1_chain := _AbelianGroupFindHom(g1, g, ["Supergroups"]);  
  hom2_chain := _AbelianGroupFindHom(g2, g, ["Supergroups"]);  

  hom1 := hom1_chain[1];
  for i in [2..Length(hom1_chain)] do
    hom1 := hom1_chain[i] * hom1;
  od;
  hom2 := hom2_chain[1];
  for i in [2..Length(hom2_chain)] do
    hom2 := hom2_chain[i] * hom2;
  od;

  m1 := MatToRowList(Copy(hom1.Mat12));
  m2 := MatToRowList(Copy(hom2.Mat12));

  Append(m1, m2);
  IsMat(m1);
  united := AbelianSubGroup(g, m1);
  smith_united := AbelianGroupSmithCreate(united);

  if path then
    AbelianGroupIsSub(g1, smith_united);
    AbelianGroupIsSub(g2, smith_united);
  fi;

  return smith_united;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupDirectProduct
#CALLING   d := AbelianGroupDirectProduct(g1, g2); or\\
#          d := AbelianGroupDirectProduct(L); 
#SHORTDOC  Returns the direct product of groups .
#LONGDOC   In the first case the direct product $d$ of group $g1$ and 
#          group $g2$ is computed. In the second case the direct product 
#          $d$ of all groups in the list is computed. \\
#          Instead of using $\bf{AbelianGroupDirectProduct}$ you may
#          use the * operator (see example).
#          
#ARGUS     g1, g2, d @ group
#          L         @ list @ list of groups  
#
#XREF      
#KASH      compute the direct product using AbelianGroupDirectProduct: @ 
#          g1 := AbelianGroupCreate([[5]]);
#          g2 := AbelianGroupCreate([[7,7],[6,8]]);
#          d := AbelianGroupDirectProduct([g1, g2]);
#          @ compute the direct product using *: @ 
#          d := g1*g2;
#          @ compute the power $\bigoplus_{i = 1}^{4}g_{1}$: @
#          p := g1^4; 
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

_AbelianGroupHomOperations := rec( Print := _AbelianGroupHomPrint,
                                  \^    := _AbelianGroupHomPower,
                                  \*    := _AbelianGroupHomMult,
                                  \+    := _AbelianGroupHomAdd,
                                  \-    := _AbelianGroupHomSubtract
                                );

AbelianGroupDirectProduct := function(arg)
  local g, m, gens, i, j, m2, m1, cols, rows,
        hom, m, v, n1, n2, g1, g2, list, n;

  if Length(arg) = 2 then
    g1 := arg[1];
    g2 := arg[2];

    if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
        and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
      Error("use AbelianGroupDirectProduct(group, group) or\n           AbelianGroupDirectProduct(list)"); 
    fi;
  elif Length(arg) = 1 and IsList(arg[1]) and Length(arg[1]) > 0 then
    list := arg[1];
    if Length(list) = 2 then
      return AbelianGroupDirectProduct(list[1], list[2]);
    elif Length(list) = 1 then
      return list[1];
    else 
      n := Length(list);
      list[n-1] := AbelianGroupDirectProduct(list[n-1], list[n]);
      Unbind(list[n]);      
      return AbelianGroupDirectProduct(list);
    fi;
  else
    Error("use AbelianGroupDirectProduct(group, group) or\n           AbelianGroupDirectProduct(list)"); 
  fi;

  m1 := MatTrans(g1.Relations);
  cols := MatCols(m1); 
  rows := MatRows(m1); 
  for i in [1..MatCols(g2.Relations)] do
    Add(m1, List([1..MatRows(g1.Relations)], j->0));
  od;
  IsMat(m1);
  m1 := MatTrans(m1);

  m2 := [];
  for i in [1..MatCols(g1.Relations)] do
    Add(m2, List([1..MatRows(g2.Relations)], j->0));
  od;
  Add(m2, MatTrans(g2.Relations));
  IsMat(m2);
  m2 := MatTrans(m2);

  m := [m1, m2];
  IsMat(m);
  if g1.Composition = g2.Composition then
    g := AbelianGroupCreate(m, g1.Composition);
  else
    g := AbelianGroupCreate(m);
    g.Composition := ""; # don't know what to do
  fi;
  if IsBound(g1.Generators) and IsBound(g2.Generators) then
    g.Generators := Concatenation(g1.Generators, g2.Generators);
  fi;

  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  if AbelianGroupIsDirectProduct(g1) and AbelianGroupIsDirectProduct(g2) then
    g.Information.IsDirectProduct := true;
  fi;

  g.DiscreteExp := _AbelianGroupStandardDiscreteExp; # ??? ??? ???

  g.Information.DiscLog := [g1, g2];
  g.DiscreteLog := _DirectProductDiscreteLog;

  n1 := MatCols(g1.Relations);
  n2 := MatCols(g2.Relations);

  # homomorphisms g -> g1 and g1 -> g
  # ---------------------------------
  m := MatId(Z, n1);
  m := MatToRowList(m);
  v := [];
  for i in [1..n1] do
    Append(v, [0]);
  od;
  for i in [1..n2] do
    Add(m, v);
  od;
  IsMat(m);
  m := MatTrans(m);

  hom := _AbelianGroupCreateHomGroupToSub(g, g1, m);

  if not IsBound(g.Subgroups) then
    g.Subgroups := [];
  fi;
  Append(g.Subgroups, [hom]);  

  hom := AbelianGroupHomCreate(g1, g, m);
  if not IsBound(g1.Supergroups) then
    g1.Supergroups := [];
  fi;
  Append(g1.Supergroups, [hom]);  

  # homomorphism g -> g2 and g2 -> g
  # --------------------------------
  v := [];
  for i in [1..n2] do
    Append(v, [0]);
  od;
  m := [];
  for i in [1..n1] do
    Add(m, v);
  od;

  m2:=MatId(Z,n2);
  Add(m,m2);
  IsMat(m);
  m := MatTrans(m);

  hom := _AbelianGroupCreateHomGroupToSub(g, g2, m);

  if not IsBound(g.Subgroups) then
    g.Subgroups := [];
  fi;
  Append(g.Subgroups, [hom]);  

  hom := AbelianGroupHomCreate(g2, g, m);
  if not IsBound(g2.Supergroups) then
    g2.Supergroups := [];
  fi;
  Append(g2.Supergroups, [hom]);  

  return g;
end;

_AbelianGroupMult := function(g, h)
  return AbelianGroupDirectProduct(g, h);
end;

_AbelianGroupEqual:=function(g,h)
 local  Opg, Oph, zz;
    if RecFields(g)=RecFields(h) then
       
       Opg:=g.operations;
       Oph:=h.operations; 
        
       Unbind(g.operations);
       Unbind(h.operations); 
       
       zz:=(g=h);
         g.operations:=Opg;
         h.operations:=Oph;       

    return zz; 

    else 
      return false;
    fi;
end;
 

_AbelianGroupPower := function(g, n)
  local h, i, t;

  h := g;
  if IsInt(n) then
    if n > 0 then
      for i in [1..n-1] do
        t := _AbelianGroupMult(h, g);
        Unbind(h); # Zwischenprodukte loeschen
        h := t;
      od;
    else
      h := AbelianGroupCreate([1]);
    fi;
    return h;
  else
    Error("exponent has to be integer"); 
  fi;  
  return h;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupSmithCreate
#CALLING   s := AbelianGroupSmithCreate(g [, generators]);
#SHORTDOC  Returns the smith normal form of a group.
#LONGDOC   Returns the smith normal form $s$ of $g$: the relation matrix
#          of $s$ is in smith normal form. The groups $s$ and $g$ are
#          isomorphic. If the second (optional)
#          parameter $generators$ is $true$ then the new generators are 
#          computed, too.
#          This is useful if you often use the discrete exp. 
#ARGUS     g, s @ group
#          generators @ boolean
#
#XREF      
#KASH      @ 
#          g := AbelianGroupCreate([[15,5,-5,10],[0,0,0,50],[6,6,6,6]]);
#          s := AbelianGroupSmithCreate(g);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB/CF 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupSmithCreate := function(arg)
  local g, WithGens, s, i, j, basis, inverse, orders, a, h, homgh, homhg, HNF,
        r, l, hom, m, elt;

  if Length(arg) > 2 or Length(arg) < 1 then 
    Error("use AbelianGroupSmithCreate(group, [, boolean])\n");
  fi;

  g := arg[1];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupSmithCreate(group, [, boolean])\n");
  fi;

  if Length(arg) >1 then
    if IsBool(arg[2]) then 
      WithGens := arg[2];
    else
      Error("use AbelianGroupSmithCreate(group) or AbelianGroupSmithCreate(group, WithGens");
    fi;
  else 
    WithGens := true;
  fi;
  
  if IsBound(g.Information) then
    if IsBound(g.Information.IsSmith) then
      return g;
    fi;
    if IsBound(g.Information.Smith) then
      return g.Information.Smith.g2;
    fi;
  fi;

  s := MatSmithTrans(g.Relations);
  HNF := MatTrans(MatHermiteColLower(MatTrans(s[2] * s[4]))); 
         # Dies ist HNF der RelationenMatrix g.Relations
  if not IsBound(g.Information) then
    g.Information := rec();
  fi;

  if MatRows(s[2]) = MatCols(s[2]) then 
    g.Information.Order := MatDet(s[2]);
  elif MatRows(s[2])>MatCols(s[2]) then
    i := 1;
    for j in [1..Minimum(MatCols(s[2]), MatRows(s[2]))] do 
      i := s[2][j][j]*i; 
    od;
    g.Information.Order := i;
  else
    g.Information.Order := 0;
  fi;
  inverse := MatInv(s[4]);

  if not IsBound(g.Information) then
    g.Information := rec ();
  fi;

  basis := [];
  orders := List([1..Minimum(MatCols(s[2]),MatRows(s[2]))], i->s[2][i][i]); 
  r := Maximum(0, MatCols(s[2]) - MatRows(s[2])); # f"ur elts ohne Relationen 
                                                  # wird Ordnung Null gesetzt
  if r>0 then
    l := List([1..r], x->0);
  else
    l := [];
  fi;
  Append(orders, l);

  g.Information.Exponent := orders[Length(orders)]; 

  s[4] := MatTrans(s[4]);

  for i in [1..Length(orders)] do
    if orders[i] = 1 then
      s[4][i] := 1;
      inverse[i] := 1;
    fi;
  od;
  s[4] := Filtered(s[4], i->i<>1);
  inverse := Filtered(inverse, i->i<>1);
  orders := Filtered(orders, i->i<>1);

  if orders in [[1], []] then                      # ???
    h := AbelianGroupCreate([[1]], g.Composition); # ???
    orders := [1];
    m := List([1..MatCols(g.Relations)], x->0);
    IsMat(m);
    inverse := m;
    s[4] := m;
  fi;

  h := AbelianGroupCreate(MatDiag(Z, orders), g.Composition);
  # Unbind damit DiscreteExp von g vererbt wird
  Unbind(h.DiscreteExp);

  h.HNF := HNF;
  if not IsBound(h.Information) then
    h.Information := rec();
  fi;
  h.Information.IsSmith := true;
  h.Information.IsDirectProduct := true;
  h.Information.Exponent := g.Information.Exponent;
  h.Information.Order := g.Information.Order;
  h.Information.Orders := orders;

  IsMat(s[4]);
  s[4] := MatTrans(s[4]);
  IsMat(inverse);

  homgh := AbelianGroupHomCreate(g, h, s[4], inverse);
  homhg := AbelianGroupHomCreate(h, g, inverse, s[4]);

  h.Isomorphic := [homhg];
  if not IsBound(g.Isomorphic) then g.Isomorphic := []; fi;
  Add(g.Isomorphic, homgh);
  if not IsBound(g.Information) then
    g.Information := rec ();
  fi;
  if not IsBound(h.Information) then
    h.Information := rec ();
  fi;
  g.Information.Smith     := homgh;  
  h.Information.SmithFrom := homhg; 

  # ist es sinnvoll, die SNF als Ober/Untergruppe einzutragen ???
  if not IsBound(g.Supergroups) then g.Supergroups := []; fi;
  Add(g.Supergroups, homgh);
  if not IsBound(g.Subgroups) then g.Subgroups := []; fi;
  Add(g.Subgroups, homgh);

  if not IsBound(h.Supergroups) then h.Supergroups := []; fi;
  Add(h.Supergroups, homhg);
  if not IsBound(h.Subgroups) then h.Subgroups := []; fi;
  Add(h.Subgroups, homhg);

  # Erzeuger ganz am Schluss berechnen, da dann bereits SNF bekannt.
  # Diese beschleunigt DiscreteExp. 
  if WithGens and IsBound(g.Generators) then
    h.Generators := [];
    for i in [1..MatCols(h.Relations)] do
      elt := AbelianGroupEltCreate(g, inverse[i]);
      h.Generators[i] := AbelianGroupDiscreteExp(elt);
    od;
  fi;

  if IsBound(h.Generators) then
    h.Basis := h.Generators;
    if not IsBound(g.Basis) then
      g.Basis := h.Generators;
    fi;  
  fi;

  if (g = h) then 
    g.Information.IsDirectProduct := true;
  fi;

  # Information  
  if not IsBound(h.Information) then
    h.Information := rec();
  fi;
  if IsBound(g.Information) and IsBound (g.Information.ExpIntModul) then
    h.Information.ExpIntModul := g.Information.ExpIntModul;
  fi;

  if IsBound(g.Information) and IsBound(g.Information.RCG) then
    if not IsBound(h.Information.RCG) then
      h.Information.RCG := rec();
    fi;
    h.Information.RCG.m0 := g.Information.RCG.m0;
    h.Information.RCG.minf := g.Information.RCG.minf;
  fi;

  if IsBound(g.Information) and IsBound(g.Information.RRR) then
    if not IsBound(h.Information.RRR) then
      h.Information.RRR := rec();
    fi;
    h.Information.RRR.m0 := g.Information.RRR.m0;
    h.Information.RRR.minf := g.Information.RRR.minf;
  fi;

  return h;
end;

_AbelianGroupRelationsMatrixCreate := function (group, B)

  return _AbelianGroupSpecialKernel(group.Relations, B);
end;

_AbelianGroupSpecialKernel := function (A, B)
  local i, M, Bl, Ml, L, K, Kt, Xt, XH, XHt, XHl, weiter;
# Hilfsfunktion

  if not (MatCols(B) = MatCols(A)) then
    Error("wrong elements");
  fi;

  Bl :=MatToRowList(B);
  M  :=-A;                  # + oder - ist egal
  Ml :=MatToRowList(M);
  L  :=Concatenation(Bl,Ml);                       
  IsMat(L);

  K:=MatKernel(L);          # main
  if MatRows(K) = 0 then
    K := List([1..MatCols(K)], x -> 0);
    IsMat(K);
  fi;
  Kt:=MatTrans(K);
  Xt:=[];
  for i in [1..MatRows(B)] do
    Append(Xt,[Kt[i]]);
  od;
  IsMat(Xt);
  XHt:=MatHermiteColLower(Xt);
  XH:=MatTrans(XHt);

  XHl:=MatToRowList(XH);
  i := MatRows(XH);
  weiter:=true;
  while weiter and i>=1 do
    if _IsZeroVec(XHl[i]) and Length(XHl)>1 then
      Unbind(XHl[Length(XHl)]);
      i := i-1;
    else 
      weiter := false;
   fi;
  od;
  IsMat(XHl);
  return XHl;
end;

_AbelianGroupCreateHomGroupToSub := function(group, subgroup, B)
local R, M, l, T, group2sub;
# creates the homomorphism from group to subgroup.
# remark: group2sub.Information.Hom12IsProc := true;
# this will be used in _AbelianGroupEltMoveToSub

   group2sub := rec( Type         := "AbelianGroupHom",
                       g1         := group,
                       g2         := subgroup,
                       operations := _AbelianGroupHomOperations
                   );

  if not IsBound(group.Information) then
    group.Information := rec( );
  fi;
  
  if not(IsBound(group2sub.Information)) then
    group2sub.Information := rec ();  
  fi; 

  group2sub.Mat21 := B;
  R := group.Relations;
  M := Copy(B);
  Append(M, R);
  IsMat(M);    

  group2sub.Information.Hermite := rec ();
  group2sub.Information.Hom12IsProc := true;

  l := MatHermiteColLowerTrans(MatTrans(M));
  T := l[3]{[1..MatRows(B)]};
  IsMat(T);
  T := MatTrans(T);
  group2sub.Information.Hermite.Mat   := MatTrans(l[2]); ###        B
  group2sub.Information.Hermite.Trans := T;              ### l[3] * -  = Hermite
  group2sub.Information.Hermite.Rang  := l[1];           ###        R 

  if not (MatTrans(l[3]) * M = MatTrans(l[2])) then
    Error("Bug in AbelianSubGroup\n");
  fi;

  return group2sub;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianSubGroup
#CALLING   s := AbelianSubGroup([g,] L [, generators]); \\
#          s := AbelianSubGroup(g, mat [, generators]);
#SHORTDOC  Creates a subgroup.
#LONGDOC   Creates a subgroup $s$ of group $g$. The subgroup $s$ is generated
#          by the elements of $list$. 
#          If you use $s := {\bf AbelianSubGroup}(g, mat);$ 
#          then the subgroup $s$
#          is generated by the elements represented by the lines of $mat$.
#          If the last (optional) parameter $generators$ is $true$ then
#          the new generators are computed, too. 
#          
#ARGUS     g, s @ group
#          L    @ list   @ list of elements of $g$ or list of exponent vectors
#          mat  @ matrix
#          generators @ boolean
#
#XREF      
#KASH      @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);;
#          @ Create subgroup with list of exponent vectors: @
#          s := AbelianSubGroup(g, [[0,1,0], [0,0,5]]);
#          @ Create subgroup with list of group elements: @
#          elt1 := AbelianGroupEltCreate(g, [0,1,0]);;
#          elt2 := AbelianGroupEltCreate(g, [0,0,5]);;
#          s := AbelianSubGroup([elt1, elt2]);
#          @ Create subgroup with matrix: @
#          mat := Mat(Z, [[0,1,0], [0,0,5]]);;
#          s := AbelianSubGroup(g, mat);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianSubGroup := function(arg)
  local i, l, B, group, elt_list,
        weiter, subgroup, sub2group, group2sub, g, T, R, XHl, M,
        n, m, elt, WithGens;  

  if Length(arg) > 3 or Length(arg) < 1 then
    Error("use AbelianSubGroup([group], elt_list [, WithGens]) or\nuse AbelianSubGroup(group, matrix [, WithGens]");
  fi;    

  if Length(arg) < 2 then
    if IsMat(arg[1]) then
      Error("use AbelianSubGroup(group, matrix)");
    fi;
    if IsList(arg[1]) then
      if arg[1] = [] then 
        Error("use AbelianSubGroup(group, elt_list)");
      fi;
      group := arg[1][1].Group;
      elt_list := arg[1];
    else
      if not IsBound(arg[1].Type) and arg[1].Type = "AbelianGroupElt" then
        Error("use AbelianSubGroup([group], elt_list)");
      fi;
      group := arg[1].Group;
      elt_list := arg[1];
    fi;
  else
    group    := arg[1];
    elt_list := arg[2];
  fi;

  if Length(arg) > 2 then
    if IsBool(arg[3]) then
      WithGens := arg[3];
    else
      Error("use AbelianSubGroup(group, elt_list, WithGens");
    fi;
  else
    WithGens := false;
  fi;

  if IsMat(elt_list) and not elt_list = [] then 
    B := elt_list;
    XHl := _AbelianGroupRelationsMatrixCreate(group, B);
  else
    if not IsList(elt_list) and elt_list.Type = "AbelianGroupElt" then
      elt_list := [elt_list];
    fi;
    if Length(elt_list) > 0 then
      B := [];
      for i in [1..Length(elt_list)] do
        Add(B, elt_list[i].expvec); 
      od;
      IsMat(B);
      # the rows of B represent a system of 
      # generators of the subgroup
      # and represent simultaneous the embedding from sugroup to group;

      XHl := _AbelianGroupRelationsMatrixCreate(group, B);
    else
      XHl := Mat(Z, [[1]]);
      B := [];
      for i in [1..MatCols(group.Relations)] do
        Add(B, 0); 
      od;
      IsMat(B);
    fi;
  fi;

# create subgroup
  subgroup := AbelianGroupCreate(XHl, group.Composition);
  # Unbind damit DiscreteExp von group vererbt wird
  Unbind(subgroup.DiscreteExp);

# install homomorphisms
# for group2sub we need hnf and use procedure _AbelianGroupEltMoveToSub

  if Length(elt_list) > 0 then
    group2sub := _AbelianGroupCreateHomGroupToSub(group, subgroup, B);
  else
    group2sub := rec( Type := "AbelianGroupHom",
              g1 := group,
              g2 := subgroup,
              operations := _AbelianGroupHomOperations
            );

    if not IsBound(group.Information) then
      group.Information := rec( );
    fi;
  
    if not(IsBound(group2sub.Information)) then
      group2sub.Information := rec ();  
    fi; 

    group2sub := AbelianGroupHomCreate(group, subgroup, MatTrans(B));
  fi;

  sub2group := AbelianGroupHomCreate(subgroup, group, B);

  if not IsBound(group.Subgroups) then
    group.Subgroups := [];
  fi;
  Add(group.Subgroups,group2sub);

  if not IsBound(subgroup.Supergroups) then
    subgroup.Supergroups := [];
  fi;
  Add(subgroup.Supergroups,sub2group);

# Generators
  if WithGens and IsBound(group.Generators) then
    if IsBound(group.Generators) then
      n := MatCols(subgroup.Relations);
      m := B;
      subgroup.Generators := [];
      for i in [1..n] do
        elt := AbelianGroupEltCreate(group, B[i]);     
        elt := AbelianGroupDiscreteExp(elt);
        Add(subgroup.Generators, elt);
      od;  
    fi;
  fi;

# Information
  if not IsBound(subgroup.Information) then
    subgroup.Information := rec();
  fi;
  if IsBound(group.Information) and IsBound(group.Information.ExpIntModul) then
    subgroup.Information.ExpIntModul := group.Information.ExpIntModul;
  fi;
   subgroup.parentgroup:=group;
  #########
  # Eintrag parentgroup gibt Gruppe, von der subgroup die Untergruppe ist
  ##########
  return subgroup;
end;


_AbelianGroupApplyHomChain := function (chain, elt)
  local i;

  if chain = false then
    return false;
  fi;  

  i:=1;
  while i <= Length(chain) do
    if chain[i].Type = "AbelianGroupHom" then
      elt := chain[i]*elt;
      if elt = false then return false; fi;
    else
      Error("Bug in MoveSystem!\n");
    fi;
    i := i+1;
  od;
  return elt;  
end;

_AbelianGroupFindDirectProduct := function(g)
  local chain, grps, id, i, look;

  chain := [];
  grps  := [];

  look := function(g)
    local l, i, id, weiter;

    if chain = [] and IsBound(g.Information) and 
                      IsBound(g.Information.IsDirectProduct) then
      if g.Information.IsDirectProduct then
        id := AbelianGroupHomCreate(g, g, MatId(Z, MatCols(g.Relations)));
        
        chain := [id];
        return false;
      fi;
    fi;

    if Position(grps, g)=false then
      Append(grps, [g]);  
    else 
      return true;    
    fi;

    if not IsBound(g.Isomorphic) then
      return true;
    else
      l := Length(g.Isomorphic);
      i := 1;
      weiter := true;
      while weiter and (i <= l) do        
        if Position(grps, g.Isomorphic[i].g2)=false then 
          Add(chain, g.Isomorphic[i]);          
          if AbelianGroupIsDirectProduct(g.Isomorphic[i].g2) then 
            weiter := false;
          else
            weiter := look(g.Isomorphic[i].g2);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
      return weiter;
    fi;
  end;

  if not look(g) then
    return chain;
  else
    return false;
  fi;
end;

_AbelianHomGroupDiscreteLog:=function(g, hom)
local g1, g2, Lhom, elt, vec, product_group, product_subgroup; 

  g1 := hom.g1;
  g2 := hom.g2;

  if (not IsBound(g.Information.HomGroup)) or
     (not IsBound(hom.Type)) or (hom.Type <> "AbelianGroupHom") or
     g.Information.HomGroup[1] <> hom.g1 or
     g.Information.HomGroup[2] <> hom.g2 then
       Error("different groups\n");
  fi;

  product_group    := g.Information.DiscLog[1];
  product_subgroup := g.Information.DiscLog[2];

  Lhom := List(AbelianGroupGenerators(g1), x->hom*x);
  Apply(Lhom, x->MatToRowList(x.expvec)[1]);
   
  Lhom := Flat(Lhom); 
  elt := AbelianGroupEltCreate(product_group, Lhom);

  elt := AbelianGroupEltMove(elt, product_subgroup);
  if (elt = false) then
    if AbelianGroupPrintLevel > 1 then
      Print("function not contained in group of homomorphisms\n");
    fi;
    return false;
  fi;
   
  elt := AbelianGroupEltCreate(g, elt.expvec);

  return elt;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianHomGroup
#CALLING   h := AbelianHomGroup(g1, g2);
#SHORTDOC  Returns the group consisting of homomorphisms from $g1$ to $g2$.
#ARGUS     g1, g2, h @ group
#
#XREF      AbelianGroupHomCreate     
#KASH      @ 
#          g1 := AbelianGroupCreate([[1,2,3],[2,4,0], [2,0,0]]);;
#          g2 := AbelianGroupCreate([[1,0,0],[0,4,0], [1,1,1]]);;
#          h := AbelianHomGroup(g1, g2);
#          @ $elt$ is an element of h: @
#          elt := AbelianGroupEltCreate(h, [0,1]);
#          @ $hom$ is a homomorphism from $g1$ to $g2$: @
#          hom := AbelianGroupDiscreteExp(elt);
#          elt1 := AbelianGroupEltCreate(g1, [1,1,0]);
#          elt2 := hom*elt1;  
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianHomGroup := function(arg)
  local g, gens, rels, i, j, k, n, h1, h2, m, g1, g2, product_group,
        product_subgroup, g_gens_exp, g1_gens, product_elts;

  if Length(arg) <> 2 then
    Error("use AbelianHomGroup(group, group)\n");
  fi;
  g1 := arg[1];
  g2 := arg[2];

  if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
          and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
    Error("use AbelianHomGroup(group, group)\n");
  fi;

  h1 := AbelianGroupSmithCreate(g1, true);
  h2 := AbelianGroupSmithCreate(g2, true);
  m := List([1..Length(h1.Relations[1])],
                             i->List([1..Length(h2.Relations[1])], j->0));
  IsMat(m);
  gens := [];
  rels := [];

  for i in [1..Length(h1.Relations[1])] do
    for j in [1..Length(h2.Relations[1])] do
      if h1.Relations[i][i] = 0 then
        # Die Faktoren ZZ m"ussen ber"ucksichtigt werden
        n := Copy(m);
        n[i][j] := 1;
        Add(gens, h2.Information.SmithFrom*AbelianGroupHomCreate(h1, h2, n)*
                           g1.Information.Smith);
      elif h2.Relations[j][j] <> 0 then
        k := h2.Relations[j][j]/Gcd(h1.Relations[i][i], h2.Relations[j][j]);
        if k<h2.Relations[j][j] then
          n := Copy(m);
          n[i][j] := k;
          Add(gens, h2.Information.SmithFrom*AbelianGroupHomCreate(h1, h2, n)*
                             g1.Information.Smith);
          Add(rels, h2.Relations[j][j]/k);
        fi;
      fi;
    od;
  od;
  if gens = [] then
    g := AbelianGroupCreate([[1]], "+");
    m := List([1..Length(g1.Relations[1])],
                             i->List([1..Length(g2.Relations[1])], j->0));
    g.Generators := [AbelianGroupHomCreate(g1, g2, m)];
  else
    if rels = [] then rels := [0]; fi;
    g := AbelianGroupCreate(MatDiag(Z, rels), "+");
    g.Generators := gens;
    if not IsBound(g.Information) then
      g.Information := rec();
    fi;
    g.Information.IsDirectProduct := true;
  fi;

  g.Information.HomGroup := [g1, g2];

  product_group := g2^AbelianGroupNumberGenerators(g1);
  g_gens_exp    := AbelianGroupGenerators(g, true);
  g1_gens       := AbelianGroupGenerators(g1);
  product_elts  := List(g_gens_exp, x->List(g1_gens, y->x*y));
  Apply(product_elts, x->Flat(List(x, y->MatToRowList(y.expvec)[1])));
  Apply(product_elts, x->AbelianGroupEltCreate(product_group, x));
  product_subgroup := AbelianSubGroup(product_group, product_elts);
  g.DiscreteLog := _AbelianHomGroupDiscreteLog;
  g.Information.DiscLog := [product_group, product_subgroup];
 
  return g;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianDualGroup
#CALLING   d := AbelianDualGroup(g);
#SHORTDOC  Returns the dual group.
#LONGDOC   Returns the dual group $d$ of group $g$. This is the group
#          consisting of homomorphisms from group $g$ to $\C$.
#          Here the homomorphisms are from $g$ to 
#          the rationals $\Q$. A rational number $q$ corresponds to
#          the complex number $e^{(i*2*\pi*q)}$. 
#       
#ARGUS     g, d @ group
#
#XREF      
#KASH      compute the dual group and apply one homomorphism: @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          d := AbelianDualGroup(g);
#          eltd := AbelianGroupEltCreate(d, [1]);
#          hom := AbelianGroupDiscreteExp(eltd);
#          eltg := AbelianGroupEltCreate(g, [1,0,0]);
#          AbelianGroupDiscreteExp(hom*eltg);
#          @ it is allowed to apply an elelment of the group to
#          an element of the correspondig dual group: @
#          eltd*eltg;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianDualGroup := function(arg)
  local i, h, dualg, orders, facs, fac, primes, primeset,
        neworders, M, maxlist, max, p, g;

  if Length(arg) <> 1 then
    Error("use AbelianDualGroup(group)\n");
  fi;
  g := arg[1];

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianDualGroup(group)\n");
  fi;

  if AbelianGroupOrder(g) = 0 then
    Error("group has element of order zero\n");
  fi;

  h := AbelianGroupSmithCreate(g, false);
  orders := [];
  for i in [1..MatCols(h.Relations)] do
    Add(orders, h.Relations[i][i]);
  od;
  
  facs := [];
  for i in [1..Length(orders)] do
    Append(facs, Factor(orders[i]));
  od;
  primes := List(facs, i->i[1]);
  primeset := Set(primes);
  maxlist := [];
  for p in primeset do
    max := 0;
    for fac in facs do
      if fac[1] = p and fac[2] > max then
        max := fac[2];
      fi;
    od;
    Add(maxlist, max);
  od;
  neworders := List([1..Length(primeset)], i->primeset[i]^maxlist[i]);
  M := MatDiag(Z, neworders);

  h := AbelianGroupCreate(M, "+"); 

  # Die Erzeuger liegen in Q und man erh"alt die zugeh. complexen Zahlen
  # mittels q -> e^(i*q*2*pi)
  h.Generators := [];
  for i in [1..MatCols(h.Relations)] do
    Add(h.Generators, 1/h.Relations[i][i]);
  od;
  
  dualg := AbelianHomGroup(g, h);
  if not IsBound(dualg.Information) then
    dualg.Information := rec();
  fi;
  dualg.Information.DualFrom := g;
  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  g.Information.Dual := dualg;

  return dualg;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupEltOrder
#CALLING   a := AbelianGroupEltOrder(elt);
#SHORTDOC  Returns the order of an element.
#LONGDOC   Returns the order of the element $elt$ in its group.
#          
#ARGUS     elt @ group element
#          a   @ integer
#
#XREF      AbelianGroupOrder, AbelianGroupEltCreate
#KASH      @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          elt := AbelianGroupEltCreate(g, [1,0,0]);
#          a := AbelianGroupEltOrder(elt);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupEltOrder := function(arg)
  local g1, i, chain, a;

  if Length(arg) <> 1 then
    Error("use AbelianGroupEltOrder(elt)");
  fi;
  a := arg[1];

  g1 := a.Group;
  chain := _AbelianGroupFindDirectProduct(g1);
  if chain = false then
    AbelianGroupSmithCreate(g1, false);
    chain := _AbelianGroupFindDirectProduct(g1);
  fi;
  a := _AbelianGroupApplyHomChain(chain, a);
  chain := 1;
  for i in [1..MatCols(a.expvec)] do
    if a.expvec[1][i]<>0 then
      chain := IntLcm(chain, a.Group.Relations[i][i]/
                              IntGcd(a.Group.Relations[i][i], a.expvec[1][i]));
    fi;
  od;
  return chain;
end;

_AbelianGroupIsHom := function (arg)
  local i, im, R, M;

  if Length(arg)<=0 or Length(arg)>1 then
    return false;
  fi;
  
  if IsBound(arg[1].Type) and arg[1].Type = "AbelianGroupHom" then
    im := AbelianGroupHomImage(arg[1]);
    if IsBound(im.Information) and IsBound(im.Information.Smith) and 
       IsBound(im.Information.Smith.g2) and 
       IsBound(im.Information.Smith.g2.HNF) then 
      R := im.Information.Smith.g2.HNF;
    else 
      R := MatTrans(MatHermiteColLower(MatTrans(im.Relations)));
    fi;
    M := Copy(arg[1].g1.Relations);
    Add(M, R);
    IsMat(M);
    M := MatTrans(MatHermiteColLower(MatTrans(M)));
    
    for i in [1..MatRows(R)] do
      if not (R[i] = M[i]) then
        return false;
      fi;
    od;
    if MatRows(M) = MatRows(R) or _IsZeroVec(M[MatRows(R)+1]) then
      return true; # Die geforderten Relationen sind enthalten  
    else 
      return false;
    fi;
  fi;
  return false;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupHomCreate
#CALLING   hom := AbelianGroupHomCreate(g1, g2, mat [, check]); \\
#          hom := AbelianGroupHomCreate(g1, g2, mat, [matinv]);
#SHORTDOC  Creates the homomorphism corresponding to a matrix.
#LONGDOC   Creates the homomorphism from group $g1$ to group $g2$ 
#          corresponding to matrix $mat$.
#          The columns of $mat$ are the images of the generators of $g1$.
#          If the fourth (optional) parameter
#          $check$ is true then it is checked whether the object $hom$ is a 
#          homomorphism.
#          If the fourth parameter is a matrix then it has to be the
#          inverse of $mat$.
#          
#ARGUS     g1, g2 @ group
#          hom    @ homomorphism @ homomorphism from $g1$ to $g2$
#          check  @ boolean
#          matinv @ matrix
#
#XREF      AbelianHomGroup
#KASH      @ 
#          g1 := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);;
#          g2 := AbelianGroupCreate([[0,2],[3,0]]);;
#          mat := Mat(Z, [[-24,24],[20,-20],[-10,10]]);;  
#          hom := AbelianGroupHomCreate(g1, g2, mat);
#          @ if you are not sure, use the fourth parameter $check$ to test 
#          whether the matrix corresponds to a homomorphism: @
#          mat := Mat(Z, [[3,0],[3,-2],[-1,2]]);;
#          hom := AbelianGroupHomCreate(g1, g2, mat, true);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupHomCreate := function(arg)
  local h;
  
  if (Length(arg) < 3) or (Length(arg) > 4) then
      Error("use AbelianGroupHomCreate(group1, group2, Matrix, [check]) or                      AbelianGroupHomCreate(group1, group2, Matrix, [Matrix])");
  fi;

  if Length(arg) >= 3 then
    if not (    arg[1].Type = "AbelianGroup" 
            and arg[2].Type = "AbelianGroup"
            and (IsMat(arg[3]) 
                 and MatRows(arg[3]) = MatCols(arg[1].Relations)
                 and MatCols(arg[3]) = MatCols(arg[2].Relations))
            or IsFunc(arg[3])
            ) then
      Error("use AbelianGroupHomCreate(group1, group2, Matrix, [, check]) or                    AbelianGroupHomCreate(group1, group2, Matrix, [, Matrix]) or                   AbelianGroupHomCreate(group1, group2, Matrix, function)");
    fi;
  fi;

  if Length(arg) >= 3 and IsFunc(arg[3]) then 
    return _AbelianGroupHomFromFunc(arg[1], arg[2], arg[3]);
  fi;

  h := rec( Type := "AbelianGroupHom",
            g1 := arg[1],
            g2 := arg[2],
            operations := _AbelianGroupHomOperations
          );

  if Length(arg)>2 then
    

       
      h.Mat12 := arg[3];  
     

  fi;

  if Length(arg)>3 and IsMat(arg[4]) then
    h.Mat21 := arg[4];
  fi;

  if Length(arg) = 4 and IsBool(arg[4]) and arg[4] then
    if not _AbelianGroupIsHom(h) then
      if AbelianGroupPrintLevel >1 then
        Print("function is no homomorphism\n");
      fi;  
      return false;
    fi;
  fi;



     return h;
    

 

end;

_AbelianGroupHomFromFunc := function(g1, g2, func)
# func has to be a homomorphic function from g1 to g2.
# returns the homomorphism from abstract group g1 to abstract group g2
# corresponding to func
local i, n, elt, mat;

  if not (IsBound(g1.Generators) and IsBound(g2.DiscreteLog)) then
    Error("no information about generators or discrete log"); 
  fi;

  n := MatCols(g1.Relations);

  mat := [];
  for i in [1..n] do
    elt := g1.Generators[i];
    elt := func(elt);
    mat[i] := AbelianGroupDiscreteLog(g2, elt).expvec;
  od;

  IsMat(mat);
  return AbelianGroupHomCreate(g1, g2, mat);
end;

_AbelianGroupOperations := rec( 
                               Print := _AbelianGroupPrint,
                               \*    := _AbelianGroupMult,
                               \^    := _AbelianGroupPower
                              # \=    := _AbelianGroupEqual
                              );

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupEqual
#CALLING   test := AbelianGroupEqual(g1, g2);
#
#SHORTDOC  Tests if the groups are equal.
#LONGDOC   {\bf AbelianGroupEqual} returns `true` iff the groups $g1$ and
#          $g2$ are contained in each other. Otherwise it returnes `false`.
#          
#ARGUS     g1, g2 @ group
#          test   @ boolean
#
#XREF      
#KASH      @ 
#          g1 := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          g2 := AbelianSubGroup(g1, [1,1,1]);
#          test := AbelianGroupEqual(g1, g2);
#          @ note: @
#          g1 = g2;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupEqual := function(arg)
  local g, g1, g2; 

  if Length(arg) <>2 then
    Error("use AbelianGroupEqual(group, group)\n");
  fi;
  g1 := arg[1];
  g2 := arg[2];

  if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
          and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
    Error("use AbelianHomGroup(group, group)\n");
  fi;

  g := _AbelianGroupFindCommomSupergroup(g1, g2);
  if g = false then return false; fi;

  return (AbelianGroupIsSub(g1, g2) and AbelianGroupIsSub(g2, g1));
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupCreate
#CALLING   g := AbelianGroupCreate(mat); \\
#          g := AbelianGroupCreate(mat, "*"); \\
#          g := AbelianGroupCreate(mat, "+");
#SHORTDOC  Creates the group with a given relation matrix.
#LONGDOC   Creates the group $g$ with the relation matrix $mat$. 
#          Use "*" as second parameter to create a multiplicative group and
#          "+" to create an additive group. This is needed by the 
#          discrete exp. The default is "*". 
#          
#ARGUS     g @ group
#          mat @ matrix or list of lists
#
#XREF      AbelianSubGroup, AbelianGroupSmithCreate
#KASH      @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupCreate := function(arg)
  local g, M;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use AbelianGroupCreate( matrix )");
  fi;

  g := rec(Type := "AbelianGroup",
           operations := _AbelianGroupOperations,
           Composition := "*",
           DiscreteExp := _AbelianGroupStandardDiscreteExp
          );

  if IsMat(arg[1]) then
    g.Relations := Copy(arg[1]);
    g.Information := rec();
    M:=MatDiag(Z,
   List([1..Minimum(MatRows(g.Relations),
                  MatCols(g.Relations))],z->g.Relations[z][z]));
    if g.Relations=M  then
      g.Information.IsDirectProduct := true;
    fi;
    if Length(arg) = 2 then
      if IsString(arg[2]) and (arg[2] = "+" or arg[2] = "*" or arg[2] = "") then
        g.Composition := arg[2];
      else
        Error("only AbelianGroupCreate( matrix ) or AbelianGroupCreate( matrix , string )supported");
      fi;
    fi;
  else
    Error("use AbelianGroupCreate( matrix )");
  fi;

  return g;
end;

_AbelianGroupEltOperations := rec( Print := _AbelianGroupEltPrint,
                                  \*    := _AbelianGroupHomMult,
                                  \+    := _AbelianGroupEltAdd,
                                  \-    := _AbelianGroupEltSubtract,
                                  \=    := _AbelianGroupEltEqual
                                );

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupEltCreate
#CALLING   elt := AbelianGroupEltCreate(g, vector);
#SHORTDOC  Creates an element in a group.
#LONGDOC   Creates the element $elt$ of $g$ with the exponent vector $vector$.
#          
#ARGUS     elt    @ group element
#          g      @ group
#          vector @ list or matrix wiht one row 
#
#XREF      AbelianGroupEltReduce
#KASH      create an element in an abstract group: @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          elt := AbelianGroupEltCreate(g, [3,2,1]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupEltCreate := function(arg)
  local a, g, exponentvector;

  if Length(arg) <> 2 then
    Error("use AbelianGroupEltCreate(group, exponentvector)\n");
  fi;
  g              := arg[1];
  exponentvector := Copy(arg[2]);

  if not (IsRec(g) and IsBound(g.Type) and g.Type = "AbelianGroup") then
    Error("use AbelianGroupEltCreate(group, exponentvector)\n");
  fi;
  
  if not(IsMat(exponentvector) 
          and MatCols(exponentvector) = MatCols(g.Relations)
          and MatRows(exponentvector) = 1) then
    Error("wrong length of exponentvector");
  fi;  
  a := rec(Type := "AbelianGroupElt",
           Group := g,
           operations := _AbelianGroupEltOperations,
           expvec := exponentvector
          );
  return a;
end;

_AbelianGroupFindSupergroups := function(arg) 
  local i, g, list;
# changes second argument (list)

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use _AbelianGroupFindSupergroups(group [, list])");
  fi;


  g := arg[1];
  if Length(arg) >=2 then
    list := arg[2];                             # changes list
  else 
    list := [];
  fi;

  if  IsBound(g.Supergroups) then
    for i in [1..Length(g.Supergroups)] do 
      if not (g.Supergroups[i].g2 in list) then 
        Add(list, g.Supergroups[i].g2);         # changes list
        list := _AbelianGroupFindSupergroups(g.Supergroups[i].g2, list);
      fi;
    od;
  fi;

  return list;
end;


_AbelianGroupFindCommomSupergroup := function(arg)
  local l1, l2, g, h, s1, s2;

  if Length(arg) <> 2 then
    Error("use _AbelianGroupFindCommomSupergroup(group, group)");
  fi;
  s1 := arg[1];
  s2 := arg[2];

  l1 := _AbelianGroupFindSupergroups(s1, [s1]);
  l2 := _AbelianGroupFindSupergroups(s2, [s2]);

  for g in l1 do
    for h in l2 do
      if g = h then  
        return g;
      fi;
    od;
  od;
 
  return false;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupIsSub
#CALLING   test := AbelianGroupIsSub(s, g);
#SHORTDOC  Tests whether $s$ is subgroup of $g$.
#LONGDOC   Tests whether group $s$ is subgroup of group $g$. {\bf AbelianGroupIsSub}
#          returns $true$ or $false$.
#          
#ARGUS     g, s @ group
#          test @ boolean
#
#XREF      AbelianSubGroup
#KASH      @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          s1 := AbelianSubGroup(g, [1,0,0]);
#          s2 := AbelianSubGroup(g, [[0,1,0], [0,0,5]]);
#          AbelianGroupIsSub(s1, g);
#          AbelianGroupIsSub(s1, s2);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupIsSub := function(arg)
  local i, j, g1, g2, super, n, m, a, g, A, weiter, is_sub,
        sub2group, group2sub;
# Returns true if g1 is subgroup of g2 and false ohterwise.
# In the first case dircet homomorphisms are installed.
# This is useful if you often move between g1 and g2.
#
# 3 args : third arg is a known common Supergroup of g1 and g2

  if Length(arg) > 3 or Length(arg) < 2 then
    Error("use AbelianGroupIsSub(group, group)\n");
  fi;

  g1 := arg[1];
  g2 := arg[2];

  if not (IsRecType(g1, "AbelianGroup")
          and IsRecType(g2, "AbelianGroup")) then
    Error("use AbelianGroupIsSub(group, group)\n");
  fi;

  if Length(arg) > 2 then
    super := arg[3];
  else
    #lets cheat: the most common case is if g1 is a direct subgrou of g2:
    if IsBound(g1.parentgroup) and g1.parentgroup = g2 then
      return true;
    fi;
    super := _AbelianGroupFindCommomSupergroup(g1, g2);
  fi;

  if IsBool(super) and not super then
    return false;
  fi;

  j := 1;
  n := MatCols(g1.Relations);
  m := MatId(Z, n);
  A := [];

  while j <= n do
    a := AbelianGroupEltCreate(g1, m[j]);
    a := AbelianGroupEltMove(a, super, ["Subgroups", "Supergroups", "Smith", "SmithFrom"]);
    a := AbelianGroupEltMove(a, g2, ["Subgroups", "Supergroups", "Smith", "SmithFrom"]);
    if a = false then
       return false;   
    else
      Add(A, a.expvec);  
    fi;    
    j := j+1;
  od;

  IsMat(A);

  sub2group := AbelianGroupHomCreate(g1, g2, A);
  if not IsBound(g1.Supergroups) then
    g1.Supergroups := [];
  fi;
  Add(g1.Supergroups, sub2group);


  if not IsBound(g2.Subgroups) then
    g2.Subgroups := [];
  fi;

  group2sub := _AbelianGroupCreateHomGroupToSub(g2, g1, A);

  Add(g2.Subgroups, group2sub);

  return true;
end;

_AbelianGroupFindHom := function(arg)
  local g1, g2, string_list, chain, grps, i, look, l, string, g;

  if Length(arg) > 3 or Length(arg) < 2 then
    Error("use _AbelianGroupFindHom(group, group [, list])");
  fi;

  if not(Length(arg)>=2 and IsBound(arg[1].Type) and IsBound(arg[2].Type) and 
         arg[1].Type = "AbelianGroup" and arg[2].Type = "AbelianGroup") then
    Error("use _AbelianGroupFindHom(group, group, [\"Subgroups\", \"Supergroups\", \"Smith\", \"SmithFrom\", \"Quotients\", \"QuotientsFrom\"])\n");
  fi;

  g1 := arg[1];
  g2 := arg[2];

  if Length(arg) = 3 then
    string_list := arg[3];
  else
    string_list := ["Subgroups", "Supergroups", "Smith", "SmithFrom", 
                    "Quotients", "QuotientsFrom"];
  fi;

  if g1 = g2 then
    return [AbelianGroupHomCreate(g1,g1,MatId(Z, MatCols(g1.Relations)))];
  fi;

  look := function(g1, g2, direction)
    local l, i, id, weiter;

    if chain = [] then
      if g1 = g2 then
        id := AbelianGroupHomCreate(g1, g2, MatId(Z, MatCols(g1.Relations)));
        chain := [id];
        return false;
      fi;
    fi;
    if Position(grps, g1)=false then
      Append(grps, [g1]);  
    fi;
    weiter := true;

    if "Supergroups" in direction and IsBound(g1.Supergroups) then
      l := Length(g1.Supergroups);
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.Supergroups[i].g2)=false then 
          Add(chain, g1.Supergroups[i]);          
          if g1.Supergroups[i].g2 = g2 then 
            return false;
          else
            weiter := look(g1.Supergroups[i].g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false; fi;
    if "Subgroups" in direction and IsBound(g1.Subgroups) then
      # the same for "Subgroups"
      l := Length(g1.Subgroups);
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.Subgroups[i].g2)=false then 
          Add(chain, g1.Subgroups[i]);          
          if g1.Subgroups[i].g2 = g2 then 
            return false;
          else
            weiter := look(g1.Subgroups[i].g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false; fi;
    if "Quotients" in direction and IsBound(g1.Quotients) then
      # the same for "Quotients"
      l := Length(g1.Quotients);
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.Quotients[i].g2)=false then 
          Add(chain, g1.Quotients[i]);          
          if g1.Quotients[i].g2 = g2 then 
            return false;
          else
            weiter := look(g1.Quotients[i].g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false; fi;
    if "QuotientsFrom" in direction and IsBound(g1.QuotientsFrom) then
      # the same for "QuotientsFrom"
      l := Length(g1.QuotientsFrom);
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.QuotientsFrom[i].g2)=false then 
          Add(chain, g1.QuotientsFrom[i]);          
          if g1.QuotientsFrom[i].g2 = g2 then 
            return false;
          else
            weiter := look(g1.QuotientsFrom[i].g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false; fi;
    if "Smith" in direction and IsBound(g1.Information) 
       and IsBound(g1.Information.Smith) then
      # the same for "Smith"
      l := 1;
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.Information.Smith.g2)=false then 
          Add(chain, g1.Information.Smith);          
          if g1.Information.Smith.g2 = g2 then 
            return false;
          else
            weiter := look(g1.Information.Smith.g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false; fi;
    if "SmithFrom" in direction and IsBound(g1.Information)
       and IsBound(g1.Information.SmithFrom) then
      # the same for "SmithFrom"
      l := 1;
      i := 1;
      while weiter and (i <= l) do        
        if Position(grps, g1.Information.SmithFrom.g2)=false then 
          Add(chain, g1.Information.SmithFrom);          
          if g1.Information.SmithFrom.g2 = g2 then 
            return false;
          else
            weiter := look(g1.Information.SmithFrom.g2, g2, direction);
          fi;
          if weiter then Unbind(chain[Length(chain)]); fi; 
        fi;
        i := i+1;
      od;
    fi;
    if not weiter then return false;
    else return true; fi;
  end;

  chain := [];
  grps  := [];
  if not look(g1, g2, string_list) then
    return chain;
  fi;

  if AbelianGroupPrintLevel > 0 then
    Print("no path found\n");  
  fi;
  return false;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupEltMove
#CALLING   elt2 := AbelianGroupEltMove(elt1, g);
#SHORTDOC  Moves an element from one group to another group.
#LONGDOC   Returns the representation $elt2$ of $elt1$ in group $g$.
#          
#ARGUS     g   @ group
#          elt1 @ group element
#          elt2 @ group element @ element of group $g$ 
#
#XREF      
#KASH      move an element from subgroup $s$ to group $g$: @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          s := AbelianSubGroup(g, [3,3,3]);
#          elt1 := AbelianGroupEltCreate(s, [2]);
#          elt2 := AbelianGroupEltMove(elt1, g);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupEltMove := function(arg)
  local hom_chain, g1, g1elt, g2, g2elt, g, string_list;

  if Length(arg) < 2 or Length(arg) > 3 then
    Error("use AbelianGroupEltMove(element, group)\n");
  fi;

  g1elt := arg[1];
  g2 := arg[2];

  if not (IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup") then
    Error("use AbelianGroupEltMove(element, group)\n");
  fi;

  if Length(arg) > 2 then
    string_list := arg[3];
  else 
    string_list := [];  
  fi;

  if (IsBool(g1elt) and not g1elt) or 
      (not IsBound(g2.Type) or not g2.Type = "AbelianGroup") then
    Error("use AbelianGroupEltMove(element, group)");
  fi;

  g1 := g1elt.Group;
  if g1 = g2 then 
    return g1elt;
  fi;
  
  if string_list = [] then
    hom_chain := _AbelianGroupFindHom(g1, g2);
  else
    hom_chain := _AbelianGroupFindHom(g1, g2, string_list);
  fi;

  g2elt := _AbelianGroupApplyHomChain(hom_chain, g1elt);
  if g2elt = false then
    return false;
  fi;
  return g2elt; 
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianQuotientGroup
#CALLING   q := AbelianQuotientGroup(g, s);
#SHORTDOC  Returns the quotient of a group and a subgroup.
#LONGDOC   Returns the quotientgroup $q$ of group $g$ and subgroup $s$.
#          
#ARGUS     g, q @ group
#          s    @ group @ subgroup of $g$
#
#XREF      AbelianGroupCanonicalQuotient
#KASH      @ 
#          g := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);
#          s := AbelianSubGroup(g, [3,3,3]);
#          q := AbelianQuotientGroup(g, s);
#          AbelianGroupSmithCreate(q);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianQuotientGroup := function (arg)
  local i, hom, P, D, hom_chain, quotient, m,
        j, n, A, weiter, a, gens, group, sub;

  if Length(arg) <> 2 then 
    Error("use AbelianQuotientGroup(group, group)\n");
  fi;
  group := arg[1];
  sub   := arg[2];

  if not (IsRec(group) and IsBound(group.Type) and group.Type = "AbelianGroup"
       and IsRec(sub) and IsBound(sub.Type) and sub.Type = "AbelianGroup") then
    Error("use AbelianQuotientGroup(group, group)\n");
  fi;

  if  not AbelianGroupIsSub(sub, group) then 
    Error("use AbelianQuotientGroup(group, sub group)");
  fi;

  hom_chain := _AbelianGroupFindHom(sub, group, ["Supergroups"]); 
  # finds the shortest way because of AbelianGroupIsSub

  hom := hom_chain[1];

  for i in [2..Length(hom_chain)] do
    hom := hom_chain[i] * hom; 
  od;  
  P := Copy(hom.Mat12);       

  D := MatToRowList(group.Relations); 
  Append(P, D);
  if group.Composition = sub.Composition then
    quotient := AbelianGroupCreate(P, group.Composition);

    quotient.DiscreteExp := function(elt)
    local h;
  
      h := elt.Group.QuotientsFrom[1].g2;
      elt := AbelianGroupEltMove(elt, h);
      return AbelianGroupDiscreteExp(elt);
    end;

    quotient.DiscreteLog := function(g, a)
    local h, elt;

      h := g.QuotientsFrom[1].g2;
      elt := AbelianGroupDiscreteLog(h, a);
      elt.Group := g;
      return elt;
    end;
  else
    Error("not supported"); # do not know what to do
  fi;

  # erst Information.RCG dann SNF, damit Information vererbt wird
  if IsBound(group.Information) and IsBound(group.Information.RCG) 
     and not IsBound(group.Information.RCG.Sub) then
    if not IsBound(quotient.Information) then
      quotient.Information := rec();
    fi;
    if not IsBound(quotient.Information.RCG) then
      quotient.Information.RCG := rec();
    fi;

    quotient.Information.RCG.m0 := group.Information.RCG.m0;
    quotient.Information.RCG.minf := group.Information.RCG.minf;
    quotient.Information.RCG.Sub := sub;
  fi;

  AbelianGroupSmithCreate(quotient, false);

  m := MatId(Z, MatCols(group.Relations));
  hom := AbelianGroupHomCreate(group, quotient, m);
  if not IsBound(group.Quotients) then
    group.Quotients := [];
  fi;
  Add(group.Quotients, hom);
  hom := AbelianGroupHomCreate(quotient, group, m);
  if not IsBound(quotient.QuotientsFrom) then
    quotient.QuotientsFrom := [];
  fi;
  Add(quotient.QuotientsFrom, hom);

  gens := AbelianGroupGenerators(group, true);
  if not(gens[1]=false) then
    quotient.Generators := gens; # Repr"asentanten
  fi;

  if IsBound(group.DiscreteLog) then
    quotient.DiscreteLog := function(g, elt)
      local a, b, qf;

      qf := g.QuotientsFrom[1].g2;
      a := qf.DiscreteLog(qf, elt);
      b := AbelianGroupEltMove(a, g);
      b := AbelianGroupEltReduce(b);

      return b;
    end;
  fi;
     quotient.subgroup:=sub;
     quotient.group:=group;
  return quotient;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupHomImage
#CALLING   g := AbelianGroupHomImage(hom [, generators]);
#SHORTDOC  Returns the image of a homomorphism.
#LONGDOC   Returns the image $g$ of the homomorphism $hom$. $g$ is a group.
#          If the second (optional) parameter $generators$ is $true$ then
#          the generators of the image $g$ are generated, too.
#          
#ARGUS     g   @ group
#          hom @ homomorphism
#          generators @ boolean
#
#XREF      AbelianGroupKernel, AbelianQuotientGroup, 
#          AbelianGroupCanonicalQuotient
#KASH      @ 
#          g1 := AbelianGroupCreate([[1,2,3],[2,4,0], [2,0,0]]);;
#          g2 := AbelianGroupCreate([[1,0,0],[0,4,0], [1,1,1]]);;
#          mat := Mat(Z, [[0,0,0],[6,0,-6],[4,0,4]]);;
#          hom := AbelianGroupHomCreate(g1, g2, mat, true);
#          g := AbelianGroupHomImage(hom);
#          AbelianGroupSmithCreate(g);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupHomImage := function (arg)
  local image, i, l, WithGens, hom;

  if Length(arg) > 2 or Length < 1 then
    Error("use AbelianGroupHomImage(hom [, WithGens])");
  fi;

  hom := arg[1];
  if Length(arg) >=2 then
    WithGens := arg[2];
  else
    WithGens := false;
  fi;
  
  if (IsBool(hom) and not hom) or (not IsBound(hom.Type) or 
      not hom.Type = "AbelianGroupHom") then
    Error("argument is no homomorphism");
  fi;

  l := [];
  for i in [1..MatRows(hom.Mat12)] do
    Add(l, AbelianGroupEltCreate(hom.g2, hom.Mat12[i]));
  od;
  image := AbelianSubGroup(hom.g2, l, WithGens);

  if not IsBound(hom.Information) then
    hom.Information := rec();
  fi;
  hom.Information.Image := image;  

  return image;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupHomKernel
#CALLING   g := AbelianGroupHomKernel(hom [,generators]);
#SHORTDOC  Returns the kernel of a homomorphism.
#LONGDOC   Returns the kernel $g$ of the homomorphism $hom$. $g$ is a group.
#          If the second (optional) parameter $generators$ is $true$ then
#          the generators of the kernel $g$ are generated, too.
#          
#ARGUS     g   @ group
#          hom @ homomorphism
#          generators @ boolean
#
#XREF      AbelianGroupHomImage, AbelianQuotientGroup, 
#          AbelianGroupCanonicalQuotient
#KASH      @ 
#          g1 := AbelianGroupCreate([[1,2,3],[2,4,0], [2,0,0]]);;
#          g2 := AbelianGroupCreate([[1,0,0],[0,4,0], [1,1,1]]);;
#          mat := Mat(Z, [[0,0,0],[6,0,-6],[4,0,4]]);;
#          hom := AbelianGroupHomCreate(g1, g2, mat, true);
#          g := AbelianGroupHomKernel(hom);
#          AbelianGroupSmithCreate(g);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupHomKernel := function (arg)
  local kernel, B, M, group, WithGens, hom;

  if Length(arg) < 1 or Length(arg) > 2 then 
    Error("use AbelianGroupHomKernel(hom [, WithGens])");
  fi;

  hom := arg[1];
  if Length(arg) >=2 then
    WithGens := arg[2];
  else
    WithGens := false;
  fi;

  if (IsBool(hom) and not hom) or (not IsBound(hom.Type) or 
      not hom.Type = "AbelianGroupHom") then
    Error("argument is no homomorphism");
  fi;
  
  B := hom.Mat12;
  group := hom.g2;
  M := _AbelianGroupRelationsMatrixCreate(group, B); # hier wird diesmal keine
                                                     # RelationenMatrix, 
                                                     # sondern Erzeuger 
                                                     # (Zeilen) berechnet

  kernel := AbelianSubGroup(hom.g1, M, WithGens); 

  if not IsBound(hom.Information) then
    hom.Information := rec();
  fi;
  hom.Information.kernel := kernel;

  # order

  return kernel;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupCanonicalQuotient
#CALLING   q := AbelianGroupCanonicalQuotient(hom);
#SHORTDOC  Returns the canonical quotient group of a homomorphism.
#LONGDOC   Let $hom$ be a homomorphism from group $g$ to group $h$.
#          Then {\bf AbelianGroupCanonicalQuotient} returns the quotient group
#          $q$ of $g$ and the kernel of hom. All homomorphisms and isomotphisms
#          are installed. This is useful if you often move between g, h and 
#          the kernel and the image of $hom$.  
#          
#ARGUS     q   @ group
#          hom @ homomorphism
#
#XREF      AbelianGroupImage, AbelianGroupKernel, AbelianQuotientGroup
#KASH      compute the canonical quotient of an homomorphism: @ 
#          g := AbelianGroupCreate([[1,2,3],[2,4,0], [2,0,0]]);;
#          h := AbelianGroupCreate([[1,0,0],[0,4,0], [1,1,1]]);;
#          mat := Mat(Z, [[0,0,0],[6,0,-6],[4,0,4]]);;
#          hom := AbelianGroupHomCreate(g, h, mat, true);
#          q := AbelianGroupCanonicalQuotient(hom);
#          AbelianGroupSmithCreate(q);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupCanonicalQuotient := function(arg)
  local hom, k, q, im, qs, ims, m, h1, h2,
        m1, m2, m3, m4;
  
  if not (Length(arg) = 1 and IsBound(arg[1].Type) and
          arg[1].Type = "AbelianGroupHom") then
    Error("use AbelianGroupCanonicalQuotient( hom )");
  fi;

  hom := arg[1];
  k   := AbelianGroupHomKernel(hom);  
  q   := AbelianQuotientGroup(hom.g1, k);  
  im  := AbelianGroupHomImage(hom);

  qs  := AbelianGroupSmithCreate(q, false);
  ims := AbelianGroupSmithCreate(im, false);

  m   := MatId(Z, MatCols(qs.Relations));
  h1  := AbelianGroupHomCreate(qs, ims, m, m);
  h2  := AbelianGroupHomCreate(ims, qs, m, m);
  
  if not IsBound(qs.Isomorphic) then
    qs.Isomorphic := [];
  fi;
  Add(qs.Isomorphic, h1);
  if not IsBound(ims.Isomorphic) then
    ims.Isomorphic := [];
  fi;
  Add(ims.Isomorphic, h2);

  m1  := q.Information.Smith.Mat12;
  m2  := ims.Information.SmithFrom.Mat12;
  m3   := m1 * m2;

  m1  := im.Information.Smith.Mat12;
  m2  := qs.Information.SmithFrom.Mat12;
  m4   := m1 * m2;

  h1  := AbelianGroupHomCreate(q, im, m3, m4);
  h2  := AbelianGroupHomCreate(im, q, m4, m3);

  if not IsBound(q.Isomorphic) then
    q.Isomorphic := [];
  fi;
  Add(q.Isomorphic, h1);
  if not IsBound(im.Isomorphic) then
    im.Isomorphic := [];
  fi;
  Add(im.Isomorphic, h2);

  return q;
end;

_AbelianGroupExtension := function(sub, quot)
local i, k, l, m, n, t, P, M, g, list, list2, A, B, elt, j, modul,
      a, mat, group2sub, sub2group, group2quot, quot2group;

# Computes a group g with subgroup sub and g/sub = quot.
# This function needs the DiscreteLog of sub. 
# Generators of sub and quot has to be in a common supergroup.

  if not (IsBound(sub.Type) and sub.Type = "AbelianGroup"
          and IsBound(quot.Type) and quot.Type = "AbelianGroup") then
    Error("use AbelianGroupExtension(group, group)\n");
  fi;

  l := MatCols(quot.Relations);
  k := MatCols(sub.Relations);
  m := MatRows(quot.Relations);
  n := MatRows(sub.Relations);

  P := [];
  for i in [1..m] do
    modul := sub.Information.Extension;
    a := 1;
    for j in [1..l] do
#      a := a * (quot.Generators[j]^quot.Relations[i][j]) mod modul; 
      a := a * _IntFastPower(quot.Generators[j], quot.Relations[i][j], modul);
    od;
    elt := AbelianGroupDiscreteLog(sub, a);
    P[i] := elt.expvec[1];
  od;
  P := -P;
  M := [];
  for i in [1..m] do
    M[i] := Flat([quot.Relations[i],P[i]]);
  od;
  for i in [1, n] do
    t := List([1..l], x->0);
    Append(t, sub.Relations[i]);
    M[m+i] := t;
  od;

  IsMat(M);
  g := AbelianGroupCreate(M);

  # homs

  mat := MatId(Z, k);
  mat := MatToRowList(mat);
  t := List([1..l], x->0);
  for i in [1..k] do
    list := Copy(t); 
    Append(list, mat[i]); 
    mat[i] := list;
  od;
  IsMat(mat);
  sub2group := AbelianGroupHomCreate(sub, g, mat);
  group2sub := AbelianGroupHomCreate(g, sub, MatTrans(mat));

  mat := MatId(Z, l);
  t := List([1..l], x->0);
  for i in [1..k] do
    Add(mat, t);
  od;
  IsMat(mat);
  group2quot := AbelianGroupHomCreate(g, quot, mat);
  quot2group := AbelianGroupHomCreate(quot, g, MatTrans(mat));

  if not IsBound(sub.Supergroups) then
    sub.Supergroups := [];
  fi;
  Add(sub.Supergroups, sub2group);

  if not IsBound(quot.Information) then
    quot.Information := rec();
  fi;
  if not IsBound(quot.Information.QuotientsFrom) then
    quot.Information.QuotientsFrom := [];
  fi;
  Add(quot.Information.QuotientsFrom, quot2group);

  if not IsBound(g.Subgroups) then
    g.Subgroups := [];
  fi;
  Add(g.Subgroups, group2sub);
 
  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  if not IsBound(g.Information.Quotients) then
    g.Information.Quotients := [];
  fi;
  Add(g.Information.Quotients, group2quot);

  # generators
  g.Generators := Copy(quot.Generators);
  Append(g.Generators, sub.Generators);
  
  return g;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       RayClassGroupToAbelianGroup
#CALLING   g := RayClassGroupToAbelianGroup(m0 [, minf] [, rels | expo]);
#SHORTDOC  Returns the ray class group modulo a congruence module.
#LONGDOC   Returns the ray class group modulo a congruence module
#           as an abstract group $g$.
#          For more information see {\bf RayClassGroup}.
#ARGUS     g    @ group
#          m0   @ ideal
#          minf @ list of integers @ infinite primes
#          rels @ matrix of integers @ gives additional relations for a quotient
#          expo @ integer @ equivalent to {\tt rels} $=$ {\tt expo*MatId}.
#XREF      RayClassGroup
#KASH      @ 
#          o := OrderMaximal(x^2+6*x+2);
#          m0 :=Elt(o, [0,3])*o;
#          g := RayClassGroupToAbelianGroup(m0, [1,2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayClassGroupToAbelianGroup := function(arg)
  local g, i, o, l;

  if Length(arg) > 3 or Length(arg) < 1 then 
    Error("use RayClassGroupToAbelianGroup(m0 [, minf] [, rels | expo ])");
  fi;

  if Length(arg)=1 then
    arg[2] := [];
  fi;
  if TYPE(arg[2]) = "KANT matrix" then
    arg[3] := arg[2];
    arg[2] := [];
  fi;

  o := arg[1];
  if IsOrder(o) then
    o := 1*o;
  fi;
  l := arg[2];
  if TYPE(arg[1]) = "KANT ideal" then
    g := AbelianGroupCreate(
                MatDiag(Z, RayClassGroup(o, l)[2]));
    if not IsBound(g.Information) then
      g.Information := rec();
    fi;
    g.Generators := List(RayClassGroupCyclicFactors(o, l), i->i[1]);
    g.Basis := g.Generators;

    if not IsBound(g.Information.RCG) then
     g.Information.RCG := rec();
    fi;
    g.Information.RCG.m0 := o;
    g.Information.RCG.minf := l;
    Ideal2EltAssure(o);
    g.Information.Name := SPrint("RayClassGroupToAbelianGroup(", o, ", ", 
                                   l, ")");

    g.DiscreteLog := function(g, i)
      local l, m;

     if IsBound(g.Information.list) then
        l := IdealRayClassRep(i, g.Information.Module, g.Information.list);
      else
        l := IdealRayClassRep(i, g.Information.Module);
      fi;
      m := AbelianGroupEltCreate(g, l);
      return m;
    end;

  else
    g := _gAssureZRCG(o, l);
  fi;

  g.Information.Module := o;
  g.Information.list := l;

  if Length(arg)=3 then
    if IsInt(arg[3]) then
      arg[3] := arg[3]*MatId(Z, MatCols(g.Relations));
    fi;
    g := AbelianQuotientGroup(g, AbelianSubGroup(g, arg[3]));
    g.Information.Name := SPrint("Quotient of RayClassGroupToAbelianGroup(",
                    o, ", ", l, ")");
  fi;
  
  return g;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       RayResidueRingToAbelianGroup
#CALLING   g := RayResidueRingToAbelianGroup(m0 [, minf]);
#SHORTDOC  Returns the ray residue ring modulo a congruence module.
#LONGDOC   Returns the multiplicative group $g$ of the ray residue ring 
#          modulo a congruence module.
#          For more information see {\bf RayResidueRing}.
#ARGUS     g    @ group
#          m0   @ ideal
#          minf @ list of integers @ infinite primes
#XREF      RayResidueRing
#KASH      @ 
#          o := OrderMaximal(Poly(Zx, [1,6,2]));
#          m0 := Ideal(Elt(o, [0,3]));
#          g := RayResidueRingToAbelianGroup(m0, [1,2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayResidueRingToAbelianGroup := function(arg)
  local g, i, o, l;

  if Length(arg) > 2 or Length(arg) < 1 then 
    Error("use RayResidueRingToAbelianGroup(m0 [, minf])");
  fi;

  if Length(arg)=1 then
    arg[2] := [];
  fi;

  o := arg[1];
  l := arg[2];
  if TYPE(arg[1]) = "KANT ideal" then
    g := AbelianGroupCreate(
              MatDiag(Z, RayResidueRing(o, l)[2]));

    if not IsBound(g.Information) then
      g.Information := rec();
    fi;
    g.Information.IsDirectProduct := true;
    g.Generators := List(RayResidueRingCyclicFactors(o, l), i->i[1]);
    g.Basis := g.Generators;

    if not IsBound(g.Information.RRR) then
      g.Information.RRR := rec();
    fi;
    g.Information.RRR.m0 := o;
    g.Information.RRR.minf := l;

    g.DiscreteLog := function(g, i)
      local l, m;
 
      if IsBound(g.Information.list) then
        l := EltRayResidueRingRep(i, g.Information.Module, g.Information.list);
      else
        l := EltRayResidueRingRep(i, g.Information.Module);
      fi;
      m := AbelianGroupEltCreate(g, l);
      return m;
    end;

  else
    g := _gAssureZRRR(o, l);
  fi;

  g.Information.Module := o;
  g.Information.list := l;
  
  return g;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupIsAut
#CALLING   L := AbelianGroupIsAut(hom [, check]);
#SHORTDOC  Tests whether a function is an automorphism.
#LONGDOC   Tests whether the function $hom$ is an automorphism.
#          {\bf AbelianGroupIsAut} returns the list $[true, hom^{-1}]$ 
#          iff $hom$ is automorphism, and the list $[false]$ otherwise.
#          If the second (optional) parameter $check$ is `false` then
#          the check whether $hom$ is endomorphism is omitted.
#
#ARGUS     hom   @ homomorphism
#          check @ boolean
#          L     @ list
#
#XREF      
#KASH      @ 
#          g := AbelianGroupCreate(Mat(Z, [[2,2], [1,3]]));;
#          mat := Mat(Z, [[0,2],[0,2]]);;                 
#          hom := AbelianGroupHomCreate(g, g, mat, true);;
#          AbelianGroupIsAut(hom);
#          @ more information: @
#          AbelianGroupPrintLevel := 1;;
#          AbelianGroupIsAut(hom);
#          AbelianGroupPrintLevel := 0;;
#          @ compute the inverse homomorphism: @
#          mat := Mat(Z, [[-1,2],[3,2]]);;
#          hom := AbelianGroupHomCreate(g, g, mat, true);;
#          hominv := hom^-1;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupIsAut := function(arg)
local g1, g2, s1, s2, hom, hom_g1s1, hom_s1g1, hom_s2g2, hom_g2s2, d, k,
      endo, inv;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use AbelianGroupIsAut(hom [, check])");
  fi;

  hom  := arg[1];
  endo := true;

  if Length(arg) >1 then
    endo := arg[2];
  fi;

  g1 := hom.g1;
  g2 := hom.g2;

  s1 := AbelianGroupSmithCreate(g1); 
  hom_g1s1 := _AbelianGroupHomMove(g1, s1);
  hom_s1g1 := _AbelianGroupHomMove(s1, g1);

  s2 := AbelianGroupSmithCreate(g2); 
  hom_g2s2 := _AbelianGroupHomMove(g2, s2);
  hom_s2g2 := _AbelianGroupHomMove(s2, g2);

  hom := hom_g2s2 * hom * hom_s1g1;

  if not(_AbelianGroupIsHom(hom)) then
    if AbelianGroupPrintLevel >0 then
      Print("function is no homomorphism\n");
    fi;
    return [false];
  elif endo and not(hom.g1 = hom.g2) then
    if AbelianGroupPrintLevel >0 then
      Print("homomorphism is no endomorphism\n");
    fi;
    return [false];
  else
    d := AbelianGroupIndex(hom.g2, AbelianGroupHomImage(hom));
    if not(d = 1) then
      if AbelianGroupPrintLevel >0 then
        Print("homomorphism is not surjective\n");
      fi;
      return [false];
    else
      k := AbelianGroupHomKernel(hom);
      if not(AbelianGroupOrder(k) = 1) then
        if AbelianGroupPrintLevel >0 then
          Print("homomorphism is not injective\n");
        fi;
        return [false];
      fi;
    fi;
  fi;    

  if MatCols(hom.Mat12) = MatRows(hom.Mat12) then
    inv := AbelianGroupHomCreate(s2, s1, MatInv(hom.Mat12));
    inv := hom_s1g1 * inv * hom_g2s2;
  else
    Error("bug in AbelianGroupIsAut");
  fi;

  return [true, inv];
end;

_AbelianGroupHomMove := function(arg)
local mat, hom, g1, g2;

  if Length(arg) <> 2 then 
    Error("use _AbelianGroupHomMove(group, group)");
  fi;
  g1 := arg[1];
  g2 := arg[2];

  mat := AbelianGroupGenerators(g1);
  Apply(mat, x-> AbelianGroupEltMove(x, g2));
  if false in mat then
    if AbelianGroupPrintLevel >0 then
      Print("no homomorphism\n");
    fi;
    return false;
  fi;
  Apply(mat, x->x.expvec);
  IsMat(mat);
  
  hom := AbelianGroupHomCreate(g1, g2, mat);
  return hom;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianFixPointGroup
#CALLING   f := AbelianFixPointGroup(hom); \\
#          f := AbelianFixPointGroup(L);
#SHORTDOC  Computes the fix point group of endomorphisms.
#LONGDOC   Computes the group consisting of elements that are
#          (pointwise) fix under each given endomorphism.
#          
#ARGUS     f   @ group
#          hom @ homomorphism
#          L   @ list @ list of homomorphisms
#
#XREF      
#KASH      compute the fix point group and apply an endomorphism: @ 
#          g := AbelianGroupCreate([[0,0],[0,60]]);;
#          mat := Mat(Z, [[1,1], [0,1]]);;
#          hom := AbelianGroupHomCreate(g, g, mat, true);
#          f := AbelianFixPointGroup(hom);
#          elt := AbelianGroupEltCreate(f, [1,2]);
#          elt := AbelianGroupEltMove(elt, g);
#          hom*elt;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianFixPointGroup := function(arg)
local d, k, g, m, h, fix, s, hom_gs, hom_sg, hom, next_fix;

  if Length(arg) <> 1 then
    Error("use AbelianFixPointGroup(hom) or AbelianFixPointGroup(list of homs)");
  else
    if IsList(arg[1]) then
      if Length(arg[1]) = 1 then
        hom := arg[1][1];
      else
        hom := arg[1][1];
        fix := hom.g1;
        for hom in arg[1] do
          next_fix := AbelianFixPointGroup(hom);
          if next_fix = false then
            return false;
          fi; 
          fix := AbelianGroupIntersect(fix, next_fix);
        od;
        return fix;
      fi;
    else
      hom := arg[1];
    fi;
  fi;

  if hom = false then return false; fi;

  g := hom.g1;

  if not(hom.g1 = hom.g2) then
    if AbelianGroupPrintLevel >0 then
      Print("hom is no endomorphism\n");
    fi;  
    return false;
  fi;

  if not(IsBound(g.Information) and IsBound(g.Information.IsSmith) and g.Information.IsSmith = true) then  
    s := AbelianGroupSmithCreate(hom.g1); 
    hom_gs := _AbelianGroupHomMove(g, s);
    hom_sg := _AbelianGroupHomMove(s, g);

    hom := hom_gs*hom*hom_sg;
    g   := s;
  fi;

  m := hom.Mat12 - MatId(Z, AbelianGroupNumberGenerators(g)); #
  h := AbelianGroupHomCreate(g, g, m);                        # main
  fix := AbelianGroupSmithCreate((AbelianGroupHomKernel(h))); # 

  return fix;
end;

###############################################################################
#                                                                             #
#                       multilinear mappings                                  #
#                                                                             #
###############################################################################

AbelianGroupMultiHomCreate    := Ignore;


#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupTensorProduct
#CALLING   t := AbelianGroupTensorProduct(L);
#SHORTDOC  Returns the tensor product of groups.
#LONGDOC   Returns the tensor product $t$ of a list $L$ of groups.
#          
#ARGUS     t @ group
#          L @ list  @ list of groups
#
#XREF      
#KASH      @ 
#          g1 := AbelianGroupCreate(MatDiag(Z, [1,2,3]));
#          g2 := AbelianGroupCreate(MatDiag(Z, [4,5]));  
#          g3 := AbelianGroupCreate(MatDiag(Z, [4, 5, 0]));
#          @ compute the tensor prosuct $t = g1 \otimes g2 \otimes g3$: @
#          t := AbelianGroupTensorProduct([g1, g2, g3]);            
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupTensorProduct := function(arg)
# returns the 'tensor product' of two or more groups:
#   AbelianGroupTensorProduct(group, group) or
#   AbelianGroupTensorProduct(list_of_groups).
# let g1 & g2 the tensor product of g1, g2.
# let #g the minimal number of generators of g. 
# 0 <= #(g1 & g2) = #g1 * #g2     ???
local g, g1, g2, l, l_new, k, i, j, list, mat, n, m, 
      hom_list, g1s, g2s;
 
  if Length(arg) > 2 or Length(arg) < 1 then
      Error("use AbelianGroupTensorProduct(group, group) or                                 AbelianGroupTensorProduct(list_of_groups) \n");
  fi;

  if Length(arg) = 2 then
    if not (IsBound(arg[1].Type) and IsBound(arg[2].Type) and                              arg[1].Type = "AbelianGroup" and arg[2].Type = "AbelianGroup") then
      Error("use AbelianGroupTensorProduct(group, group) or                                 AbelianGroupTensorProduct(list_of_groups) \n");
    fi;

    g1 := arg[1];
    g2 := arg[2];

    # main begin

    if IsBound(g1.Information) and IsBound(g1.Information.TensorInfo) then
      hom_list := g1.Information.TensorInfo;
      g1s := g1;
    else
      g1s := AbelianGroupSmithCreate(g1);
      hom_list := [g1.Information.Smith];
    fi;

    g2s := AbelianGroupSmithCreate(g2);
    Add(hom_list, g2.Information.Smith);

    n := MatCols(g1s.Relations);
    m := MatCols(g2s.Relations);

    k := 1;
    list := [];
    for i in [1..m] do
      for j in [1..n] do
        list[k] := Gcd(g1s.Relations[j][j], g2s.Relations[i][i]);
        k := k+1;
      od;
    od;
    mat := MatDiag(Z, list);

    g := AbelianGroupCreate(mat);
    if not IsBound(g.Information) then 
      g.Information := rec();
    fi;
    g.Information.TensorInfo := hom_list;
    
    return g;

    # main end
    
  elif IsList(arg[1]) then
    if Length(arg[1]) = 1 then
      return arg[1][1];
    fi;
    l  := arg[1];
    g1 := l[1];
    g2 := l[2];
    g  := AbelianGroupTensorProduct(g1, g2);
    l_new := l{[2..Length(l)]};
    l_new[1] := g;
    Unbind(g1);     # Zwischenprodukte loeschen
    if Length(l_new) = 1 then 
      return l_new[1];
    else
      return AbelianGroupTensorProduct(l_new);
    fi;
  else 
    Error("use AbelianGroupTensorProduct(group, group) or                                 AbelianGroupTensorProduct(list_of_groups) \n");
  fi;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianMultiHomGroup
#CALLING   L := AbelianMultiHomGroup([$g_{1},\ldots,g_{n}$], h);
#SHORTDOC  Computes the group of multilinear mappings.
#LONGDOC   Computes the group of multilinear mappings from $d = 
#          g_{1}\times\cdots\times g_{n}$ to $h$. 
#          {\bf AbelianMultiHomGroup} returns a list with three enties:
#          \begin{enumerate}
#          \item group of multilinear mappings from 
#          $g_{1}\times\cdots\times g_{n}$ to $h$. 
#          \item direct product $d = g_{1}\times\cdots\times g_{n}$
#          \item tensor product 
#          $t = g_{1}\otimes\cdots\otimes g_{n}$.
#          \end{enumerate}
#          You need $d$ and $t$ only if you want to create multilinear
#          mappings from $g_{1}\times\cdots\times g_{n}$ to $h$ corresponding
#          to a given matrix using the function 
#          {\bf AbelianGroupMultiHomCreate}.
#          
#ARGUS     $ g_{1},\ldots, g_{n}, h$ @ group
#          L @ list
#
#XREF      AbelianGroupMultiHomCreate
#KASH      @ 
#          z  := AbelianGroupCreate(Mat(Z,[[0]]));;
#          z2 := AbelianGroupCreate(Mat(Z,[[2]]));;
#          z4 := AbelianGroupCreate(Mat(Z,[[4]]));;
#          z8 := AbelianGroupCreate(Mat(Z,[[8]]));;
#          l  := AbelianMultiHomGroup([z,z4,z8], z2*z8);
#          @ mg is the group of multilinear mappings from 
#          $d = z \times z4 \times z8$ to $h = z2 \times z8$: @
#          mg := l[1]; 
#          @ d is the direct product $z \times z4 \times z8$: @
#          d  := l[2]; 
#          @ t is the tensor product $z \otimes z4 \otimes z8$: @
#          t  := l[3]; 
#          @ elt is element of group of multilinear mappings from 
#          $d = z \times z4 \times z8$ to $h = z2 \times z8$: @
#          elt  := AbelianGroupEltCreate(mg, [1,0]); 
#          @ f is a multilinear mapping from $d = z\times z4 \times z8$ 
#          to $h = z2 \times z8$: @
#          f := AbelianGroupDiscreteExp(elt); 
#          a := AbelianGroupEltCreate(d, [1,1,1]);
#          f * a;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianMultiHomGroup := function(arg)
local g1, g2, t, i, hom_group, group_list, temp;
# let g1 x g2 the tensor product of g1, g2.
# AbelianGroupMultiHomGroup([g11, g12, g13], g2) computes the group of
# trilinear mappings  g11 * g12 * g13 -> g2 
# if direct product g1 := g11 * g12 * g13 and
# tensor product    t := g11 x g12 x g13 alrady known use the optional 
# 3rd and 4th parameter: AbelianGroupMultiHomGroup([g11, g12, g13], g2, g1, t)
# result: [multi_hom_group, direct product g1, tensor product t]

  if Length(arg) <> 2 then
    Error("use AbelianMultiHomGroup(list_of_groups, group)");
  fi;

  group_list := arg[1];
  g2 := arg[2];

  if IsList(group_list) and  Length(group_list) <= 1 then
    Error("use AbelianMultiHomGroup(list_of_groups, g2) with length of list_of_groups >1 \n");
  fi;

  if Length(arg) >= 3 then
    g1 := arg[3];
    t := arg[4];
  else
    g1  := group_list[1];
    for i in group_list{[2..Length(group_list)]} do
      temp := g1 * i;
      Unbind(g1); # Zwischenprodukte loeschen
      g1 := temp;
    od;
    t := AbelianGroupTensorProduct(group_list);
  fi;
 
  hom_group := AbelianHomGroup(t, g2);

  # MultiHom(g11 * g12 * g13, g2) =~ Hom(g11 x g12 x g13, g2)

  for i in [1..MatCols(hom_group.Relations)] do
    hom_group.Generators[i] := AbelianGroupMultiHomCreate(g1, g2, t, hom_group.Generators[i].Mat12);
  od;
  
  return [hom_group, g1, t]; 
end;

_AbelianGroupEltDirectToTensor := function (arg)
# moves an element from a direct product to the corresponding tensor
# product. e.g. you may use the 2nd and 3rd argument returned by
# AbelianMultiHomGroup
local i, g1_elt, t, list1, list2, l, k, vec, vec_old, vec_new, 
      temp, len1, len2, add_list1, add_list2, elt, L, hom_list,
      elt_g1, elt_g2;

  g1_elt := arg[1];
  t := arg[2];

  hom_list := t.Information.TensorInfo;
  vec_old := g1_elt.expvec[1];
  vec_new := [];

  list1 := List(hom_list, x->MatCols(x.g1.Relations)); 
  len1 := Length(list1);
  add_list1 := [1];
  for i in [1..len1] do
    add_list1[i+1] := add_list1[i] + list1[i];
  od; 

  list2 := List(hom_list, x->MatCols(x.g2.Relations)); 
  len2 := Length(list2);
  add_list2 := [1];
  for i in [1..len2] do
    add_list2[i+1] := add_list2[i] + list2[i];
  od; 

  for i in [1..Length(hom_list)] do
    elt_g1 := AbelianGroupEltCreate(hom_list[i].g1, 
              vec_old{[add_list1[i]..add_list1[i+1]-1]});
    elt_g2 := hom_list[i] * elt_g1;
    vec_new{[add_list2[i]..add_list2[i+1]-1]} := elt_g2.expvec[1]; 
  od;  

  l := List(list2, x -> x-1);

  L := MultiCounterInit(len1);
  k := 1;
  vec := [];
  while MultiCounterInc (L, l) do
    temp := 1;
    for i in [1..len2] do
      temp := temp * vec_new[add_list2[i]+L[i]];
    od;
    vec[k] := temp;
    k := k+1;
  od;

  elt := AbelianGroupEltCreate(t, vec);
  return elt;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianGroupMultiHomCreate
#CALLING   f := AbelianGroupMultiHomCreate(d, g, t, mat);
#SHORTDOC  Creates a multilinear mapping.
#LONGDOC   Creates a multilinear mapping from direct product $d$ to group $g$.
#          This multilinear mapping corresponds to the homomorphism from
#          the tensor product $t$ to group $g$ with the matrix $mat$.
#          It is recommended to use the second and third entry returned by
#          {\bf AbelianMultiHomGroup} (see example).
#          It is checked if $mat$ corresponds to a homomorphism.
#          For a second possibility to create multilinear mappings see
#          {\bf AbelianMultiHomGroup}.
#          
#ARGUS     d, g, t  @ group
#          mat      @ matrix or list of lists
#
#XREF      AbelianMultiHomGroup      
#KASH      @ 
#          g1 := AbelianGroupCreate(Mat(Z,[[2]]));;
#          g2 := AbelianGroupCreate(Mat(Z,[[4]]));;
#          h := AbelianGroupCreate(Mat(Z,[[8,0],[0,2]]));;
#          l  := AbelianMultiHomGroup([g1,g2], h);
#          @ direct product $d = g1\times g2$: @ 
#          d  := l[2]; 
#          @ tensor product $t = g1\otimes g2$: @ 
#          t  := l[3]; 
#          @ create the multilinear mapping: @
#          f := AbelianGroupMultiHomCreate(d, h, t, [[0,1]]);
#          elt := AbelianGroupEltCreate(d, [1,1]); 
#          AbelianGroupEltReduce(f*elt, true);
#          f := AbelianGroupMultiHomCreate(d, h, t, [[1,0]]);
#          AbelianGroupPrintLevel := 1;
#          f := AbelianGroupMultiHomCreate(d, h, t, [[1,0]]);
#          AbelianGroupPrintLevel := 0;
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          HB 18.09.98 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianGroupMultiHomCreate := function(arg)
# Creates a multilinear mapping corresponding to hom from tensor product
# to g2 with matrix mat.
# e.g. you may use the 2nd and 3rd argument returned by
# AbelianMultiHomGroup
local hom_t_to_g2, hom_g1_to_g2, g1, g2, t, mat;

  if Length(arg) <> 4 then 
    Error("use AbelianGroupMultiHomCreate(group, group, group, mat)\n");
  fi;
  g1  := arg[1];
  g2  := arg[2];
  t   := arg[3];
  mat := arg[4];

  if not (IsRec(g1) and IsBound(g1.Type) and g1.Type = "AbelianGroup"
          and IsRec(g2) and IsBound(g2.Type) and g2.Type = "AbelianGroup"
          and IsBound(t.Type) and t.Type = "AbelianGroup"
          and (IsMat(mat) or IsList(mat))) then
    Error("use AbelianGroupMultiHomCreate(group, group, group, mat)\n");
  fi;
  
  hom_t_to_g2  := AbelianGroupHomCreate(t, g2, mat, true);
  if hom_t_to_g2 = false then 
    return false;
  fi;

  hom_g1_to_g2 := rec( Type         := "AbelianGroupHom",
                         g1         := g1,
                         g2         := g2,
                         operations := _AbelianGroupHomOperations
                     );

  hom_g1_to_g2.Type := "AbelianGroupMultiHom";
  hom_g1_to_g2.Tensor := t;
  hom_g1_to_g2.HomTensor := hom_t_to_g2;

  return hom_g1_to_g2;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianDualHom
#CALLING   dualhom := AbelianDualHom(hom); \\
#          
#SHORTDOC  Creates the dual homomorphism of a given homomorphism.
#LONGDOC   Creates the dual homomorphism from the group $g2^{*}$  
#          to group $g1^{*}$ of a homomorphism hom: $g1 \rightarrow g2$.
#          The dual homomorphism is defined by: $\text{hom}^{*}:g2^{*} \rightarrow g^{*}$
#          and $\text{hom}^{*}: \chi \rightarrow \chi * hom$.
#           
#          
#          
#          
#          
#ARGUS     hom    @ homomorphism @ homomorphism from $g1$ to $g2$
#          
#          
#          
#
#XREF      AbelianGroupHomCreate, AbelianHomGroup
#KASH      @ 
#          g1 := AbelianGroupCreate([[0,1,2],[5,6,0],[0,4,5]]);;
#          g2 := AbelianGroupCreate([[0,2],[3,0]]);;
#          mat := Mat(Z, [[-24,24],[20,-20],[-10,10]]);;  
#          hom := AbelianGroupHomCreate(g1, g2, mat, true);
#          dualhom:=AbelianDualHom(hom);
#          
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          hb 16.03.99 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianDualHom:=function(arg)
local hom, n, g1, g1_, g1Gen, g1_Gen, L, 
      g1GenHom, Hom, g2, g2_, g2_Gen, g2_GenHom;

if (Length(arg)=1 or Length(arg)=3) then 
  
   hom:=arg[1];
     

 
if IsRec(hom) then
 if IsBound(hom.Type) then
   if hom.Type="AbelianGroupHom" then 
  

  g1:=hom.g1;
  g2:=hom.g2;

 if Length(arg)=1 then 
 
  g1_:=AbelianDualGroup(g1);
  g2_:=AbelianDualGroup(g2);

else 
     
    g2_:=arg[2];
    g1_:=arg[3];
fi;

  g1_Gen:=AbelianGroupGenerators(g1_);
  g2_Gen:=AbelianGroupGenerators(g2_);

  #g2_GenHom:=List(g2_Gen,z->AbelianGroupDiscreteExp(z));
  
  g1Gen:=AbelianGroupGenerators(g1);

 # Hom:=AbelianGroupDiscreteExp(hom);
  g1GenHom:=List(g1Gen,z->hom*z);

  g2_Gen:=List(g2_Gen,z->List(g1GenHom,y->z*y)); 
  
  L:=List(g1_Gen,z->List(g1Gen,y->z*y)); 
  n:=Lcm(List(Flat(L),z->Den(z)));
  n:=Lcm(n,Lcm(List(Flat(g2_Gen),z->Den(z))));
  L:=MatTrans(Z,n*L);
  g2_Gen:=n*g2_Gen;
  g2_Gen:=List(g2_Gen,z->MatTrans(Z,[z]));
  g2_Gen:=List(g2_Gen,z->MatSolve(L,z,n)[1]);
  g2_Gen:=List(g2_Gen,z->MatToColList(z)[1]);
  g2_Gen:=Mat(Z,g2_Gen);

 return AbelianGroupHomCreate(g2_,g1_,g2_Gen,true);

else return Error("usage: <AbelianGroupHom>");
 fi;
 else return Error("usage: <AbelianGroupHom>");
  fi;
  else return Error("usage: <AbelianGroupHom>");
   fi;


else return Error("usage: <AbelianGroupHomomorphism, dualgroup, dualgroup>");

fi;

end;  	
  
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       AbelianRayGroupImbed
#CALLING   dualhom := AbelianRayGroupImbed(G,g); \\
#          
#SHORTDOC  Imbeds a ClassGroup into a RayClassGroup.
#LONGDOC   If $I^{({ \goth f})}/P_{ \goth f}$ is a ray class group with  
#          conductor ${ \goth f}$ and $I^{ \goth f'}/U_{ \goth f'}$ is 
#          a classgroup with conductor ${\goth f'}$, then we may
#          imbed $U_{\goth f'}$ canonically into 
#          $I^{({\goth f})}/P_{\goth f}$: $U_{\goth f'}^{({\goth f})}/P_{\goth f}$ 
#          is a subgroup of $I^{({\goth f})}/P_{\goth f}$.
#          
#          
#          
#ARGUS     G    @ abstact rayclassgroup
#          g    @ abstact classgroup
#          
#          
#
#XREF      RayClassGroupToAbelianGroup
#KASH      @
#          O:=OrderMaximal(x^3-x^2-9*x+8);; 
#          G :=RayClassGroupToAbelianGroup(25*O);;
#          g:=RayClassGroupToAbelianGroup(1*O);;
#          u:=AbelianRayGroupImbed(G,g);  
#          
#          
#          
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#          hb 16.03.99 written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

AbelianRayGroupImbed:=function(G,g)
local A, BG, M, hom, ff;
  
  if IsRecType(G, "AbelianGroup") and IsBound(G.Information.RCG) and 
     IsRecType(g, "AbelianGroup") and IsBound(g.Information.RCG) then
    ff:=RayConductor(g);
    if IdealIsIntegral(G.Information.RCG.m0/g.Information.RCG.m0) and
       ForAll(ff[2],z->z in G.Information.RCG.minf) then        
  
      BG:=AbelianGroupGenerators(G);
      BG:=List(BG,z->AbelianGroupDiscreteExp(z));
      BG:=List(BG,z->AbelianGroupDiscreteLog(g,z));
      BG:=List(BG,z->MatToRowList(z.expvec)[1]);
      M:=Mat(Z,BG);

      hom:=AbelianGroupHomCreate(G,g,M);

      return AbelianGroupHomKernel(hom);
      return AbelianQuotientGroup(G, AbelianGroupHomKernel(hom));
  
    else 
      Error("ray of the second group is not a divisor of the first");
    fi;

  else
    Error("usage <abstact rayclassgroup, abstact classgroup>");
  fi;
 
end;
###############################################################################



