
GetIdealEps := function(ord, prim)
# Determines a possible ideal idealEps prime to prim for the calculation
# of the eta-quotient eps in schertz's P-function
local x, discr, idealEps;

    discr := OrderDisc (ord);
    if discr=(-3) or discr=(-4) then
      idealEps := 1*ord;
    else
      x := 11;
      while ((x+1) mod 12)<>0 or IntGcd(x,prim)<>1
            or ImQuadLegendre(discr,x)<>1 do
        x := NextPrime(x);
      od;
      idealEps := Factor(x*ord)[1][1];
    fi;

    return idealEps;
end;



GetIdealD := function(ord)
# Determines a possible ideal idealD for the calculation
# of the number D for schertz's generating element theta = P-D
local x, discr, idealD;

    discr := OrderDisc (ord);
    if discr=(-3) then
      idealD := Factor(2*ord)[1][1] * Factor(3*ord)[1][1];
    elif discr=(-4) then
      idealD := Factor(2*ord)[1][1] * Factor(5*ord)[1][1];
    elif ImQuadLegendre(discr,2)=(-1) then
      x := 5;
      while ((x+1) mod 3)<>0 or ImQuadLegendre(discr,x)<>1 do
        x := NextPrime(x);
      od;
      idealD := Factor(x*ord)[1][1];
    elif ImQuadLegendre(discr,3)<>(-1) then
      idealD := Factor(2*ord)[1][1] * Factor(3*ord)[1][1];
    elif ImQuadLegendre(discr,2)=1 then
      idealD := 2*ord;
    else
      x := 7;
      while ((x+1) mod 4)<>0 or ImQuadLegendre(discr,x)<>1 do
        x := NextPrime(x);
      od;
      idealD := Factor(x*ord)[1][1];
    fi;

    return idealD;
end;



GetDConjugates := function(DGenConj, idealD)
# Calculates the conjugates of D (=spur(P(...))) using all the conjugates
# DGenConj of the Element P(...)
local j, k, discr, DSumConj;

    DSumConj := [];
    discr := OrderDisc (IdealOrder(idealD));
    for j in [1..Length(DGenConj)] do
      if discr=(-3) then
        DSumConj[j] := DGenConj[j][1];
      elif discr=(-4) then
        DSumConj[j] := DGenConj[j][1];
      elif ImQuadLegendre(discr,2)=(-1) then
        DSumConj[j] := 0;
        if ImQuadLegendre(discr,3)<>(-1) then
          for k in [1..Length(DGenConj[1])] do
            DSumConj[j] := DSumConj[j] - DGenConj[j][k];
          od;
        fi;
      elif ImQuadLegendre(discr,3)<>(-1) then
        DSumConj[j] := DGenConj[j][1];
      elif ImQuadLegendre(discr,2)=1 then
        DSumConj[j] := DGenConj[j][1];
      else
        DSumConj[j] := 0;
        for k in [1..Length(DGenConj[1])] do
          DSumConj[j] := DSumConj[j] - DGenConj[j][k];
        od;
        DSumConj[j] := DSumConj[j] * ImQuadLegendre(2,Abs(IdealNorm(idealD)));
      fi;
    od;

    return DSumConj;
end;



alphaConjugates := function(classReprPrim, f)
# Calculates the conjugates of the integral factor alpha (mainly a quotient
# of eta-values) used for the rayclassfield-generator theta=alpha*(P-D)
local j, fprim, prim, p, pid, xi, ord, discr,
      hilfid, hilfidBasis, reprA, alpha;

    if Length(Factor(f)) = 1 then
      ord := IdealOrder(f);
      discr := OrderDisc (ord);
      fprim := Factor(f)[1][1];
      prim := 6 * discr * IdealNorm (Product(classReprPrim));
      prim := FactorProd (prim);
      p := 5;
      while (prim mod p)=0 or ImQuadLegendre(discr,p)<>1 or
            (IdealIsPrincipal(fprim*Factor(p*ord)[1][1])=false and
             IdealIsPrincipal(fprim*Factor(p*ord)[2][1])=false) do
        p := NextPrime(p);
      od;
      pid := Factor(p*ord)[1][1];
      xi := IdealIsPrincipal(fprim*pid/IdealNorm(pid));
      if xi=false then
        pid := IdealNorm(pid)/pid;
        xi := IdealIsPrincipal(fprim*pid/IdealNorm(pid));
      fi;
      alpha := [];
      for j in [1..Length(classReprPrim)] do
        hilfid := pid^2*IdealNorm(classReprPrim[j])/classReprPrim[j];
        hilfidBasis := EltCon(Elt(ord,MatToColList(IdealBasisUpperHNF(hilfid)[2])[2]),1);
        if Im(hilfidBasis)<0 then hilfidBasis := -hilfidBasis; fi;
        reprA := Abs(MatElt(IdealBasisUpperHNF(classReprPrim[j])[2],1,1));
        alpha[j] := EltCon(xi,1)*DedekindEta(hilfidBasis/(reprA*p^2))/
                                 DedekindEta(hilfidBasis/reprA);
      od;
    else
      alpha := List([1..Length(classReprPrim)],x->1);
    fi;

    return alpha;
end;



EpsForSchertzPConjugates := function(classReprPrim, idealEps)
# Calculates the conjugates of the factor
#   eps*(reprA/(2*pi))^2/eta(hilfidBasis/reprA)^4
# (mainly a quotient of eta-values) in schertz's P-function belonging to the
# ideal idealEps and the given representants classReprPrim of the ideal classes
local j, reprA, epsA, hilfid, hilfidBasis,
      hilf1, hilf2, hilf3, epsQuotient, ord;

  epsQuotient := [];
  ord := IdealOrder(idealEps);
  epsA := Abs(MatElt(IdealBasisUpperHNF(idealEps)[2],1,1));
  for j in [1..Length(classReprPrim)] do
    reprA := Abs(MatElt(IdealBasisUpperHNF(classReprPrim[j])[2],1,1));
    if (OrderDisc(ord)<>(-3)) and (OrderDisc(ord)<>(-4)) then
      hilfid := idealEps^2*IdealNorm(classReprPrim[j])/classReprPrim[j];
      hilfidBasis := EltCon(Elt(ord,MatToColList(IdealBasisUpperHNF(hilfid)[2])[2]),1);
      if Im(hilfidBasis)<0 then hilfidBasis := -hilfidBasis; fi;
      hilf3 := DedekindEta(hilfidBasis/reprA)^3;
      hilf2 := DedekindEta(hilfidBasis/(reprA*epsA))^2;
      hilf1 := DedekindEta(hilfidBasis/(reprA*epsA^2));
      epsQuotient[j] := (reprA/(2*pi))^2*hilf1/(hilf2*hilf3);
    else
      hilfid := IdealNorm(classReprPrim[j])/classReprPrim[j];
      hilfidBasis := EltCon(Elt(ord,MatToColList(IdealBasisUpperHNF(hilfid)[2])[2]),1);
      if Im(hilfidBasis)<0 then hilfidBasis := -hilfidBasis; fi;
# if OrderDisc(ord)=-4 the value hilfidBasis have to be normalized additionally;
# ATTENTION: for simplicity it is supposed at this point, that classReprPrim=[1*ord] !
#            if classReprPrim was calculated using 'IdealClassGroupRepr()' or
#            'IdealClassGroupReprPrim()', this will always be the case
      if OrderDisc(ord)=-4 then
        hilfidBasis := Comp(0,Im(hilfidBasis));
      fi;
      hilf1 := DedekindEta(hilfidBasis/reprA)^4;
      epsQuotient[j] := (reprA/(2*pi))^2/hilf1;
    fi;
  od;

  return epsQuotient;
end;




SchertzPConjugates := function(value, classReprPrim, resRepr, idealEps)
# Calculates all the conjugates P(value|(classReprPrim[j]*resRepr[k])^(-1))
# of schertz's P-value P(value|ord) given by the representants classReprPrim
# of the ideal classes and the representants resRepr of the residue classes
local j, k, eps, reprA, reprBasis, hilf1, hilf2, hilf3, temp, zzahl,
      ord, schertzP;

# debugging:
# Print("\ndebugging ...\n");
  schertzP := [];
  ord := IdealOrder(idealEps);
  eps := EpsForSchertzPConjugates (classReprPrim, idealEps);
# debugging:
# Print("eps = ",eps,"\n");
# Print("IdealOrder(eps) = IdealOrder(repr) ?   ",
#       ord=IdealOrder(classReprPrim[1]),"\n");
  for j in [1..Length(classReprPrim)] do
    temp := [];
    zzahl := EltCon(value,1)*IdealNorm(classReprPrim[j]);
    reprBasis := EltCon(Elt(ord,MatToColList(IdealBasisUpperHNF(classReprPrim[j])[2])[2]),1);
# debugging:
# Print("reprBasis = ",reprBasis,"\n");
# Print("World(reprBasis): ",World(reprBasis),"\n");
    if Im(reprBasis)<0 then reprBasis := -reprBasis; fi;

# --> hier stuerzt das Programm manchmal ab !!!
    hilf2 := -Conj(reprBasis);
# --> hier stuerzt das Programm manchmal ab !!!

    hilf3 := Comp(Abs(MatElt(IdealBasisUpperHNF(classReprPrim[j])[2],1,1)),0);
# debugging:
# Print("classReprPrim[",j,"] = ",classReprPrim[j],"\n");
#    hilf3 := IdealBasisUpperHNF(classReprPrim[j]);
# Print("nach HNF: ",hilf3,"\n");
#    hilf3 := MatElt(hilf3[2],1,1);
# Print("MatElt(hilf3[2],1,1): ",hilf3,"\n");
#    hilf3 := Comp(Abs(hilf3),0);
# debugging:
# Print("\nRestklasse=",classReprPrim[j],",   Gitter = ( ",hilf2," , ",hilf3," )\n");
    for k in [1..Length(resRepr)] do
      hilf1 := zzahl*EltCon(resRepr[k],1);
# debugging:
# Print("z-Wert = ",hilf1,"\n");
      if (OrderDisc(ord)<>(-4)) and (OrderDisc(ord)<>(-3)) then
        temp[k] := eps[j]*WeierstrassP(hilf1,hilf2,hilf3);
      elif OrderDisc(ord)=(-4) then
        temp[k] := (eps[j]*WeierstrassP(hilf1,hilf2,hilf3))^2;
      else
        temp[k] := (eps[j]*WeierstrassP(hilf1,hilf2,hilf3))^3;
      fi;
    od;
    schertzP[j] := temp;
  od;

  return schertzP;
end;




ImQuadRayField := function(arg)
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#USERDEF
#FXN       ImQuadRayField
#SHORTDOC  Determines the ray class field modulo ideal f over an 
#          imaginary quadratic field.
#LONGDOC   Let $K$ be a imaginary quadratic field with maximal
#          order $\O|K$, $f$ an integral $\O|K$--ideal
#          and $K|{(f)}$ the ray class field over $K$ corresponding to $f$.
#          Integral
#          generators for $K|{(f)}$ over the Hilbert class field $K|{(1)}$ are
#          for example elements of the form
#          \[ \Theta = \alpha\,\Big( P(1\mid f)-D \Big), \]
#          where $P(1\mid f)$ is a normalized singular value
#          of the Weierstrass $\wp$--function
#          and $\alpha$ and $D$ are suitable choosen elements
#          of $K|{(1)}$ (see \cite{Scherz,Scherz2}).
#          Furthermore we have an integral basis for $K|{(f)}/K|{(1)}$ 
#          given by
#          \begin{eqnarray*} \O|{K|{(f)}} = \O|{K|{(1)}} +
#          \Theta\:\O|{K|{(1)}} &+&
#          \alpha^{-1}\,\Theta^2\:\O|{K|{(1)}} + \\ &+& \ldots +
#          \alpha^{-n+2}\,\Theta^{n-1}\:\O|{K|{(1)}} \end{eqnarray*}
#          The minimal polynomial for $\Theta$ is calculated using complex
#          approximations one gets by evaluating the defining infinite
#          series of the Dedekind $\eta$--function and the
#          Weierstrass $\wp$--function.
#
#          If {\tt "char"} is specified only the characteristic polynomial
#          for $\Theta$ over $K$ is computed (which might be reducible).
#          In case of {\tt "field"} also the Hilbert class field (and the
#          representation of the ray class field over 
#          the Hilbert class field) is
#          determined, and if {\tt "maxord"} or nothing is specified
#          (but that might cause a long output)
#          one gets the maximal order of the
#          ray class field as an extension of the maximal order of the Hilbert
#          class field.
#
#          {\em Note:} For numerical reasons don't use prime ideal powers
#          as a module whenever possible.
#          So for example if 2 is not inert in $K$, $p|2$ is a prime ideal
#          in $K$ above 2 and $f$ is prime to $p|2$, then we
#          have $K|{(f)}=K|{(p|2f)}$ and one should 
#          use {\tt ImQuadRayField}($p|2f$)
#          instead of {\tt ImQuadRayField}($f$).
#
#CALLING   O := ImQuadRayField( f [,"char" $\mid$ "field" $\mid$ "maxord"] );
#ARGUS     f @ ideal @ integral ideal in imaginary quadratic field
#          O @ order @ ray class field modulo {\tt f}
#DEBUG
#XREF      RayClassField, ImQuadHilbert, OrderHilbert.
#PACKAGE   lib/ImQuad
#KASH      @
#          o := OrderMaximal(Z, 2, -7);
#          f := Factor(2*o)[1][1] * Factor(7*o)[1][1];
#          O := ImQuadRayField(f, "maxord");
#          @
#          Here Hilbert class field and imaginary field are the same (E1).
#          F[1] is the maximal order of the ray class field, compare also the
#          corresponding discriminants:@
#          Factor(OrderDisc(OrderCoefOrder(O)));
#          Factor(OrderDisc(O));
#
#**HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***C***C***C***C***C***C***C***C***C***C***C***C***C***C***C***C***C***/
# Determines the characteristic polynomial over the imaginary quadratic field
# (relating to ideal f) for a generating integral element alpha*(P-D) of the
# rayclassfield modulo f (the polynomial may be reducible), the rayclassfield
# itself and the maximal order of the rayclassfield.
# Numerical approximations for a generating element alpha*(P-D) and its
# conjugates are calculated using schertz's P-function.
# The theoretical background is described in three articles of Schertz,
# published in
#   J. Reine Angew. Math. 398 (1989), 105-129
#   J. Number Theory 34 (1) (1990), 41-53
#   J. Number Theory 34 (2) (1990), 229-234
#
local j, k, l, temp, temp2, prec, oprec, precOld, debug, flag, ord, f,
      idealD, idealEps, hilbertRepr, hilbertReprPrim, prime,
      fResidues, idealDResidues, fValue, DValue,
      fGenerator, DGenerator, DHilbert, alpha,
      theta, thetaPol, oK1All, thetaK1Pol, oKfEqnOrder,
      oK1Abs, alphaElt, Tden, Tmat, oKf;

    f := Copy(arg[1]);
    debug := false;
    flag := false;
    if Length(arg)>1 and arg[2]<>"debug" then
      flag := arg[2];
    fi;
    if arg[Length(arg)]="debug" then
      debug := true;
    fi;
    ord := IdealOrder(f);
    if OrderDegAbs(ord)<>2 or OrderDisc(ord)>0 then
      Error("ImQuadRayField: Field must be imaginary quadratic!\n");
    fi;
    if not OrderIsMaximal(ord) then
      Error("ImQuadRayField: Only maximal orders are supported!\n");
    fi;

    Print("resulting degrees:\n");
    temp := [ImQuadResNum(f),ImQuadClassNum(ord)];
    Print("  [RayField:Hilbert]=",temp[1],
          "  ,  [Hilbert:Imaginary]=",temp[2],"\n");

# usage of new precision
    precOld := [Prec(),OrderPrec(ord)];
    prec := Maximum(Prec(),20+temp[1]*temp[2]);
    prec := Prec(2*Round(prec/2));
    oprec := Maximum(prec,2*Round(OrderPrec(ord)[1]/2));
    OrderPrec(ord, oprec);
#    ord := OrderMaximal(Order(OrderPoly(Zx,ord)));
#    if not OrderIsSubfield(IdealOrder(f),ord) then
#      Prec(precOld[1]); OrderPrec(precOld[2]);
#      return false;
#    fi;
#    f := IdealMove(f,ord);
    if debug then
      Print("ord = \n",ord,"\n");
      Print("EltCon(OrderBasis(ord)[2],1) = ",EltCon(OrderBasis(ord)[2],1),"\n");
    fi;

# ideals for D and Eps
    idealD := GetIdealD (ord);
    idealEps := GetIdealEps (ord, 6*IdealNorm(f*idealD));
    if debug then
      Print("f = ",f,"\n");
      Print("idealD = ",idealD,"\n");
      Print("idealEps = ",idealEps,"\n");
    fi;

# representants for ideal and residue classes
    hilbertRepr := IdealClassGroupRepr (ord);
    prime := FactorProd (IdealNorm (f*idealD*idealEps));
    hilbertReprPrim := IdealClassGroupReprPrim (ord, prime, hilbertRepr);
    fResidues := ImQuadRayResidues (f);
    idealDResidues := ImQuadRayResidues (idealD);
    if debug then
      Print("prime = ",prime,"\n");
      Print("hilbertReprPrim = ",hilbertReprPrim,"\n");
      Print("fResidues = ",fResidues,"\n");
      Print("idealDResidues = ",idealDResidues,"\n");
    fi;

# calculation of the hilbert class field
    if flag<>"char" then
      Print("\ncalculation of Hilbert class field ...\n");
# debugging:
#      oK1All := ImQuadHilbert (ord, hilbertReprPrim, "roots", "debug");
      oK1All := ImQuadHilbert (ord, hilbertReprPrim, "roots");
      if oK1All=false then
        Prec(precOld[1]); OrderPrec(ord, precOld[2][1], precOld[2][2]);
        return false;
      fi;
      if debug then
        Print("oK1 =\n",oK1All[1],"\n");
        Print("oK1generator = ",oK1All[2],"\n");
      fi;
      if flag=false or debug then
        Print("Hilbert class field:\n",oK1All[1],"\n");
      fi;
    fi;

    if temp[1]=1 then return oK1All[1]; fi;  # CF was here

# values for calculation of P(fValue) and D=Spur(P(DValue))
    temp := IdealClassReprPrim (f, prime);
    fValue := temp[2];
    DValue := IdealClassReprSmall (f/idealD, "mult");
    temp := IdealClassReprPrim (DValue[1]^(-1), idealD);
    DValue := fValue*temp[2]/DValue[2];
    if debug then
      Print("fValue = ",fValue,"\n");
      Print("DValue = ",DValue,"\n");
    fi;

# conjugates for P(fValue), P(DValue), alpha and D=Spur(P(DValue))
    Print("\ncalculation of ray class field ...\n");
    l := 1;
    prec := prec/2; oprec := oprec/2;
    repeat
      l := l+1;
      prec := Prec(2*prec); oprec := OrderPrec(ord,2*oprec)[1];
      Print("used real precision is ",prec," places\n");
      fGenerator := SchertzPConjugates (fValue, hilbertReprPrim, fResidues, idealEps);
      DGenerator := SchertzPConjugates (DValue, hilbertReprPrim, idealDResidues, idealEps);
      alpha := alphaConjugates (hilbertReprPrim, f);
      DHilbert := GetDConjugates (DGenerator, idealD);
      if debug then
        Print("fGenerator = ",fGenerator,"\n");
        Print("DGenerator = ",DGenerator,"\n");
        Print("alpha = ",alpha,"\n");
        Print("DHilbert = ",DHilbert,"\n");
      fi;

# conjugates of the integral generator theta=alpha*(P(fValue)-D)
      theta := [];
      for j in [1..Length(fGenerator)] do
        temp := [];
        for k in [1..Length(fGenerator[1])] do
          temp[k] := alpha[j]*(fGenerator[j][k]-DHilbert[j]);
        od;
        theta[j] := Copy(temp);
      od;

# characteristic polynomial for theta over the imaginary quadratic field
      thetaPol := PolyFromConjugatesOverImQuad (ord, Concatenation(theta), 1);
    until thetaPol<>false or l>3;

    if thetaPol=false then
      Print("real precision of ",prec," places was not sufficient for calculation!\n\n");
      Prec(precOld[1]); OrderPrec(ord, precOld[2][1], precOld[2][2]);
      return false;
    fi;
    if flag="char" then
      Prec(precOld[1]); OrderPrec(ord,precOld[2][1], precOld[2][2]);
      return thetaPol;
    fi;
    if flag=false or debug then
      Print("characteristic polynomial over quadratic imaginary field:\n");
      Print("  ",thetaPol,"\n\n");
    fi;

# generation of the rayclassfield over the hilbert classfield
    thetaK1Pol := PolyFromConjugatesOverHilbert (oK1All[1], oK1All[2], theta, 1);
    if thetaK1Pol=false then
      Prec(precOld[1]); OrderPrec(precOld[2][1]);
      return false;
    fi;
    oKfEqnOrder := Order (thetaK1Pol);
    if flag="field" then
      Prec(precOld[1]); OrderPrec(precOld[2][1]);
      OrderPrec(oKfEqnOrder,precOld[2][1],precOld[2][2]);
      return oKfEqnOrder;
    fi;
    if flag=false or debug then
      Print("ray class field:\n",oKfEqnOrder,"\n");
    fi;

# calculation of a relativ integral basis
    Print("\ncalculation of representation with integral relative basis...\n");
    alphaElt := oK1EltFromConjugates (oK1All[1], oK1All[2], alpha);
    if alphaElt=false then
      Print("real precision of ",prec," places was not sufficient for calculation!\n\n");
      Prec(precOld[1]); OrderPrec(precOld[2][1]);

      return false;
    fi;
    oK1Abs := OrderMaximal (OrderAbs (oK1All[1]));
    alphaElt := EltMove (alphaElt, oK1Abs);
    temp := PolyToList (thetaK1Pol);
    Apply(temp,x->EltMove(x,oK1Abs));
    if debug then Print("oKfEqnOrder: start\n"); fi;
    oKfEqnOrder := Order (Poly (PolyAlg(oK1Abs), temp));
    if debug then Print("oKfEqnOrder: end\n"); fi;
    if Abs(EltNorm(alphaElt,Z))=1 or Length(fResidues)<=2 or
       IdealIsIntegral(2/f) then
      oKf := oKfEqnOrder;
    else
#      alphaElt := EltMove (alphaElt, oKfEqnOrder);
#      temp := List([1..Length(fResidues)],x->EltMove(1,oKfEqnOrder));
#      for j in [3..Length(fResidues)] do
#        temp[j] := temp[j-1] * alphaElt;
#      od;
#      temp := List([1..Length(temp)],x->OrderBasis(oKfEqnOrder)[x]/temp[x]);
#      oKf := Order (oKfEqnOrder, temp);
#-----------
      Tden := EltNorm(alphaElt,Z);
      temp2 := Tden / alphaElt;
      temp := List([1..Length(fResidues)],x->EltMove(1,oK1Abs));
      for j in [3..Length(fResidues)] do
        temp[j] := temp[j-1] * temp2;
      od;
      temp2 := 1;
      for j in [1..Length(fResidues)-2] do
        temp2 := Tden * temp2;
        temp[Length(fResidues)-j] := temp[Length(fResidues)-j] * temp2;
      od;
      temp[1] := temp[1] * temp2;
      Tden := temp2;
      Tmat := MatDiag (oK1Abs, temp);
      if debug then Print("oKf: start\n"); fi;
#      oKf := Order (oKfEqnOrder, Tmat, Tden);
      oKf := Order (oKfEqnOrder, Tmat/Tden, 1);
      if debug then Print("oKf: end\n"); fi;
#-----------
    fi;
    OrderSetMaximal(oKf);
    if flag=false or debug then
      Print("integral relative basis over Hilbert class field:\n  ",OrderBasis(oKf),"\n\n");
    fi;
    Prec(precOld[1]); OrderPrec(precOld[2][1]); 

    return oKf;
end;
