
FactorProd := function(d)
# Returns the product of the primefactors of the integer d
local j, dFactors, prod;

    dFactors := Factor (Abs(d));
    prod := 1;
    for j in [1..Length(dFactors)] do
      prod := prod * dFactors[j][1];
    od;

    return prod;
end;


power1 := function(power)
# calculates (-1)^power
    if (power mod 2)=1 then
      return -1;
    else
      return 1;
    fi;
end;


ImQuadLegendre := function(mm, nn)
# Generalised Legendre (reziprocity) symbol m over n for imaginary quadratic fields
local m, n, res, change;

    m := Copy(mm);
    n := Copy(nn);
    if IntGcd(m,n)<>1 then
      res := 0;
    elif (m=0) or (n=0) then
      res := 1;
    else
      if n<0 then n := -n; fi;
      while (m mod 4)=0 do m := m/4; od;
      while (n mod 4)=0 do n := n/4; od;
      res := 1;
      while (m<>1) and (n<>1) do
        if (m mod 2)=0 then
          res := res*power1((n*n-1)/8);
          m := m/2;
        elif (n mod 2)=0 then
          res := res*power1((m*m-1)/8);
          n := n/2;
        elif m<0 then
          res := res*power1((n-1)/2);
          m := -m;
        elif m<n then
          res := res*power1(((n-1)*(m-1))/4);
          change := m;
          m := n;
          n := change;
        else
          m := m mod n;
        fi;
      od;
    fi;

    return res;
end;



ImQuadClassNum := function(ord)
# Calculates the class number of the imaginary quadratic field belonging to ord
# FH: seems to be quite slowly
local sum, j, bound;

    if OrderDegAbs(ord)<>2 or OrderDisc(ord)>0 then
      Error("ImQuadClassNum: Field must be imaginary quadratic!\n");
    fi;

    sum := 0;
    if (OrderDisc(ord) mod 2) = 1 then
      bound := (1-OrderDisc(ord))/2;
    else
      bound := -OrderDisc(ord)/2;
    fi;

    if OrderDisc(ord)=-3 or OrderDisc(ord)=-4 then
      return 1;
    else
      for j in [1..bound-1] do
        sum := sum+ImQuadLegendre(OrderDisc(ord),j);
      od;
      return sum/(2-ImQuadLegendre(OrderDisc(ord),2));
    fi;
end;



IdealEulerPhi := function(id)
# Euler's phi-function phi(id)=|(ord/id)^*| for integral ideals id
local p, classes;

    classes := 1;
    for p in Factor(id) do
      classes := classes * (IdealNorm(p[1])-1) * IdealNorm(p[1])^(p[2]-1);
    od;

    return classes;
end;



ImQuadResNum := function(f)
# Determines the number of multiplicative residues modulo ideal f
# and modulo torion units in an imaginary quadratic field
local j, ord, calc, zeta;

    ord := IdealOrder(f);
    calc := 1;
    zeta := OrderTorsionUnit(ord);
    for j in [1..OrderTorsionUnitRank(ord)-1] do
      if EltIsInIdeal(zeta^j-Elt(ord,1),f) then
        calc := calc+1;
      fi;
    od;

    return IdealEulerPhi(f)*calc/OrderTorsionUnitRank(ord);
end;

