


RayResidueIsCalculated := function(f, res)
# If a representant for the multiplikative residue res modulo ideal f
# and modulo torion units still has been calculated, true is returned,
# otherwise false
local  j, resCoeff, ord, test, calc, zeta;

#Print("RayResidueIsCalculated\n");

    resCoeff := EltToList(res);
    ord := IdealOrder(f);
    zeta := Copy(OrderTorsionUnit(ord));

    calc := false;
    for j in [1..OrderTorsionUnitRank(ord)-1] do
      test := EltToList(EltIdealReduce(zeta*res,f,"HNF_POS"));
#Print("test:",test,"\n");
      if (test[2]<resCoeff[2]) or ((test[2]=resCoeff[2]) and
         (test[1]<resCoeff[1])) then calc := true; fi;
      zeta := zeta*OrderTorsionUnit(ord);
    od;
    return calc;
end;



ImQuadRayResidues := function(f)
# Calculates representants for the multiplikative residues modulo ideal f
# and modulo torion units in an imaginary quadratic field
local res, ord, f1, f2,
      n1, n2, nMat, n1max, n2max, calculated,
      addelem, testelt, one;

    res := [];
    ord := IdealOrder(f);
    one := 1*ord;

    calculated := 0;
    nMat := IdealBasisUpperHNF(f)[2];
    n1max := Abs(MatElt(nMat,1,1));
    n2max := Abs(MatElt(nMat,2,2));
    f1 := FactorProd(n2max);
    f2 := 1*ord; for n1 in Factor(f/n2max) do f2:=f2*n1[1]; od;
    for n2 in [0..n2max-1] do
      for n1 in [0..n1max-1] do
        testelt := Elt(ord,[n1,n2]);
#Print("testelt:",testelt,"\n");
        if testelt<>0 and IntGcd(EltNorm(testelt,Z),f1)=1 
                      and testelt*ord+f2=one then
          addelem := not RayResidueIsCalculated (f, testelt);
        else addelem := false; fi;
#Print("addelem:",addelem,"\n");
        if addelem then
          calculated := calculated+1;
          res[calculated] := testelt;
        fi;
      od;
    od;

    return res;
end;






IdealClassReprSmall := function(arg)
# Finds a small integral representant for the ideal class given by the ideal
# arg[1]; if Length(arg)>1 and arg[2]="mult", then also the used multiplicative
# factor small is returned
local j, repr, small;
    repr := arg[1];
    repr := repr^-1;
    repr := Lat(repr);
    small := EltMove(LatShortestElt(repr)[1],IdealOrder(arg[1]));
    if Length(arg)>1 and arg[2]="mult" then
      return [arg[1]*small,small];
    fi;
    return arg[1]*small;
end;


IdealClassGroupRepr := function(ord)
# Calculates (small) integral representants for the ideal class group in order ord
local class, classFactors, runValues,
      j, k, dim, pos, prodPos, small;

    OrderClassGroup (ord);
#    OrderClassGroup (ord, 500, "fast", "Euler");
#    if not OrderClassGroupCheck(ord) then
#      Print("OrderClassGroup with 500 and fast failed ... --> use safe OrderClassGroup");
#      OrderClassGroup (ord, "Euler");
#    fi;
    classFactors := OrderClassGroupCyclicFactors (ord);
    dim := Length(classFactors);
    runValues := List([1..dim],x->0);
    prodPos := [];

    class := [];
    prodPos[1]:=1*ord;
    pos := 0;
    j := 1;
    while j>=1 do
#   calculate representants by multiplying out the cyclic factors
      if j=dim then
        for k in [1..classFactors[dim][2]] do
          pos := pos + 1;
          class[pos] := Copy(prodPos[dim]);
          if k<classFactors[dim][2] then
            prodPos[dim] := prodPos[dim]*classFactors[dim][1];
#            prodPos[dim] := IdealClassReprSmall (prodPos[dim]);
          fi;
        od;
        j := j-1;
        if j>=1 then
          prodPos[j] := prodPos[j]*classFactors[j][1];
#          prodPos[j] := IdealClassReprSmall (prodPos[j]);
        fi;
      else
        runValues[j] := runValues[j]+1;
        if runValues[j]>classFactors[j][2] then
          runValues[j] := 0;
          j := j-1;
          if j>=1 then
            prodPos[j] := prodPos[j]*classFactors[j][1];
#            prodPos[j] := IdealClassReprSmall (prodPos[j]);
          fi;
        else
          j := j+1;
          prodPos[j] := Copy(prodPos[j-1]);
        fi;
      fi;
    od;

    return class;
end;




IdealClassReprPrimSmall := function(idOld,prim)
# Finds a small integral representant new (prime to the rational integer
# prim) for the element of the ideal class given by the ideal idOld, and
# returns [new,new/idOld]; the used method is enumeration of small lattice
# points
local j, elt, lllLat, temp, denom, lBound, uBound, found, change;

    lllLat := LatLLL(Lat(idOld^-1));
    temp:=[1];
    Append (temp,List([2..Length(MatToColList(LatBasis(lllLat)))],x->0));
    temp := LatEltLength (LatElt (lllLat,temp));
    denom := IdealNorm(idOld);
    change := 1;
    lBound:=0; uBound := 10;
    found := false;

    repeat
      LatEnumLowerBound(lllLat,lBound*temp);
      LatEnumUpperBound(lllLat,uBound*temp);
      while LatEnum(lllLat) do
        elt := EltMove(LatEnumElt(lllLat),IdealOrder(idOld));
        if IntGcd(Norm(elt)*denom,prim)=1 then
          found := true;
          if Abs(Norm(elt/change))<1 then change:=elt; fi;
        fi;
      od;
      lBound:=uBound; uBound := 4*uBound;
    until found or uBound>160;

    return [idOld*change,change];
end;




IdealClassReprPrim := function(idOld, prim)
# Calculates an integral representant new (prime to prim) for the element of
# the ideal class group given by the ideal idOld, and returns [new,new/idOld]
local L2, p, q, add, L, pL, eL, small, new;

    L := Factor(1/idOld);

    L2 := Factor(prim*IdealOrder(idOld));
    for p in L2 do
      add := true;
      for q in L do
        if p[1]=q[1] then
          add:=false;
        fi;
      od;
      if add then Add(L,[p[1],0]); fi;
    od;

    new := [];
    pL := List(L,x->x[1]);
    eL := List(L,x->x[2]);
    new[2] := EltApproximation(pL,eL);
    new[1] := new[2]*idOld;

# the following seems to be useful to get small representants, but
# LatLLL() sometimes doesn't work in the KASH-version used by me - MK, Sept'96
    small := IdealClassReprPrimSmall(new[1],IdealMin(prim*IdealOrder(idOld)));
    new[1] := small[1];
    new[2] := new[2]*small[2];

    return new;
end;





IdealClassGroupReprPrim := function(arg)
# Calculates integral representants prime to arg[2] for the ideal class
# group in order arg[1]; if still calculated, representants for the ideal class
# group can be given in arg[3]
local j, ord, f, class;

    ord := arg[1];
    if Length(arg)=2 then
      f := arg[2]*ord;
      class := IdealClassGroupRepr(ord);
    else
      f := arg[2]*ord;
      class := Copy(arg[3]);
    fi;

    for j in [1..Length(class)] do
      if (class[j]+f)<>1*ord then
        class[j] := IdealClassReprPrim(class[j], f)[1];
      fi;
    od;

    return class;
end;











