#/homes/algebra/kant/V4/src/kash/lib/KashPvmMaster.g
#Print("KashPvm functions installed.\n");

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       PvmGetB
#CALLING   L := PvmGetB();
#SHORTDOC  Waits for any data to arrive.
#LONGDOC   Calls alternately {\tt PvmGet} and {\tt Sleep(1)} until valid
#          data arrives.
#ARGUS     L @ list
#CSRC
#PACKAGE   lib
#XREF      PvmGet.
#EXAMPLE   See {\tt PvmGet} or {\tt PvmClassgroup} (in
#          {\tt PvmClassgroup.kash})
#          for an example.
#**HISTORY*|*HISTORY***HISTORY***HISTORY**HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***|***C***C***C***C***C***C**C***C***C***C***C***C***C***C***C***C***/

PvmGetB := function()
  local d;
  d := PvmGet();
  while d = false do
    Sleep(1);
    d := PvmGet();
  od;
  return d;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       PvmGetAnswer
#CALLING   L := PvmGetAnswer(true | false);
#SHORTDOC  Waits for a valid answer to arrive from the slave.
#LONGDOC   This function performs a while-loop until a valid answer arrives
#          (i.~e.~a data list starting with 1).
#          
#          Master only: if the parameter is {\tt true} all incoming data is
#          echoed to the screen. The function will return after the first answer
#          arrived.
#ARGUS     L @ list @ first entry is in $\{1, 2, 4\}$, all others arbitrary
#CSRC
#PACKAGE   lib
#XREF      PvmGet.      
#EXAMPLE   See {\tt PvmClassgroup} (in {\tt PvmClassgroup.kash})
#          for an example.
#**HISTORY*|*HISTORY***HISTORY***HISTORY**HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***|***C***C***C***C***C***C**C***C***C***C***C***C***C***C***C***C***/

PvmGetAnswer := function(arg)
  local d, out;

  out := function(st)
    local  i;
    for i in [st..Length(d)-1] do
      Print(d[i], ", ");
    od;
    Print(d[Length(d)]);
  end;

  if Length(arg) <>1 or TYPE(arg[1]) <> "boolean" then
    return Error("usage: PvmGetAnswer( true | false)");
  fi;


  while true do
    d := PvmGetB();

    if d[1]=2 and arg[1] then
      Print("Error on   ", d[2], " : ");
      out(3);
    fi;
    if d[1]=4 and arg[1] then
      Print("Print from ", d[2], " : ");
      out(3);
    fi;
    if d[1]=1 then
      if arg[1] then
        Print("Answer: ");
        out(2);
        Print("\n");
      fi;
      return d;
    fi;
  od;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       PvmRead
#CALLING   PvmRead();
#SHORTDOC  Reads (and prints) all data already received from the slave(s).
#LONGDOC   The incoming data is ``beautyfied'' before printing.
#ARGUS
#CSRC
#PACKAGE   lib
#XREF      PvmGet.
#EXAMPLE
#**HISTORY*|*HISTORY***HISTORY***HISTORY**HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***|***C***C***C***C***C***C**C***C***C***C***C***C***C***C***C***C***/


PvmRead := function()
  local d, m, f, out;

  out := function(st)
    local  i;
    for i in [st..Length(d)-1] do
      Print(d[i], ", ");
    od;
    Print(d[Length(d)]);
  end;

  d := PvmGet();

  while (d<>false) do
    if d[1]=2 then
      Print("Error on   ", d[2], " : ");
      out(3);
    fi;
    if d[1]=4 then
      Print("Print from ", d[2], " : ");
      out(3);
    fi;
    if d[1]=1 then
      Print("Answer: ");
      out(2);
      Print("\n");
    fi;
    d := PvmGet();
  od;
end;

#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       PvmPread
#CALLING   PvmPread();
#SHORTDOC  Reads and prints all arrived data in plain style, i.e. as provided
#          by \kash.
#LONGDOC  
#ARGUS
#CSRC
#PACKAGE   lib
#XREF      PvmGet.
#EXAMPLE
#**HISTORY*|*HISTORY***HISTORY***HISTORY**HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***|***C***C***C***C***C***C**C***C***C***C***C***C***C***C***C***C***/
PvmPread := function()
  local d, m, f, out;

  d := PvmGet();

  while (d<>false) do
    Print(d, "\n");
    d := PvmGet();
  od;
end;
