if not IsBound(Info) then
  Info := Ignore;
fi;

if not IsBound(XIsIdeal) then
  Read("../lib/RCLF/ideal.g");
fi;
if not IsBound(OrderMergeWithAutomorphisms) then
  Read("../lib/RCLF/test4.g");
fi;
if not IsBound(ModKernel) then
  Read("../lib/RCLF/Kernel.g");
fi;
if not IsBound(GetGens) then
  Read("../lib/RCLF/GetGens.g");
fi;
Read("../lib/RCLF/Root.g");
Read("../lib/RCLF/test3.g");

if not IsBound(_g) then
  Read("../lib/RCLF/gasman.g");
fi;

PRINTLEVEL("LLL", 1);
PRINTLEVEL("HOLZ", 1);  #PolyZeros...
PRINTLEVEL("ORDER_ELT_DIV", -1); #divide in equation orders
PRINTLEVEL("ELT_SIMPLIFY", -10); # sometimes simplification

RayGensVar := function(id, inf, modu, index, fxn, start)
  local i, j, k, l, p, fp, o, sol, lp, m, nm;

  o := XIdealOrder(id);
  p := NextPrime(start);
  nm := XIdealNorm(id);
  if IsInt(modu) then 
    nm := IntLcm(nm,  modu);
  else 
    nm := IntLcm(nm, XIdealNorm(modu));
  fi;
  sol := [];
  lp := [];
  GASMAN("g");
  repeat
    if IntGcd(p, nm)=1 then
      fp := XIdealFactor(XIdeal(o, p));
      fp := Filtered(fp, j->XIdealNorm(j[1])<150 or XIdealDegree(j[1])=1);
      for j in fp do
        Add(sol, fxn(j[1]));
        Add(lp, j[1]);
        if Length(sol)>=Length(sol[1][1]) then
          i := sol;
          IsMat(i);
          i := MatHermiteColLower(MatTrans(i));
          i := MatToRowList(i);
          k := 1;
          for l in [1..Minimum(Length(i), Length(i[1]))] do
            k := k*i[l][l];
          od;
          if Gcd(k, index) = 1 then
            m := List(lp, XIdealMin);
            Apply(m, Log);
            m := Maximum(m);
            Info("after reduction: Max(Log(Min(gens))) = ", m, "\n");
            GASMAN("g");
            return [lp, sol];
          fi;
        fi;
      od;
    fi;
    p := NextPrime(p);
    GASMAN("g");
  until false;
end;

RayGens := function(id, inf, modu, start)
  return RayGensVar(id, inf, modu, 0, i->IdealRayClassRep(i, id, inf), start);
end;

IntChineseRemainder := function(a)
  local i, j;

  for i in [2..Length(a)] do
    j := IntXGcd(a[1][2], a[i][2]);
    a[1][1] := a[1][1]*j[2][2]*a[i][2] + a[i][1]*j[2][1]*a[1][2];
    a[1][2] := a[1][2] * a[i][2];
  od;
  return a[1][1] mod a[1][2];
end;

Kappa := function(deg) # Cohen, page 26
  local f, i, j, r, s, t;

  f := Factor(deg);
  r := 1;

  t := [];
  for i in f do
    if i[1] = 2 then
      if   i[2] = 1 then Add(t, rec(kappa := 1, modul := 2, ord := 1));
      elif i[2] = 2 then Add(t, rec(kappa := 3, modul := 4, ord := 2));
      elif i[2] >=3 then Add(t, rec(kappa := 5, modul := 2^i[2],
                                                          ord := 2^(i[2]-2)));
                         Add(t, rec(kappa := 2^i[2]-1, modul := 2^i[2],
                                                          ord := 2));
      fi;
    else
      j := FF(i[1]);
      s:=FFPrimitiveElt(j);
      j:=MatToRowList(MatMove(Mat(j,[[s]]),Z))[1][1];
      if i[2]>1 and j^(i[1]-1) mod i[1]^2 <> 1 then
        Add(t, rec(kappa := j+i[1], modul := i[1]^i[2],
                                       ord := i[1]^(i[2]-1)*(i[1]-1)));
      else
        Add(t, rec(kappa := j, modul := i[1]^i[2],
                                       ord := i[1]^(i[2]-1)*(i[1]-1)));
      fi;
    fi;
  od;
  Info("Kappa (", deg,")=", t, "\n");
  return t;
end;

OrderCyclotomicAdd := function(o, deg)
  local c, O, Or, N, deg_o, i, j, kappa, l, k, d;

  if TYPE(o) = "KANT ring" then
    if deg = 2 then
      O := o;
    else
      O := OrderCyclotomic(deg);
    fi;
    i := function(x)
      return XIdeal(Z, XIdealNorm(x));
    end;
    return [O, O, i, XOrderDeg(O)];
  fi;

  if XOrderTorsionUnitRank(o) mod deg<>0  then
    c := OrderCyclotomic(deg);
    O := OrderMergeWithAutomorphisms(o, c);
    Or := Order(PolyMove(OrderPoly(O[4]), o));
    OrderInstallHom(Or, O[2], XEltMove(OrderBasis(c)[2], O[2]));
    OrderInstallHom(c, Or, OrderBasis(Or)[2]);
    i := [XEltMove(OrderBasis(OrderEquationOrder(o))[2],
                                    OrderEquationOrder(O[2])),
          XEltMove(OrderBasis(c)[2], OrderEquationOrder(O[2])),
          OrderBasis(OrderEquationOrder(O[2]))[2]];
    Apply(i, EltToList);
    IsMat(i);
    j := MatKernel(i);
    i := -XEltMove(j[1][1]/j[1][3], o)*
                        XEltMove(OrderBasis(OrderEquationOrder(o))[2], Or)-
          XEltMove(j[1][2]/j[1][3], o)*XEltMove(OrderBasis(c)[2], Or);
    OrderInstallHom(OrderEquationOrder(O[2]), Or, i);
    OrderInstallHom(O[2], Or, true);
    i := List(OrderAutomorphisms(O[2], []), i->XEltMove(EltAutomorphism(
                                   XEltMove(OrderBasis(Or)[2], O[2]), i), Or));
    i := Set(i);
    OrderAutomorphisms(Or, i);
    O := OrderAbs(O[2]);
    d := O;
    i := Factor(deg*o);
    for j in i do
      Or := OrderPMaximal(Or, j[1]);
      Or := OrderSimplify(Or);
    od;
    OrderSetMaximal(Or);
    Info("Start RelOrderMax\n");
    O := RelOrderMax(Or);
    Info("Start LLL\n");
    OrderSetMaximal(O);
    Info("Start Short\n");
    c := OrderShort(O);
    Info("Stop Short\n");
    if c<>false then O := c; else c := O; fi;
    Info("Start LLL\n");
    O := OrderLLL(O);
    O := OrderSimplify(O); # CF new
    j := XEltMove(OrderBasis(OrderEquationOrder(O))[2], d);
    j := List(OrderAutomorphisms(d, []), i->XEltMove(EltAutomorphism(j, i),
                                                       OrderEquationOrder(O)));
    OrderAutomorphisms(OrderEquationOrder(O), j);
    Info("Stop LLL\n");
    deg_o := XOrderDeg(Or);
    N := XIdealNorm;
    c := XEltMove(OrderBasis(c)[2], Or);
  else
    O := o;
    Or := O;
    deg_o := 1;
    N := x->x;
    c := XOrderTorsionUnit(O)^(XOrderTorsionUnitRank(O)/deg);
  fi;

  j := [];

  kappa := Kappa(deg);
  l := MultiCounterInit(Length(kappa));

  while MultiCounterInc(l, List(kappa, k->k.ord-1)) do
    k := [];
    if deg mod 8 = 0 then
      k[1] := [kappa[1].kappa^l[1]* kappa[2].kappa^l[2] mod kappa[1].modul,
               kappa[1].modul];
      for i in [3..Length(l)] do
        Add(k, [kappa[i].kappa^l[i] mod kappa[i].modul, kappa[i].modul]);
      od;
    else
      for i in [1..Length(l)] do
        Add(k, [kappa[i].kappa^l[i] mod kappa[i].modul, kappa[i].modul]);
      od;
    fi;
    k := c^IntChineseRemainder(k);
    i := Eval(OrderPoly(Or), k);
    if i=0 then Add(j, k); fi;
  od;

  OrderAutomorphisms(Or, j);

  Info("Abs. Order\n", O, "\nRel. Order\n", Or, "\n");

  return [O, Or, N, deg_o];
end;

KummerCreate := function(k, p, S)  # p is NO prime!!!
  local st, i, j, k, kl, a, beta, zeta;
  st := rec(baseField := k, deg := p);
  if IsList(S) then
    st.gens := Copy(S);
  else
    st.gens := [S];
  fi;

  zeta := XOrderTorsionUnit(k)^(XOrderTorsionUnitRank(k)/p);
  st.zeta := zeta;
  st.operations:=rec(Print := function(i)
      Print(i.baseField, "\n", i.deg, " ", i.gens, "\n");
    end
  );
  return st;
end;


KummerArtinPrimeIdeal := function(st, p)
  local OK, x, i, f, j, k, base, zeta, a, b, pow, mu, zeta_pow;

  if not IsBound(st.artinPrime) then
    st.artinPrime := [];
  fi;
  for i in [1..Length(st.artinPrime)] do
    if st.artinPrime[i][1] = p then return st.artinPrime[i][2]; fi;
  od;
  a := [];
  XIdealBasis(p);
  XIdealResidueField(p);

  zeta := st.zeta;

  for i in [1..Length(st.gens)] do
    mu := st.gens[i];
    mu := XEltToFFE(mu, p);
    mu := mu^((XIdealNorm(p)-1)/st.deg);
    mu := XFFEToElt(mu, p);
    j := 0;
    OK :=0;
    while j<st.deg do
      if XEltIsInIdeal(mu-zeta^j, p) then
        Add(a, j);
        OK := OK +1;
      fi;
      j := j+1;
    od;
    if OK<>1 then
      Error("Frobenius not found (or too many found)");
    fi;
  od;
  Add(st.artinPrime, [p, a]);
  IsMat(a);
  return a;
end;

KummerArtinIdeal := function(st, id)
  local i, x, j, m;

  if not IsBound(st.artin) then
    st.artin := [];
  fi;
  for i in [1..Length(st.artin)] do
    if st.artin[i][1] = id then
    return st.artin[i][2]; fi;
  od;

  j := Factor(id);
  m := KummerArtinPrimeIdeal(st, j[1][1])*j[1][2];
  for i in [2..Length(j)] do
    m := m + KummerArtinPrimeIdeal(st, j[i][1])*j[i][2];
  od;
  for i in [1..Length(m[1])] do
    m[1][i] := m[1][i] mod st.deg;
  od;

  Add(st.artin, [id, m]);
  return m;
end;

_with_reduce := true;
XEltLogs := function(e)
  if IsInt(e) then
    return Mat(R, [[Log(Abs(e))]]);
  else
    return EltLogs(e);
  fi;
end;
EltL := function(e, lp)
  local a, b;
  a := MatToRowList(XEltLogs(e))[1];
  for b in lp do
#    Add(a, Log(XIdealNorm(b))*EltValuation(b, e));
    Add(a, EltValuation(b, e));
  od;
  return a;
end;

Redu := function(sb)
  local Redu2, Redu3, logs, sunit, redex, i, r, unit, redex2, redex3,
        j, k, l, scale, gen, q, bound, b, finish, op, pp, np;

  r := Sum(XOrderSig(XEltOrder(sb.S[1][1])))-1;
  if r=0 then
    return;
  fi;

  HEAP(XEltOrder(sb.S[1][1]), "CHECK_LOG_PRECISION", 1);
  op := OrderPrec(EltOrder(sb.S[1][1]))[1];
  pp := Prec();
  i := Maximum(List(Flat(sb.S[2]), Abs));
  i := Floor(Log(i)/Log(10))+Length(sb.S[2][1])+20; # not perfect (should work)!!
  OrderPrec(XEltOrder(sb.S[1][1]), i);
  Prec(i);
  np := i;
#  np := pp;

  logs := [];
  for i in sb.S[1] do
    Add(logs, EltL(i, sb.SMod));
  od;
  sunit  := List(logs, i->i{[r+2..Length(i)]}); 
  logs   := List(logs, i->i{[1..r+1]});

  IsMat(logs);
  IsMat(sunit);
  sunit := MatMove(sunit, Z);
  IsMat(sb.S[2]);

  unit := sb.S[2]*sunit;
  unit := unit{[r+1..Length(unit)-1]};
  IsMat(unit);

  Redu3 := function(start)
    local l, i, j, k, gen, redex, genex, ml, val;

    genex  := MatToRowList(start)[1]{[r+1..MatCols(start)-1]};
    gen  := MatMove(Mat(Z, [genex])*unit, Z);
    
    ml := MatMLLL(MatTrans(unit));  
    k := MatMove(ml[1], Z)^-1;
    i := MatToColList(k);
    Apply(i, x->Lcm(List(x, Denominator)));

    redex := List(gen[1], x->0);
    for j in [1..Length(i)] do
      redex := redex + sb.deg*i[j]*IntQuo(gen[1][j], sb.deg*i[j])*
         MatToColList(k)[j];
    od;
    redex := MatToColList(ml[2]*MatTrans(Z, [redex]))[1];
    
    j := MatToRowList(start)[1];
    redex := Concatenation(j{[1..r]}, j{[r+1..Length(j)-1]} - 
                                                  redex, [j[Length(j)]]);

    val := Mat(Z, [redex{[r+1..Length(j)-1]}])*unit;
    val := MatToRowList(val)[1];
    return [val, redex];
  end;


  Redu2 := function(start)
    local l, i, j, k, gen, redex, crit, noncrit, genex, ml, val;

    genex  := MatToRowList(start)[1]{[r+1..MatCols(start)-1]};
    gen  := Mat(Z, [genex])*unit;
    redex := 0*genex;
    
    ml := MatMLLL(MatTrans(unit)*sb.deg);  
    # find critical indices:
    i := MatToRowList(ml[1]);
    for j in i do Apply(j, x->Minimum(1, Abs(x))); od;
    i := Sum(i);
    #critical are at least those k with i[k] >1!
    Apply(i, x->Minimum(1, x-1));
    i := List([1..Length(i)], x->i[x]*x);
    crit := Filtered(i, x->x<>0);
    #but we have to take a closure...
    i := Set(crit);
    crit := [];
    l := MatToRowList(ml[1]);
    Apply(l, x->List(x, k->k<>0));
    while i<> [] do
      j := i[1];
      SetAdd(crit, j);
      SetRemove(i, j);
      k := 0;
      repeat
        k := Position(List(l, x->x[j]), true, k);
        if k<> false and Position(crit, k)=false then
            SetAdd(i, k);
        fi;
      until k=false;
    od;

    noncrit := Set([1..MatCols(ml[1])]);
    SetSubtract(noncrit, crit);

    i := MatToColList(ml[1]);
    for j in noncrit do
      l := Position(List(i[j], x->x=0), false);
      k := i[j][l];
      if k>0 and gen[1][l]>sb.deg * k then  
        redex[j] := IntQuo(gen[1][l], sb.deg * k);
      elif k<0 and gen[1][l]>-sb.deg * k then
        redex[j] := -IntQuo(gen[1][l], -sb.deg * k);
      fi;
    od;

    #now the hard part...  (improvement: find a decomposion if possible)
    if crit <> [] then
      for i in noncrit do gen[1][i] := 0; od;
      i := MatToColList(ml[1]){crit};
      for j in i do Apply(j, Abs); od;
      i := Sum(i);
      i := MatSolve(ml[1], 1/2*MatTrans(Z, [i]))[1];
      i := MatToColList(i)[1];
      j := MatSolve(ml[1], MatMove(MatTrans(gen), Q))[1];
      j := MatToColList(j)[1];
      j := -i+j;
      Apply(j, Round);
      redex := redex + j;
    fi;
    redex := MatToColList(ml[2]*MatTrans(Z, [redex]))[1];
   
    
    j := MatToRowList(start)[1];
    redex := Concatenation(j{[1..r]}, j{[r+1..Length(j)-1]} - 
                                                  sb.deg*redex, [j[Length(j)]]);

    val := Mat(Z, [redex{[r+1..Length(j)-1]}])*unit;
    val := MatToRowList(val)[1];
    return [val, redex];
  end;

  redex2 := Redu2(sb.unram_mat);
  Info("Starting with ", redex2, "\n");
  finish := false;
  while not finish do
    finish := true;
    for i in [2..sb.deg-1] do
      if IntGcd(sb.deg, i)=1 then
	j := i*redex2[2];
	Apply(j, x->x mod sb.deg);
	IsMat(j);
	redex3 := Redu2(j);
#        Print(redex3, "\n");
	if Sum(redex2[1])>Sum(redex3[1]) then
	  redex2 := redex3;
          finish := false;
	  Info("now ", redex2, "\n");
	fi;
      fi;
    od;

    for i in [1..sb.deg-1] do
      if IntGcd(sb.deg, i)=1 then
	j := i*redex2[2];
	Apply(j, x->x mod sb.deg);
	IsMat(j);
	redex3 := Redu3(j);
#        Print(redex3, "\n");
	if Sum(redex2[1])>Sum(redex3[1]) then
	  redex2 := redex3;
          finish := false;
	  Info("now ", redex2, "\n");
	fi;
      fi;
    od;
  od;


  redex := redex2[2];
  IsMat(redex);

  unit := sb.S[2]*logs;

  Prec(pp);
  OrderPrec(EltOrder(sb.S[1][1]), op);

  unit := MatMove(unit, R);
  logs := unit;
  

  gen  := redex*unit;
  unit := unit {[1..r]};
  IsMat(unit);

  l := Lat(MatTrans(unit)*sb.deg, "basis");
  
  k := MatOrthogonal(LatBasis(l)); 
  b := MatTrans(k[1])*k[1];
  scale := MatDiag(R, List([1..Length(b)], i->1/b[i][i]));
  gen := gen * k[1]*scale * MatTrans(k[2]);
  
  LatEnumRefVec(l, LatElt(l, MatToRowList(gen)[1]));

  j := LatShortestElt(l)[1];

  Info(LatEltLength(j), "\t", LatEltLength(LatEnumRefVec(l)-j), "\n");
  
  j := LatEltToList(j);
  while Length(j)<Length(sb.unram_mat[1]) do
    Add(j, 0);
  od;
  j := MatToRowList(redex)[1] - sb.deg * j;
  Info("Maximum j is ", j, "\n");
  IsMat(j);
 
  i := j * logs;
  j := j*sb.S[2];
  l := i*MatTrans(i);
  q := EltL(sb.unramified[Length(sb.unramified)][1], []);
  IsMat(q);
  q := q*MatTrans(q);

  l := MatMove(l, R);
  q := MatMove(q, R);

  Info(q[1][1], "\t", l[1][1], "\n");
  if q[1][1]>l[1][1] then
    IsMat(j);
    IsMat(sb.S[1]);
    unit := EltPowerProduct(XEltOrder(sb.S[1][1][1]), sb.S[1], j);
    sb.S[1] := Flat(MatToRowList(sb.S[1]));
    sb.unramified[Length(sb.unramified)][3] := 1/unit;
    Info("Reduced ", sb.unramified[Length(sb.unramified)][1], 
                      " to ", unit, "\n");
    sb.unramified[Length(sb.unramified)][1] := unit;
  else
    Info("Reduce on ", sb.unramified[Length(sb.unramified)][1], 
                      " not successful\n");
  fi;
  sb.S[2] := MatToRowList(sb.S[2]);
end;

BaseCreate := function(o, deg, exact, rg, INF, CFT)
  local sb, m, S, mat, i, j, x, k, t, lp, O, c, Or, N, st, ff, rgO, inf,
        deg_o, kappa, gens, ineu, max_ineu, rt, l, n, ord, cg, add_ideal;

  rt := Runtime();

  if TYPE(o) = "record" and o.deg = deg then
    sb := o;
    o := sb.field;
    c := sb.c;
  else
    sb := rec(field := o, deg:=deg);
    sb.operations:=rec(Print := function(i)
        Print(i.field, "\nGenerators:", i.unramified, "\n");
      end
    );
    c := OrderCyclotomicAdd(o, deg);
    sb.c := c;
    sb.CFT := CFT;
  fi;
  RayClassGroup(rg, INF);
  sb.rg := rg; sb.inf := INF;


  O     := c[1];
  Or    := c[2];
  N     := c[3];
  deg_o := c[4];

  sb.Or := Or;
  Info("Time up to now: ", Runtime()-rt, "\n");
  Info(O, "\n");
  ## XOrderClassGroup(O, XOrderDeg(O)*25, "euler", "fast");
  ##OrderPrec(O, 100);
  XOrderClassGroup(O, XOrderDeg(O)*40, "euler", "fast");
  Info("Time up to now: ", Runtime()-rt, "\n");
  Info(O, "\n");

  rgO := XIdealMove(rg, O);
  S := [];
  # better start with the prime we already know: deg and rg!
  Add(S, deg*rgO);
  Apply(S, x->Factor(XIdeal(XIdealOrder(x),Min(x))));
  S := Flat(S);
  S := Filtered(S, XIsIdeal);
  S := Set(S);
  Info("S:(before enlarging):\n", S, "\n\n");
  # now check, if this generates the classgroup:
  if XOrderClassGroup(O)[1]>1 then
    cg := XOrderClassGroup(O)[2];
    m := List(S, i->List(IdealClassRep(i), j->j[2]));
    Append(m, MatDiag(Z, List(XOrderClassGroup(O)[2], j->
                  IntGcd(deg^10, j))));
    IsMat(m);
    #Print(m, "\n");
    m := MatHermiteRowMod(XOrderClassGroup(O)[1], m);
    n := Copy(m);

    i := 2;
    j := [];
    k := 0;
    l := Product(List([1..Length(cg)], i->m[i][i]));
#    while IntGcd(XOrderClassGroup(O)[1],l) > 1 do
    while IntGcd(deg,l) > 1 do
      if k>=Length(j) then
        k := 1;
        j := Factor(i*O);
        i := NextPrime(i);
      else
        k := k+1;
      fi;
      if Position(S, j[k][1])= false then
        t := List(IdealClassRep(j[k][1]), j->j[2]);
        #Print(t, "\n");
        ord := Lcm(List([1..Length(t)], x->cg[x]/Gcd(t[x], cg[x])));
        Info("Ord(", j[k][1], ") is ", ord, " l now ", l, "\n");
        if true or not _with_reduce or ord< deg*deg then
          Add(n, t);
          IsMat(n);
          n := MatHermiteRowMod(XOrderClassGroup(O)[1], n);
          if Product(List([1..Length(cg)], i->n[i][i])) <l
                             then
            m := Copy(n);
            Add(S, j[k][1]);
            l := Product(List([1..Length(XOrderClassGroup(O)[2])], i->n[i][i]));
            Info("l now: ", l, "\n");
          else 
            n := Copy(m);
          fi;
        fi;
      fi;
    od;
  fi;
  Info("S:(generating the Classgroup):\n", S, "\n\n");
  Apply(S, x->Factor(XIdeal(XIdealOrder(x),Min(x))));
  S := Flat(S);
  S := Filtered(S, XIsIdeal);
  S := Set(S);
  Info("S:(finally):\n", S, "\n\n");
  Info("Time up to now: ", Runtime()-rt, "\n");
  m := Product(S);
  Info("module: ", m, "\n");
  sb.SMod := S;
  S := XOrderSUnits(O, S, "raw");
  Info("SUnits (without Torsion, raw) ", S, "\n\n");
  l := [];
  i := MatToColList(S[2]);
  for j in i do
    k := 1;
    if Maximum(List(j, Abs))>20 then
      for n in [1..Length(j)] do
        k := k*S[1][n]^(j[n] mod deg);
      od;
    else
      for n in [1..Length(j)] do
        k := k*S[1][n]^j[n];
      od;
    fi;
    Add(l, k);
    Add(j, 0);
  od;
  Add(i, Concatenation(List([1..Length(i[1])-1], j->0), [1]));
  Info("\"SUnits\" (without Torsion) ", l, "\n\n");
  Add(l, XOrderTorsionUnit(O));
  Info("SUnits:\n", l, "\n\n");
  Info("Time up to now: ", Runtime()-rt, "\n");
  sb.S := [Concatenation(S[1], [XOrderTorsionUnit(O)]), i];
  S := l;
  st := KummerCreate(O, deg, S);

  sb.st := st;

  rgO := XIdealNorm(m);
  m := m*Lcm(List(sb.S[1], i->XIdealNorm(XIdeal(i))))*deg;

  inf := List([1..XOrderSig(O)[1]], i->i);
  Info("f:\n", m, inf, "\n\n");

  Info("Time up to now: ", Runtime()-rt, "\n");
  gens := RayGensVar(m, inf, rgO*XOrderIndex(O), RayClassGroup(rg, INF)[1], 
                               i->KummerArtinPrimeIdeal(sb.st, i), 1000);  
  sb.gens := gens;
  Info("RayGens:\n", gens[1], "\n\n");

  sb.RG := m;
  sb.INF := inf;

  mat := [];
  for i in gens[1] do
    Add(mat, IdealRayClassRep(N(XIdealMove(i, Or)), rg, INF));
  od;
  IsMat(mat);
  sb.mat := mat;

  ineu := 1;
  sb.unramified := [];
  t := RayClassGroup(rg,INF)[2];
  if CFT<>[] then
    max_ineu := Length(CFT);
  else
    max_ineu := Length(t);
  fi;
  while CFT=[] and (t[ineu] mod deg<>0 or 
        (exact and IntGcd(t[ineu]/deg, deg)<>1)) do
    ineu := ineu +1;
  od;
  while ineu<= max_ineu do
    mat := Copy(sb.mat);

    if CFT<>[] then
      Append(mat, CFT[ineu]);
      IsMat(mat);
      t := MatKernel(mat, deg); #### CHECK, probably without deg!!
      t := MatToRowList(t);
      for i in [1..Length(t)] do
        t[i] := t[i]{[1..Length(sb.mat)]};
        if ForAll(t[i], j->j=0) then
          Unbind(t[i]);
        fi;
      od;
     t := Set(t);
     IsMat(t);
    else
      for i in [1..Length(t)] do
        if i=ineu then
          t[i] := deg;
        else
          t[i] := 0;
         for j in mat do j[i] := 0; od;
        fi;
      od;

      if not IsMat(mat) then Error("s.th. went wrong"); fi;
      t := MatKernel(mat, deg);
    fi;

    mat := [];

    k := List(gens[1], i->KummerArtinPrimeIdeal(sb.st, i));
    IsMat(k);
    k := t*k;
    k := MatToRowList(k);
    for i in k do
      Apply(i, j->j mod deg);
    od;
    IsMat(k);

    mat := MatTrans(k);

    sb.critical_matrix := mat;
    IsMat(mat);
    k := MatKernel(mat, deg);

    mat := k;

  Info("Time up to now: ", Runtime()-rt, "\n");
    if Length(mat)>1 then
#      Error("kernel should be of dimension 1");
      Info("Kernel of dim>1\n", mat, "\n");
      mat := MatHermiteRowMod(deg, mat);
      Info("Kernel of dim>1\n", mat, "\n");
      i := 1;
      while i<=Length(mat) and mat[i][i] mod deg = 0 do 
        i := i+1; 
      od;
      mat := mat[i];
      IsMat(mat);
      Info("New matrix\n", mat, "\n");
    fi;
  

    #reduce kern (find a basis that yields minimal generators)
    #we do this per generator only (for now)


    k := [];
    n := MatTrans(Z, sb.S[2])*MatTrans(mat);
    n := MatToColList(n)[1];
    mat := MatToRowList(mat);

    for i in [1..Length(mat)] do
      x := Gcd(mat[i]);
      mat[i] := mat[i]/x;
      n := n/x;
      l := List(n, j->j mod (deg/x));
      t := Sum(l);
      for j in [2..deg/x] do
        if IntGcd(deg/x, j) = 1 then
          if Sum(List(n, t-> (t*j) mod (deg/x)))<t then
            l := List(n, t-> (t*j) mod (deg/x));
            t := Sum(l);
          fi;
        fi;
      od;

      t := 1;
      for j in [1..Length(sb.S[1])] do
        t := t*sb.S[1][j]^l[j];
      od;

      if t<>1 then
        n := 1;
        for j in [1..Length(sb.S[1])] do
          n := n*(1/sb.S[1][j])^l[j];
        od;
        Add(sb.unramified, [t, deg/x, n]);
        
      else
        Error("trivial element");
      fi;
    od;
    IsMat(mat);
    sb.unram_mat := mat;

    t := RayClassGroup(rg,INF)[2];
    ineu := ineu +1;
    while CFT=[] and (ineu <= Length(t) and
      (t[ineu] mod deg<>0 or (exact and IntGcd(t[ineu]/deg, deg)<>1))) do
      ineu := ineu +1;
    od;

    if _with_reduce then
      Redu(sb);
    fi;
  od;

  GASMAN("g");
  Info(sb, "\n");
  Info("Time for BaseCreate: ", Runtime()-rt, "\n");
  Info("OK\n****************************************\n\n");
  return sb;
end;

Intersect := function(sb, deg, exact, offset, CFT)
# We have
#             O = o(zeta_deg, alpa^1/deg'), deg'|deg
#            /
#          o(zeta_deg)   = G
#         /               /
#        o               Q
# We are looking for o(alpha^1/deg) (NOT deg') approx....
# rg, inf is the RayClassGroup in o, RG, INF in G
  local auto1, auto2, kappa, mu, ox, deg_o, degs, i, j, k, l, m, n,
        continueOneAutomorphism, zeta, auto, artin_prime, O, o,
        artin_prime_calc, artin_ideal, artin_ideal_calc, ray, t, gens, ineu,
        automorphism_mult, factor_g, rest_g, automorphism_apply,
        automorphism_find, new, a, b, max_ineu, Newton3;
        

  Newton3 := function(elt, deg, o) # EltMinPoly by lin. alg.
    local L, i, j, m, me, den;

    me := EltToList(elt);
    IsMat(me);
    me := MatTrans(me);
    m := EltRepMat(elt);
    if m[1]<>1 then m := m[2]/m[1];
    else m := m[2];
    fi;
    L := [EltToList(Elt(XEltOrder(elt), 1)), Copy(me)];
    IsMat(L[1]);
    L[1] := MatTrans(L[1]);

    for i in [3..deg+1] do
      L[i] := m * L[i-1];
      Info("Newton3: Step ", i, "\n");
    od;
    Apply(L, i->MatToColList(i)[1]);

    while not IsInt(L[1][1]) and not IsRat(L[1][1]) and XEltOrder(L[1][1])<>o do
      Apply(L, i->Flat(List(i, EltToList)));
    od;
    den := function(a)
      if IsRat(a) then
        return Denominator(a);
      elif IsInt(a) then
        return 1;
      else
        return EltDen(a);
      fi;
    end;
    den := Lcm(List(L, i->Lcm(List(i, den))));
    IsMat(L);
    sb.L := L;  ### CF
    if o=Z or o=Q then 
      L := MatKernel(L);
    else
      L := ModKernel(L, true);
    fi;
    if L=false or MatRows(L)>1 or MatRows(L)=0 then return false; fi;
    L := MatToRowList(L)[1];
    L := Reversed(L);
    if L[1]<>1 then L := (1/L[1])*L; fi;
    return [Poly(PolyAlg(o), L), den];
  end;

  auto1 := function(elt, data)
    local i, j, k, tmp, ddata, elt_i, mu;

    if Length(data)=2 then
      tmp := EltToList(data[2]);
      j := List(tmp, i->i<>0);
      if Length(Filtered(j, i->i=true)) <> 1 then
        Error("should not happen");
      fi;
      j := Position(j, true);
      mu := XEltMove(sb.unramified[offset][1], XOrderCoefOrder(O));
      ddata := tmp[j];
      tmp := [EltToList(Elt(O, 1))];
      for i in [2..XOrderDeg(O)] do
       tmp[i] := List([1..XOrderDeg(O)], k->0);
       tmp[i][((i-1)*(j-1)) mod XOrderDeg(O) +1] :=
             ddata^(i-1)*mu^(IntQuo((i-1)*(j-1), XOrderDeg(O)));
      od;
      IsMat(tmp);

      ddata := [];
      j := OrderBasis(O, OrderEquationOrder(O));
      for i in j do
        i := EltToList(i);
        IsMat(i);
        i := i*tmp;
        i := MatToRowList(i)[1];
        Add(ddata, i);
      od;
      IsMat(ddata);
      data[3] := ddata;
      data[4] := 0; data[5] := 0; data[6] := 0;
      data[7] := [];
    fi;

    i := Position(List(data[7], j->j[1]), elt);
    if i<>false then
      data[6] := data[6]+1;
      return data[7][i][2];
    fi;
    i := Position(List(data[7], j->j[1]), -elt);
    if i<>false then
      data[6] := data[6]+1;
      return -data[7][i][2];
    fi;


    if XEltOrder(elt) = XOrderCoefOrder(O) then
      data[4] := data[4] +1;
      i := EltAutomorphism(elt, data[1]);
      Add(data[7], [elt, i]);
      return i;
    fi;

    data[5] := data[5] +1;

    elt_i := elt;
    elt := EltToList(elt);
    Apply(elt, i->EltAutomorphism(i, data[1]));
    IsMat(elt);
    elt := elt * data[3];
    elt := Elt(O, MatToRowList(elt)[1]);
    Add(data[7], [elt_i, elt]);
    GASMAN("g");
    return elt;
  end;

  auto2 := function(elt, data)
    local i, j, tmp, ddata, elt_i;

    if Length(data)=1 then
      tmp := [Elt(XEltOrder(data[1]), 1)];
      for i in [2..XOrderDeg(O)] do
        tmp[i] := data[1] * tmp[i-1];
      od;
      data[2] := tmp;
      data[4] := 0; data[5] := 0; data[6] := 0;
      data[7] := [];
    fi;

    if XEltOrder(elt) = XOrderCoefOrder(O) then
      data[4] := data[4]+1;
      return elt;
    fi;

    i := Position(List(data[7], j->j[1]), elt);
    if i<>false then
      data[6] := data[6]+1;
      return data[7][i][2];
    fi;
    i := Position(List(data[7], j->j[1]), -elt);
    if i<>false then
      data[6] := data[6]+1;
      return -data[7][i][2];
    fi;


    elt_i := elt;
    data[5] := data[5]+1;

    elt := EltToList(elt);
    for i in [1..XOrderDeg(O)] do
      elt[i] := elt[i]*data[2][i];
    od;
    elt := Elt(O, elt);
    Add(data[7], [elt_i, elt]);
    GASMAN("g");
    return elt;
  end;

  continueOneAutomorphism := function(kappa)
    local mu0, zeta_to_kappa, auto_data, y, z, order, mu1, i, j, k, zz, yy, mui;

    zeta_to_kappa := zeta^kappa;
    mui := XEltMove(sb.unramified[offset][3], XEltOrder(mu))^kappa;
    mu0 := EltAutomorphism(mu, zeta_to_kappa)*mui;

    mu0 := EltRoot2(XEltMove(mu0, OrderMaximal(OrderAbs(XEltOrder(mu)))), degs,
                                                                 sb.RG);
#    mu0 := XEltMove(mu0, OrderMaximal(OrderAbs(XEltOrder(mu))));
#    mu0 := Factor(x^degs-mu0); Print(mu0, "\n"); FLDin();
#    mu0 := PolyToList(mu0[1][1])[2];
                                   
    if mu0=false then
      Error("Field was not abelian");
    fi;

    mu0 := XEltMove(mu0, XEltOrder(mu));

    if XOrderDeg(O)>kappa then
      mu1 := List([1..XOrderDeg(O)], i->0);
      mu1[kappa+1] := mu0;
      mu1 := Elt(O, mu1);
    else
      mu1  := mu0*OrderBasis(OrderEquationOrder(O), O)[2]^kappa;
#      Error("this cetainly should never happen eccept for p=8");
    fi;


    zeta_to_kappa := Position(
            OrderAutomorphisms(OrderEquationOrder(XEltOrder(zeta_to_kappa)),[]),
            XEltMove(zeta_to_kappa,
                                 OrderEquationOrder(XEltOrder(zeta_to_kappa))));

    # get the order of the continued Automorphism:
    auto_data := [zeta_to_kappa, mu1];

    z := OrderBasis(OrderEquationOrder(O), O)[2];

    y := z;
    yy := zeta;
    order := 0;
    repeat
      y := auto1(y, auto_data);
      yy := auto1(yy, auto_data);
      order := order+1;
    until yy = zeta;
    yy := order;
    zz := Elt(XOrderCoefOrder(O), 1);
    if y<>z then
      Info("2nd loop\n");
      Info("y: ", y, "\nz: ", z, "\n", "order: ", order, "\n");
      mu1 := EltToList(y)[2];
      zz := mu1;
      repeat
        mu1 := mu1*zz;
        order := order+yy;
      until mu1=1;
    fi;
    Info("Order: ", order, "\n");
    GASMAN("g");

    return rec(data := Copy(auto_data), ord := order, fxn := auto1,
                              order_o := yy, order_zeta := zz);
  end;
  
  automorphism_find := function(a)
    return Position(List(auto.grp, j->Position(j[3], a)<>false), true);
  end;

  artin_prime := [];
  artin_prime_calc := function(p)
    local i, j, k, l, phi_mu, phi_zeta, curr_mu, curr_zeta,
          test_fxn, mc, is_OK, tmp, mc_new, next, mc_tmp, pe,
          mc_bound;

    Unbind(tmp);
    for i in artin_prime do
      if i[1]=p then
        if IsBound(tmp) then
          Error("to many auto's found");
        fi;
        tmp := Copy(i[2]);
      fi;
    od;
    if IsBound(tmp) then return tmp; fi;

    pe := XIdealMin(p)^XIdealRamIndex(p);
    phi_mu := List([1..degs], i->0);

    j := EltPowerMod(mu, IntQuo(XIdealNorm(p),degs), pe);
    phi_mu[(XIdealNorm(p) mod degs) +1] := j;
    phi_mu := Elt(O, phi_mu);

    phi_zeta :=  zeta^(XIdealNorm(p) mod deg);

    is_OK := function(a, b)
      if EltDen(a)>1 then
        a := a*EltDen(a)*(1/EltDen(a) mod pe);
      fi;
      if EltDen(b)>1 then
        b := b*EltDen(b)*(1/EltDen(b) mod pe);
      fi;
      a := a-b;
      a := EltToList(a);
      if XOrderCoefOrder(XEltOrder(a[1]))<>Z then
        a := List(a, EltToList);
        a := Flat(a);
      fi;
      if IsRec(p) and IsBound(p.Type) and p.Type = "ZIdeal" then
        while (not(XEltOrder(a[1]) = XIdealOrder(p))) do
          a := List(a, EltToList);
          a := Flat(a);
        od;
      fi;

      return ForAll(a, i->XEltIsInIdeal(i, p));
    end;

    Unbind(tmp);
    for i in auto.grp do
      if is_OK(i[2], phi_zeta) and is_OK(i[1], phi_mu) then
        Add(artin_prime, [p, i[3][1]]);
        if IsBound(tmp) then
          Error("to many auto's found");
        fi;
        tmp := Copy(i[3][1]);
      fi;
    od;
    if IsBound(tmp) then return tmp; fi;
    if IsBound(auto.finish) then
      Error("Auto not found");
    fi;
    auto.grp[1] := [auto.mu, auto.zeta, [Copy(auto.mc)]];

    for i in [1..Length(auto.Auto)] do
      j := 1;

      curr_mu   := auto.mu;
      curr_zeta := auto.zeta;

      repeat
        curr_mu := auto.Auto[i].fxn(curr_mu, auto.Auto[i].data);
        curr_zeta := auto.Auto[i].fxn(curr_zeta, auto.Auto[i].data);
        k := 1;
        while k<=Length(auto.grp) and
           (auto.grp[k][1] <> curr_mu or auto.grp[k][2] <> curr_zeta) do
          k := k+1;
        od;
        if k<=Length(auto.grp) then
          # found a relation!!
          mc := Copy(auto.grp[k][3][1]);
	  if mc[i]<>0 then Error("asa"); fi;
          mc[i] := j;
          auto.Auto[i].order_o := mc;
          j := 0;
        else
          # new element, so add the full coset...
          for k in auto.grp do
            if k[3][1][i]=j-1 then
              phi_mu := auto.Auto[i].fxn(k[1], auto.Auto[i].data);
              phi_zeta := auto.Auto[i].fxn(k[2], auto.Auto[i].data);
              mc := Copy(k[3][1]);
              mc[i] := j;
              Add(auto.grp, [phi_mu, phi_zeta, [mc]]);
            fi;
          od;
          j := j+1;
        fi;
      until j=0;
    od;
    # now the table ...
    mc       := MultiCounterInit(Length(auto.Auto));
    mc_bound := List([1..Length(auto.Auto)], i->auto.Auto[i].ord-1);
    while MultiCounterInc(mc, mc_bound) do
      if automorphism_find(mc) = false then
        j := Copy(mc);
        for i in Reversed([2..Length(auto.Auto)]) do
          for k in [1..i-1] do
            j[k] := j[k] +
                  auto.Auto[i].order_o[k]*IntQuo(j[i],
                                                auto.Auto[i].order_o[i]);
          od;
          j[i] := j[i] mod auto.Auto[i].order_o[i];
        od;
        j[1] := j[1] mod auto.Auto[1].order_o[1];
        Add(auto.grp[automorphism_find(j)][3], Copy(mc));
      fi;
    od;


    auto.finish := 1;
    return artin_prime_calc(p);
  end;

  artin_ideal := [];
  artin_ideal_calc := function(id)
    local i, j, k, f;

    for i in artin_ideal do
      if i[1] = id then
        return Copy(i[2]);
      fi;
    od;

    f := Factor(id);
    j := List([1..Length(auto.Auto)], i->0);
    for i in f do
      j := j + i[2]*artin_prime_calc(i[1]);
    od;
    for i in [1..Length(j)] do
      j[i] := j[i] mod auto.Auto[i].ord;
    od;
    Add(artin_ideal, [id, j]);
    return Copy(j);
  end;

  automorphism_mult := function(a, b)
    local j, k, m;
    k := auto.grp[a][3][1]+auto.grp[b][3][1];
    for m in [1..Length(auto.Auto)] do
      k[m] := k[m] mod auto.Auto[m].ord;
    od;
    k := MatToRowList(k)[1];
    return Position(List(auto.grp, j->Position(j[3], k)<>false), true);
  end;

  automorphism_apply := function(i, elt) # change
    local j, k, l, m, n;
    k := auto.grp[i][3][1];
    for l in [1..Length(k)] do
      for j in [1..k[l]] do
        elt := auto.Auto[l].fxn(elt, auto.Auto[l].data);
      od;
    od;
    return elt;
  end;

  i := KASHLEVEL("factor");
  KASHLEVEL("factor", 0);
  O := Order(sb.Or, sb.unramified[offset][2],
                                    XEltMove(sb.unramified[offset][1], sb.Or));
  KASHLEVEL("factor", i);
  o := XOrderCoefOrder(sb.Or);

  degs := XOrderDeg(O);
  mu  := -PolyToList(OrderPoly(O))[degs+1];
  ox := PolyAlg(XOrderCoefOrder(O));

  if deg=1 then return 1; fi;
  if deg=2 or (o=Z and XIdealOrder(sb.rg)<>Z) then
    if o=Z then
      i := XOrderTorsionUnit(sb.Or)^(XOrderTorsionUnitRank(sb.Or)/deg);
    else
      Error("");
    fi;
    zeta := i;
    j := rec(gens := [sb.unramified[offset][1]],
             zeta := zeta,
             deg := sb.unramified[offset][2]);
    new := RayGensVar(sb.rg, sb.inf,  
                    sb.unramified[offset][2]*XDisc(sb.field), 0, 
                    i->IdealRayClassRep(i, sb.rg, sb.inf), 1000);
    for i in [1..Length(new[1])] do
      a := KummerArtinPrimeIdeal(j, new[1][i]);
      a := a[1][1];
      new[1][i] := [new[1][i], a];
    od;


    return rec(base := OrderPoly(O), 
                   cf := [, [x-OrderBasis(O, O)[2],
                               x-zeta * OrderBasis(O, O)[2]],
                             sb.rg, new, sb, O, offset],
                   operations := rec(Print := function(a) Print(a.base); end));
  fi;


  deg_o := XOrderDeg(XOrderCoefOrder(O));
  zeta  := OrderBasis(OrderEquationOrder(XOrderCoefOrder(O)),
                                   XOrderCoefOrder(O))[2];

  i := GetGens(XOrderCoefOrder(O));  # Improve: if Length(Kappa)=1...
  Info("I need to extend the following auto's: ", i, "\n");

  auto := [];
  Add(auto, rec(data := [zeta^(deg/degs)],
                               ord := degs, fxn := auto2));

  for j in [1..Length(i)] do
    k := 1;
    while OrderAutomorphisms(XOrderCoefOrder(O), [])[i[j]]<> zeta^k do
      k := k+1;
    od;
    Add(auto, continueOneAutomorphism(k));
  od;


  auto := rec(Auto := auto, mc := List([1..Length(auto)], i->0),
              grp := [],
              mu := OrderBasis(OrderEquationOrder(O), O)[2],
              zeta := zeta);


  ray := [XIdealLcm(XIdeal(o, sb.unramified[offset][2]*4),
             sb.c[3](XIdealMove(sb.RG, sb.Or))),
                             List([1..XOrderSig(o)[1]], j->j)];

  l := [];
  Info("RayModule: ", ray[1], ray[2], "\n");
#  Info("RayGroup :", RayClassGroup(ray[1], ray[2]), "\n");
#  gens := RayGens(ray[1], ray[2], sb.unramified[offset][2]*XDisc(o)* #THINK
#                       sb.c[3](XEltMove(sb.unramified[offset][1], sb.Or)), 0);
  gens := RayGensVar(ray[1], ray[2], sb.unramified[offset][2]*XDisc(o), #THINK
                       0,
                      i->Mat(Z, [artin_prime_calc(i)]), 1000);
  Info("Gens:\n", gens, "\n");

  for j in gens[1] do
    Add(l, IdealRayClassRep(j, sb.rg, sb.inf));
  od;

  if CFT<> [] then
    Append(l, CFT[offset]);
    IsMat(l);
    l := MatKernel(l, deg);  ########################
    l := MatToRowList(l);
    for i in [1..Length(l)] do
      l[i] := l[i]{[1..Length(gens[1])]};
      if ForAll(l[i], j->j=0) then
        Unbind(l[i]);
      fi;
    od;
    l := Set(l);
    IsMat(l);
  else
    t :=  RayClassGroup(sb.rg, sb.inf)[2];
    ineu := 0;
    for i in [1..Length(t)] do
      if t[i] mod deg = 0 and (not exact or IntGcd(t[i]/deg, deg)=1) then
        ineu := ineu +1;
      fi;
      if offset = ineu and t[i] mod deg = 0 and
        (not exact or IntGcd(t[i]/deg, deg)=1) then
        t[i] := deg;
      else
        t[i] := 0;
        for j in l do j[1][i] := 0; od;
      fi;
    od;

    if not IsMat(l) then Error("s.th. went wrong"); fi;
    l := MatKernel(l, deg);
    if not IsMat(l) then Error("s.th. went wrong"); fi;
  fi;
  t :=  gens;
  Append(l, MatId(Z, Length(gens[1]))*deg);
  IsMat(l);

  j := [];
  for k in [1..MatRows(l)] do
    m := List([1..Length(auto.Auto)], i->0);
    for i in [1..Length(t[1])] do
      m := m+artin_prime_calc(gens[1][i]) *l[k][i];
    od;
    for i in [1..Length(auto.Auto)] do
      m[i] := m[i] mod auto.Auto[i].ord;
    od;
    Add(j, Copy(m));
  od;

  Info("j: ", Length(j), "\tafter reduction: ");
  for i in [1..Length(j)] do
    j[i] := Copy(auto.grp[automorphism_find(j[i])][3][1]);
  od;
  j := Set(j);
  Info(Length(j), "\n");



  n := [];
  for i in j do
    l := 1;
    for k in [1..Length(i)] do
      l := IntLcm(l, auto.Auto[k].ord/IntGcd(auto.Auto[k].ord, i[k]));
    od;
    Add(n, l-1);
  od;
  Info("Expo bound:", n, "\n");
  l := MultiCounterInit(Length(n));
  IsMat(j);
  i := [];
  while MultiCounterInc(l, n) do
    IsMat(l);
    k := l*j;
    for m in [1..Length(auto.Auto)] do
      k[1][m] := k[1][m] mod auto.Auto[m].ord;
    od;
    k := MatToRowList(k)[1];
    Add(i, Position(List(auto.grp, j->Position(j[3], k)<>false), true));
    if Length(i)>100 then
      Info("Li: ", Length(i), "\tafter reduction: ");
      i := Set(i);
      Info(Length(i), "\n");
      if Length(i)*deg*XOrderDeg(o)>= OrderDegAbs(O) then
        Info("Leaving...(", Length(i), ", ", deg, ", ",
                                  XOrderDeg(o), ", ", OrderDegAbs(O), ")\n");
        n := -1;
      fi;
    fi;
    l := l[1];
  od;

  Info("Li: ", Length(i), "\tafter reduction: ");
  i := Set(i);
  Info(Length(i), "\n", "set: ", i, "\n");
  if Length(i)*deg*XOrderDeg(o)<> OrderDegAbs(O) then
    Error("rel. minpoly of wrong degree");
  fi;

  factor_g := [1];
  rest_g   := List([1..Length(auto.grp)]);
  SetSubtract(rest_g, i);
  while Length(rest_g)>0 do
    j := rest_g[1];
    for k in i do
      SetSubtract(rest_g, [automorphism_mult(j, k)]);
    od;
    Add(factor_g, j);
  od;

  #get Generating element: the factor group is cyclic of order deg!!
  rest_g := i;
  j := 1;
  while j<Length(factor_g) do
    j := j+1;
    k := 1;
    l := factor_g[j];
    repeat
      k := k+1;
      l := automorphism_mult(l, factor_g[j]);
    until Position(rest_g, l)<>false;
    if k = Length(factor_g) then
      l := factor_g[2];
      factor_g[2] := factor_g[j];
      factor_g[j] := l;
      j := Length(factor_g)+1;
    fi;
  od;

  Info("factor_g:", factor_g, "\n\n");
  Info("Group is:\n");
  for k in [1..Length(auto.grp)] do
    Info(k, "\t", auto.grp[k][3], "\n");
  od;

  # get images for a set of generators of sb.rg, sb.inf:
  new := RayGens(sb.rg, sb.inf, ray[1]*sb.unramified[offset][2]*XDisc(o), 1000);
  for i in [1..Length(new[1])] do
    a := automorphism_find(artin_prime_calc(new[1][i]));
    if a=1 then
      new[1][i] := [new[1][i], 0];
    else
      b := List(rest_g, i->automorphism_mult(i, a));
      a := factor_g[2];
      j := 1;
      while Position(b, a)=false do
        a := automorphism_mult(a, factor_g[2]);
        j := j+1;
      od;
      new[1][i] := [new[1][i], j];
    fi;
  od;

  m := [];
  i := rest_g;
  for j in i do
    rest_g := [];
    for k in factor_g do
      Add(rest_g, auto.grp[automorphism_mult(j, k)][1]);
    od;
    Add(m, rest_g);
  od;

  repeat
    n := false;
    j := 1;
    while not n and j<= Length(m) do
      k := j+1;
      while not n and k<=Length(m) do
        if m[j][1] = m[k][1] then
          for n in [1..Length(i)] do
            for rest_g in [1..Length(factor_g)] do
              m[n][rest_g] := m[n][rest_g] +
                    auto.grp[automorphism_mult(i[n], factor_g[rest_g])][2];
            od;
          od;
          #Print("LOOP\n");
          n := true;
        fi;
        k := k+1;
      od;
      j := j+1;
    od;
  until n = false;

  rest_g := List(m, i->[i[1], i[2]]);
  m := [Sum(m)];
  GASMAN("g");

  for i in m do
    j := EltSimplify(i[1]);
    if j<>1 then
      Info("Minpoly of: ", j, "\n");
      if not IsPrime(deg) then  # Cohen: if IsPrime => Trace is Primitive
        i := [];
        for k in factor_g do
          n := automorphism_apply(k, j);
          if Position(i, n)<>false then
            j := Elt(O,0);
          fi;
          Add(i, n);
        od;
        if j=0 and Length(m)=1 then  
          Info("Trace not Primitive, trying Norm\n");
          Add(m, x->[Product(List(rest_g, o->o[1])), 
                     Product(List(rest_g, o->o[2]))]);
        fi;
        if j=0 and Length(m)>1 then
          Error("need more coefficients, computation stopped");
        fi;
      fi;
      if j<>0 then
        Info("..automorphism finished..\n");
        j := Newton3(j, deg, o);
        sb.den_max := j[2];
        j := j[1];
        Info("..is : ", j, "\n");
        j := PolyToList(j);
        Apply(j, EltSimplify);
        Apply(j, k->XEltMove(k, o));
        return rec(base := Poly(PolyAlg(o), j),
                   cf := [, [x-i[1], x-i[2]], ray[1],
                            new, sb, , offset],
                   operations := rec(Print := function(a) Print(a.base); end));
      else
        Info("is : ", false, "\n");
      fi;
    fi;
  od;
  Error("no Polynomial found");
end;

RayClassField := function(arg)
  local usage, ord, ideal, inf, deg, sol, cg, i, j, sb, mat, SplitGroup;

  usage := "RayClassField( ord | ideal [, list of infinite places] [, deg] );\n";

  SplitGroup := function(id, inf, m)
    local l, cg, n, M, i, j, k, s, inf;


    M := Copy(m);
    cg := RayClassGroup(id, inf);

    n := MatDiag(Z, cg[2]);
    Append(n, m);
    IsMat(n);
    m := MatHermiteRowMod(cg[1], n);
    m := m{[1..Length(cg[2])]};
    IsMat(m);

    if IsPrime(MatDet(m)) then
      return [M];
    fi;

    l := MatSmithTrans(m);
    l[3] := l[3]^-1;
    l[4] := l[4]^-1;
    s := [];
    for i in [1..MatCols(l[2])] do
      if l[2][i][i]<>1 then
        m := Factor(l[2][i][i]);
        for j in m do
          M := MatDiag(Z, List(cg[2], k->1));
          M[i][i] := j[1]^j[2];
          Add(s, l[3]*M*l[4]);
        od;
      fi;
    od;

    return s;
  end;


  deg := 0;
  inf := [];

  if Length(arg)=0 then
    Error(usage);
  elif Length(arg)>0 and not IsOrder(arg[1]) and not XIsIdeal(arg[1]) and
       not IsRecType(arg[1], "AbelianGroup") then
    Error(usage);
  fi;

  if IsRecType(arg[1], "AbelianGroup") then
    ideal := arg[1].Information.RCG.m0;
    inf := arg[1].Information.RCG.minf;
    ord := XIdealOrder(ideal);
    deg := arg[1].Relations;
    if Length(arg)>1 then
      if IsInt(arg[2]) then
        deg := Concatenation(deg, arg[2]*MatId(Z, MatCols(deg)));
        IsMat(deg);
      else
        deg := Concatenation(deg, arg[2]);
        IsMat(deg);
      fi;
    fi;
  else
    if XIsIdeal(arg[1]) then
      ideal := arg[1];
      ord := XIdealOrder(ideal);
    else
      ord := arg[1];
      ideal := 1*ord;
    fi;
    if Length(arg)>1 then
      if not IsInt(arg[2]) and not IsList(arg[2]) then
        Error(usage);
      elif IsInt(arg[2]) or TYPE(arg[2])="KANT matrix" then
        deg := arg[2];
      else
        inf := arg[2];
      fi;
      if Length(arg)>2 and not IsInt(arg[3]) and TYPE(arg[3])<> "KANT matrix"
        or Length(arg)>3 then
        Error(usage);
      elif Length(arg)>2 then
        deg := arg[3];
      fi;
    fi;
  fi;
  cg := RayClassGroup(ideal, inf);

  if not IsInt(deg) then
    mat := deg;
    mat := SplitGroup(ideal, inf, mat);
    sol := function(m)
      local i, j;
      m := Concatenation(m, MatDiag(Z, RayClassGroup(ideal, inf)[2]));
      IsMat(m);
      m := MatHermiteRowMod(cg[1], m);
      j := 1;
      for i in [1..Length(cg[2])] do
        j := j*m[i][i];
      od;
      return j;
    end;
    deg := List(mat, sol);
    SortParallel(deg, mat);
    Info("set of degrees: ", Set(deg), "\n");
    j := [[deg[1], [mat[1]]]];
    for i in [2..Length(mat)] do
      if deg[i] = deg[i-1] then
        Add(j[Length(j)][2], mat[i]);
      else
        Add(j, [deg[i], [mat[i]]]);
      fi;
    od;
    sol := [];
    mat := j;
    for i in mat do
      sb := BaseCreate(ord, i[1], true, ideal, inf, i[2]);
      for j in [1..Length(sb.unramified)] do
        Add(sol, Intersect(sb, i[1], true, j, i[2]));
      od;
    od;
  else
    mat := [];
    if IntGcd(cg[1], deg)=1 then
      return [];
    fi;

    cg[1] := [];
    for i in cg[2] do
      Append(cg[1], Factor(i));
    od;
    cg := Set(cg[1]);
    Info("set of degrees: ", cg, "\n");
    sol := [];

    if deg<>0 then
      sb := BaseCreate(ord, deg, false, ideal, inf, mat);
      for j in [1..Length(sb.unramified)] do
        Add(sol, Intersect(sb, deg, false, j, mat));
      od;
    else
      for i in cg do
        sb := BaseCreate(ord, i[1]^i[2], true, ideal, inf, mat);
        for j in [1..Length(sb.unramified)] do
          Add(sol, Intersect(sb, i[1]^i[2], true, j, mat));
        od;
      od;
    fi;
  fi;
  return sol;
end;
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#
#FXN       RayClassField
#
#SHORTDOC  Given a congruence module, this function computes generating
#          polynomials for the class fields belonging to the cyclic
#          factors of the corresponding ray class group.
#
#LONGDOC   Using the methods of \cite{Fi2}, this function computes
#          generating polynomials for the class fields belonging to the
#          cyclic factors of a decomposition of the corresponding ray class
#          group as into factors of prime power order.
#
#          For each cyclic factor this function will first construct a field
#          containing certain roots of unity. Since this field may have
#          a very large degree, computation may be very lengthy. So
#          be careful if the ray class group has large cyclic factors.
#
#CALLING   c := RayClassField( o | a [, [inf | inf, deg | mat] | deg | mat ] );
#
#ARGUS     c @ list @ containg polynomials of {\tt o[x]}
#          o @ order @ for Hilbert class field computations
#          a @ ideal @ used as (part of) an congruence module as for {\tt RayClassGroup}
#          inf @ list @ of infinite places as for {\tt RayClassGroup}
#          deg @ integer @ if your interesed only in the {\tt deg} part of the class field
#          mat @ matrix @ rows of {\tt mat} define a factorgroup of the RayClassGroup
#
#CSRC
#
#DEBUG     Info := Print;
#
#XREF      RayClassGroup, RayDiscSig, ImQuadRayField, RayConductor, RayConductorTest.
#
#PACKAGE   lib/RCLF
#
#EXAMPLE
#
#KASH      @o := OrderMaximal(Z, 2, 10);
#          c := RayClassField(o);
#          O := OrderMaximal(c[1]);
#          OrderDisc(O);XOrderSig(O);
#          RayClassGroup(4*o);
#          RayClassGroup(4*o, [1]);
#          c:= RayClassField(4*o, [1]);
#          O := OrderMaximal(RelOrderMerge(Order(c[1]), Order(c[2])));
#          XOrderSig(O);
#          RayDiscSig(4*o, [1])[1] = XDisc(O);
#          G := RayClassGroupToAbelianGroup(24*o);
#          RayClassField(G);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY**HISTORY***HISTORY***HISTORY***HISTORY
#
#**C***C***|***C***C***C***C***C***C**C***C***C***C***C***C***C***C***C***C***/

