# int * ZIdeal funktioniert nicht (aber ZIdeal * int ist ok)
# Z * int und int * Z funktioniert nicht (es sollte ein Z Ideal rauskommen)

XIdealIsPrime := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      if IsPrime(arg[1].Generator) then
        return true;
      else
        return false;
      fi;
    fi;    
    return IdealIsPrime(arg[1]);
  else
    Error("usage: IdealIsPrime( <ideal> )");
  fi;
end;

XIdealLcm := function(arg)
local a, b, r;

  if Length(arg)=2 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" and
       IsRec(arg[2]) and arg[2].Type = "ZIdeal" then
        a := arg[1].Generator;      
        b := arg[2].Generator;      
        r := ZIdealCreate(Lcm(a, b));
        return r;
    fi;    
    return IdealLcm(arg[1], arg[2]);
  else
    Error("usage: IdealLcm( <ideal>, <ideal> )");
  fi;
end;

XIdealGcd := function(arg)
local a, b, r;

  if Length(arg)=2 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" and
       IsRec(arg[2]) and arg[2].Type = "ZIdeal" then
        a := arg[1].Generator;      
        b := arg[2].Generator;      
        r := ZIdealCreate(Gcd(a, b));
        return r;
    fi;    
    return Gcd(arg[1], arg[2]);             ### IdealDcd
  else
    Error("usage: IdealGcd( <ideal>, <ideal> )");
  fi;
end;

XIdealMin := function(arg)
local r;

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      r := Abs(arg[1].Generator);
      return r;
    fi;    
    return IdealMin(arg[1]);
  else
    Error("usage: IdealMin( <ideal> )");
  fi;
end;

XIdealOrder := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      return Z;
    fi;    
    return IdealOrder(arg[1]);
  else
    Error("usage: IdealOrder( <ideal> )");
  fi;
end;

XEltIsInIdeal := function(arg)

  if Length(arg)=2 then
    if IsInt(arg[1])  and IsRec(arg[2]) and arg[2].Type = "ZIdeal" then
      return arg[1] mod arg[2].Generator = 0;
    fi;    
    return EltIsInIdeal(arg[1], arg[2]);
  else
    Error("usage: IdealOrder( <elt>, <ideal> )");
  fi;
end;

XIdealRamIndex := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      if XIdealIsPrime(arg[1]) then
        return 1;
      else
        Error("Ideal is required to be prime!");
      fi;
    fi;    
    return IdealRamIndex(arg[1]);
  else
    Error("usage: IdealRamIndex( <ideal> )");
  fi;
end;

XIdealNorm := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      return arg[1].Generator;
    fi;    
    return IdealNorm(arg[1]);
  else
    Error("usage: IdealNorm( <ideal> )");
  fi;
end;

XIdealFactor := function(arg)
local l;

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      l := Factor(arg[1].Generator);
      Apply(l, x->[ZIdealCreate(x[1]), x[2]]);
      return l;
    fi;    
    return IdealFactor(arg[1]);
  else
    Error("usage: IdealFactor( <ideal> )");
  fi;
end;

XIdealBasis := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      return [arg[1].Generator];
    fi;    
    return IdealBasis(arg[1]);
  else
    Error("usage: IdealBasis( <ideal> )");
  fi;
end;

XIdealResidueField := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      if XIdealIsPrime(arg[1]) then
        return FF(arg[1].Generator);
      else
        Error("Ideal is required to be prime!");
      fi;
    fi;    
    return IdealResidueField(arg[1]);
  else
    Error("usage: IdealResidueField( <ideal> )");
  fi;
end;

XIdeal := function(arg)

  if Length(arg)=1 then
    if IsInt(arg[1]) then
      return ZIdealCreate(arg[1]);
    fi;    
    return Ideal(arg[1]);
  elif Length(arg)=2 then
    if TYPE(arg[1])="KANT ring" and IsInt(arg[2]) then
      return  ZIdealCreate(arg[2]);
    fi;
    return Ideal(arg[1], arg[2]);
  else
    if Length(arg)=3 then
      return Ideal(arg[1], arg[2], arg[3]);
    fi;
  fi;
end;

XIsIdeal := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and IsBound(arg[1].Type) and arg[1].Type = "ZIdeal" then
      return true;
    fi;
    return IsIdeal(arg[1]);
  else
    Error("usage: IsIdeal( <object> )");
  fi;
end;

XIdealDegree := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      if XIdealIsPrime(arg[1]) then
        return 1;
      else
        Error("Ideal is required to be prime!");
      fi;
    fi;    
    return IdealDegree(arg[1]);
  else
    Error("usage: IdealDegree( <ideal> )");
  fi;
end;

XIdealIsPrincipal := function(arg)

  if Length(arg)=1 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      return arg[1].Generator;
    fi;
    return IdealIsPrincipal(arg[1]);
  else
    Error("usage: IdealIsPrincipal( <ideal> )");
  fi;
end;

XOrderTorsionUnit := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return -1;
    fi;
    return OrderTorsionUnit(arg[1]);
  else
    Error("usage: OrderTorsionUnit( <order> )");
  fi;
end;

XOrderTorsionUnitRank := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return 2;
    fi;
    return OrderTorsionUnitRank(arg[1]);
  else
    Error("usage: OrderTorsionUnitRank( <order> )");
  fi;
end;

XOrderClassGroup := function(arg)

  if Length(arg)>0 then
    if TYPE(arg[1]) = "KANT ring" then
      return [1, [1]];
    fi;
    if Length(arg)=1 then
      return OrderClassGroup(arg[1]);
    elif Length(arg)=2 then
      return OrderClassGroup(arg[1], arg[2]);
    elif Length(arg)=3 then
      return OrderClassGroup(arg[1], arg[2], arg[3]);
    elif Length(arg)=4 then
      return OrderClassGroup(arg[1], arg[2], arg[3], arg[4]);
    else
      Error("usage: OrderClassGroup( <order> )");
    fi;
  else
    Error("usage: OrderClassGroup( <order> )");
  fi;
end;

XOrderClassGroupCyclicFactors := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return [[XIdeal(1), 1]];
    fi;
    return OrderClassGroupCyclicFactors(arg[1]);
  else
    Error("usage: OrderClassGroupCyclicFactors( <order> )");
  fi;
end;

XIdealMove := function(arg)

  if Length(arg)=2 then
    if IsRec(arg[1]) and arg[1].Type = "ZIdeal" then
      return XIdeal(arg[2], arg[1].Generator);
    else
      return IdealMove(arg[1], arg[2]);        
    fi;
  else 
    Error("usage: IdealMove( <ideal>, <order> )");
  fi;
end;

XOrderSUnits := function(arg)       ### @@@ !!! Aufrufe, ... !!! @@@ ###
local l, T, lb;

  if Length(arg)>1 and Length(arg)< 4 then
    if arg[1]=Z then
      if XIsIdeal(arg[2]) then
        l := XIdealFactor(arg[2]);
        Apply(l, x-> x[1]);
      fi;

      if IsList(arg[2]) then
        l := arg[2];
        lb := List(l, x-> (TYPE(x)="record" and x.Type="ZIdeal"));
        if (false in lb) then 
         Error("second argument has to be a list of prime ideals over correkt ring");
       fi;      
        Apply(l, x-> x.Generator);
        lb := List(l, x-> IsPrime(x));
        if (false in lb) then 
          Error("second argument has to be a list of prime ideals over correkt ring");
        fi;      

        if Length(arg)=3 and arg[3] = "raw" then
          T := MatId(Z, Length(l));
          return [l, T, 1];
        else
          return [l, 1];
        fi;
      else 
        Error("not supported");
      fi;
    else
      if Length(arg) = 1 then 
        return OrderSUnits(arg[1]);
      fi;
      if Length(arg) = 2 then 
        return OrderSUnits(arg[1], arg[2]);
      fi;
      if Length(arg) = 3 then 
        return OrderSUnits(arg[1], arg[2], arg[3]);
      fi;
    fi;
  else
    Error("usage: OrderSUnits( <order> [, < ideal >] [, \"raw\"] )");
  fi;

end;

XDisc := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return 1;
    else
      return Disc(arg[1]);
    fi;
  else 
    Error("usage: Disc( <order> )");
  fi;
end;

XEltToFFE := function(arg)

  if Length(arg)=2 then
    if IsRec(arg[2]) and arg[2].Type = "ZIdeal" then
      return FFElt(XIdealResidueField(arg[2]), arg[1] mod arg[2].Generator);
    else 
      return EltToFFE(arg[1], arg[2]); 
    fi;
  else
    Error("usage: EltToFFE( <elt>, < prime ideal > )");
  fi;
end;

XFFEToElt := function(arg)
local m;

  if Length(arg)=2 then
    if IsRec(arg[2]) and arg[2].Type = "ZIdeal" then
      m := Mat(FFEltFF(arg[1]), [[arg[1]]]);
      m := MatMove(m, Z);
      return m[1][1];
    else
      return FFEToElt(arg[1], arg[2]);
    fi;
  else
    Error("usage: FFEToElt( <elt>, < prime ideal > )");
  fi;
end;

XOrderSig := function(arg)
  
  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return [1, 0];
    else
      return OrderSig(arg[1]);
    fi;
  else 
    Error("usage: OrderSig( <order> )");
  fi;
end;

XEltOrder := function(arg)

  if Length(arg)=1 then
    if IsInt(arg[1]) then
      return Z;
    else
      return EltOrder(arg[1]);
    fi;
  else 
    Error("usage: EltOrder( <elt> )");
  fi;
end;

XOrderCoefOrder := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return Z;
    else
      return OrderCoefOrder(arg[1]);
    fi;
  else 
    Error("usage: OrderCoefOrder( <elt> )");
  fi;
end;

XOrderDeg := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return 1;
    else
      return OrderDeg(arg[1]);
    fi;
  else 
    Error("usage: OrderDeg( <order> )");
  fi;
end;

XOrderDegAbs := OrderDegAbs; #### noch implementieren

XEltMove := function(arg)

  if Length(arg)=2 then
    if TYPE(arg[2]) = "KANT ring" then
      if IsInt(arg[1]) then
        return arg[1];
      else
        Error("not supported");
      fi;
    else
      return EltMove(arg[1], arg[2]);
    fi;
  else 
    Error("usage: EltMove( <algebraic element>, <order> )");
  fi;
end;

XOrderIndex := function(arg)

  if Length(arg)=1 then
    if TYPE(arg[1]) = "KANT ring" then
      return 1;
    fi;    
    return OrderIndex(arg[1]);
  else
    Error("usage: OrderIndex( <order> )");
  fi;
end;
