# Database: use channel 3 for Statistic, and CountQuery
#                       2 Query
# so running CountQuery inside Query doesn't hurt.
#
# globals:
# DbStatisticCreate  can only be run by user postgres
# DbStatistic
# DbQuery
# DbCountQuery
# DbExplainQuery
# DbNextOrder
# DbDeleteOrder      can only be run by user postgres, delete order with
#                    the same polynomial
# DbUpdateOrderGalois  can only be run by user postgres, update galois
#                    information of all orders with the same polynomial
# DbUpdateOrderDisc  can only be run by user postgres, update disc::float8
#                    information of all orders with the same polynomial
# GenericQuery
# GenericNext
# GenericNames
# DbChangeNames
# FindSamePoly
# DbOpen
# DbClose
#
# variables:
# _Db_in_query    value 0, 1, 2
#                    0 initial value or no current query
#                    1 begin was successful
#                    2 declare cf ... was successful
# _Db_has_opened        true, false
#                    true during this kash-session a DbOpen was successful
#                    false inital value, no open
#
#
#
#
if not IsBound(String) then
  ReadLib("string");
fi;

_Db_in_query := KASHLEVEL("KASH_OVERWRITE_INTERNAL_FXNS");
KASHLEVEL("KASH_OVERWRITE_INTERNAL_FXNS", 2);
if not IsBound(Info) then
  Info := Ignore;
fi;
KASHLEVEL("KASH_OVERWRITE_INTERNAL_FXNS", _Db_in_query);

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#BERLIN
#USERDEF
#
#FXN       DbStatisticCreate
#
#SHORTDOC  Creates the statistic data for KNFT0
#
#LONGDOC   Creates the statistic data for KNFT0. This function can only
#          be used by certain authorized users.
#
#CALLING   DbStatisticCreate ();
#
#ARGUS
#
#XREF
#
#PACKAGE   db.kash
#
#KASH      
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbStatisticCreate := function()
  local q, deg, no_r;

  q := "drop table KNFT0_CONT";
  Info(q, "\n");
  if not _DbGenericRetrievalQuery(q, 3) then return false ; fi;

  q := Concatenation("create table KNFT0_CONT (count INT, degree INT, ",
                       "sig_re INT, ",
                       "sig_im INT, min_dsc float8, max_dsc float8, ",
                       "min_reg float4, max_reg float4)");
  Info(q, "\n");
  if not _DbGenericRetrievalQuery(q, 3) then return false; fi;

  for deg in [2..10] do
    if deg mod 2 <> 0 then no_r := 1; else no_r := 0; fi;
    while no_r<=deg do
      q := Concatenation(
          "insert into KNFT0_CONT ",
          "select count(degree), max(degree), max(sig_real), " ,
                "(max(degree) - max(sig_real)) /2,",
                "min(discriminant) ,",
                "max(discriminant) ,",
                "min(regulator)    ,",
                "max(regulator)     ",
        "from KNFT0 ",
        "where degree= ", String(deg),
        " and sig_real = ", String(no_r));

      Info(q, "\n");
      if not _DbGenericRetrievalQuery(q, 3) then return false; fi;
      no_r := no_r+2;
    od;
  od;

  q := "delete from KNFT0_CONT where count=0";
  Info(q, "\n");
  if not _DbGenericRetrievalQuery(q, 3) then return false; fi;

  return true;

end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbStatistic
#
#SHORTDOC  Shows the statistical information of the default database.
#
#LONGDOC   Shows the statistical information of the default database.
#          The shown invariants are degree, signature and the range of
#          discriminant and regulator.
#
#CALLING   DbStatistic()
#
#ARGUS
#
#XREF
#
#PACKAGE   db.kash
#
#KASH      @
#          DbOpen();
#          DbStatistic();
#          DbClose(); # <!IGNORE>
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbStatistic := function()
  local q, n, i, str, sum;

  q := "select * from KNFT0_CONT";
  Info(q, "\n");
  if not _DbGenericRetrievalQuery(q, 3) then return false; fi;

  n := _DbNumberOfColumns(3);
  if n=false then return false; fi;
  n := n-1;

  for i in [0..n] do
    Print(_DbGetColumnName(i, 3), "\t");
  od;
  Print("\n");

  sum := 0;
  str := _fmt();

  while _DbNextRecordFromChannel(3) do
    sum := sum + _DbGetColumnValue(0, 3);
    for i in [0..n] do
      if i=4 or i=5 then
        _fmt("%12R");
        Print(Round(_DbGetColumnValue(i, 3))*1.0);
      elif i=6 or i=7 then
        _fmt("%10.3R");
        Print(_DbGetColumnValue(i, 3));
      else
        Print(_DbGetColumnValue(i, 3), "\t");
      fi;
    od;
    Print("\n");
  od;
  _fmt(str);
  Print("Total of ", sum, " fields.\n");
end;

LocateSubString := function(big, small, start)
  local i, j, s;

  big := big{[start..Length(big)]};

  if big=small then return 1; fi;
  if Length(big)<=Length(small) then return false; fi;

  for i in [1..Length(big) - Length(small)+1] do
    s :=big{[i..i+Length(small)-1]};
    if s = small then
      return i+start-1;
    fi;
  od;
  return false;
end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbChangeNames
#
#SHORTDOC  Change the names for database attributes to the internal
#          representation.
#
#BROADDOC  The names of the entries in the database are --- mostly ---
#          cryptic and short. This function allows a user to use names
#          that are more readable, by changing them into the internal
#          names. Furthermore, by extending this function one is allowed
#          to create own ``nick-names''.\\
#
#          \sloppy\begin{tabular}{p{5cm}|p{7cm}}
#            $[$integral basis flag$]$  & see FLD manual \\
#            $[$number of units$]$      & unit rank \\
#            $[$LFUNIT flag$]$           & see FLD manual  \\
#            $[$number roots of unity$]$  & number of torsion units \\
#            $[$fundamental units flag$]$ & see FLD manual  \\
#            $[$Galois group flag$]$      & see FLD manual \\
#            $[$class number$]$          & \\
#            $[$number of real zeroes$]$  & \\
#            $[$coefficients of defining polynomial$]$  & \\
#            $[$integral basis transformation matrix$]$ & \\
#            $[$coefficients unit system$]$ & \\
#            $[$coefficient torsion generator$]$ & \\
#            $[$KANT name Galois group$]$ & special: in a {\tt Query}, this will
#                        result in a table join to get the number from the name. \\
#            Galgrp & the {\em number} of the Galois group \\
#            is_abelian & special: in a {\tt Query}, this will
#                        result in a table join to get the desired information from
#                        a Galois-group database.             \\
#            is_abel    & see: is_abelian \\
#            is_cyclic  & see: is_abelian \\
#            is_even    & see: is_abelian \\
#            is_odd     & see: is_abelian \\
#            is_primitive & see: is_abelian \\
#            is_solvable  & see: is_abelian \\
#          \end{tabular}
#          \newpage
#          \sloppy \begin{tabular}{p{5cm}|p{7cm}}
#            gal_order & order of the Galois group, see: is_abelian \\
#            $[$number cyclic factors in class group$]$ & \\
#            $[$list of orders of cyclic factors in class group$]$ & \\
#            $[$number necessary prime ideals for class group description$]$ &
#               query for the number of prime ideals in our representation of
#               the class group. \\
#            $[$class group prime ideal description$]$ & \\
#            $[$class group exponent vectors$]$        & \\
#            $[$minimal polynomial (expression in x)$]$  & \\
#            $[$minimal polynomial$]$                    & \\
#            polynomial                              & \\
#            $[$signature real$]$                        & \\
#            $[$signature imaginary$]$                   & \\
#            $[$decimal length of discriminant$]$        & \\
#            $[$discriminant long ascii$]$               & discriminant as ascii string\\
#            $[$regulator long ascii$]$                  & regulator as ascii string\\
#            $[$discriminant known$]$                    & \\
#            disc                                    & \\
#            $[$regulator known$]$                       & \\
#          \end{tabular}
#
#CALLING   outs := DbChangeNames(ins);
#
#ARGUS     outs @ string
#          ins  @ string
#
#XREF
#
#PACKAGE   db.kash
#
#KASH      @
#          DbChangeNames("degree = 3 and [number of real zeroes]=1");
#          DbChangeNames("disc>0 and disc<100 and degree=4");
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbChangeNames :=function(ins)
  local i, j, change;
  change := [
[  "FLDF2",  "[integral basis flag]" ],
[  "FLDF3",  "[number of units]" ],
[  "FLDF4",  "[LFUNIT flag]" ],
[  "FLDF5",  "[number roots of unity]" ],
[  "FLDF6",  "[fundamental units flag]"],
[  "FLDF7",  "[Galois group flag]"],
[  "class_number",  "[class number]"],
[  "sig_real",  "[number of real zeroes]"],
[  "poly_coef", "[coefficients of defining polynomial]"],
[  "ib_matrix", "[integral basis transformation matrix]"],
[  "FLDF12", "[coefficients unit system]"],
[  "FLDF13", "[coefficient torsion generator]"],
[  "*GAL1*", "[KANT name Galois group]"],
[  "*GAL1*", "Galgrp"],
[  "*GAL2*", "is_abelian"],
[  "*GAL2*", "is_abel"],
[  "*GAL3*", "is_cyclic"],
[  "*GAL4*", "is_even"],
[  "*GAL5*", "is_odd"],
[  "*GAL6*", "is_primitive"],
[  "*GAL7*", "is_solvable"],
[  "*GAL8*", "gal_order"],
[  "FLDF15", "[number cyclic factors in class group]"],
[  "FLDF16", "[list of orders of cyclic factors in class group]"],
[  "FLDF17", "[number necessary prime ideals for class group description]"],
[  "FLDF18", "[class group prime ideal description]"],
[  "FLDF19", "[class group exponent vectors]"],
[  "FLDF20", "[minimal polynomial (expression in x)]"],
[  "FLDF20", "[minimal polynomial]"],
[  "FLDF20", "polynomial"],
[  "sig_real", "[signature real]"],
[  "sig_im", "[signature imaginary]"],
[  "disc_dec_len", "[decimal length of discriminant]"],
[  "disc_ascii", "[discriminant long ascii]"],
[  "reg_ascii", "[regulator long ascii]"],
[  "disc_known", "[discriminant known]"],
[  "discriminant", "disc"],
[  "reg_known", "[regulator known]"]];

  for i in change do
    j := 1;
    while j<>false and j<Length(ins) do
      j:= LocateSubString(ins, i[2], j);
      if j<>false then
        ins := Concatenation(ins{[1..j-1]}, i[1],
                     ins{[j+Length(i[2])..Length(ins)]});
        j := j+Length(i[1]) -1;
        IsString(ins);
      fi;
    od;
  od;
  return ins;
end;

_Db_create_where := function(ins)
  local qs, ins, sort, i, add_galinfo, where;

  ins := DbChangeNames(ins);
  add_galinfo := false;
  ins := Concatenation(" where ", ins);
  qs := "declare cf cursor for select * from KNFT0 ";

  i := LocateSubString(ins, "*GAL1*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         "name",
                         ins{[i+Length("*GAL1*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL2*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " is_abelean ",
                         ins{[i+Length("*GAL2*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL3*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " is_cyclic ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL4*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " is_even ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL5*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " not is_even ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL6*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " is_primitive ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL7*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " is_solvable ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  i := LocateSubString(ins, "*GAL8*", 1);
  if i<>false then
    add_galinfo := true;
    ins := Concatenation(ins{[1..i-1]},
                         " ord ",
                         ins{[i+Length("*GAL3*")..Length(ins)]});
  fi;

  if add_galinfo then
    qs := Concatenation(", GALINFO ", ins,
                     " and galois_grno=galno ");
  else
    qs := ins;
  fi;

  return qs;
end;

_Db_in_query := 0;
_Db_startQuery := function()
  if _Db_in_query=2 then
    Info("close cf", "\n");
    if not _DbGenericRetrievalQuery("close cf", 2) then return false; fi;
    _Db_in_query :=1;
  fi;
  if _Db_in_query=1 then
    Info("end", "\n");
    if not _DbGenericRetrievalQuery("end", 2) then return false; fi;
    _Db_in_query := 0;
  fi;

  Info("begin", "\n");
  if not _DbGenericRetrievalQuery("begin", 2) then return false; fi;
  _Db_in_query := 1;
  return true;
end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbQuery
#
#SHORTDOC  Performs a query.
#
#LONGDOC   Firstly this function will convert the strings using
#          {\tt DbChangeNames}. Afterwards the query will be sent to
#          the database, so subsequent calls to {\tt DbNextOrder}
#          may be used to get those orders.
#
#          Note that this function will return {\tt true} iff
#          the query is syntactically correct, regardless whether
#          matching orders exist or not.
#          The actual search starts when performing {\tt DbNextOrder}.
#
#          When present, the orders will be returned in the order
#          described by {\tt sort}, if not the database will
#          return them in arbitrary order.
#
#          If you are interested in the number of orders matching your
#          query, use {\tt DbCountQuery} with the same {\tt where} string.
#
#          As {\tt sort} string you can use e.g. "degree", "disc", 
#          "regulator", "sig_im", "sig_real" or "[class number]".
#
#          For further information about design and implementation
#          see \cite{DaWe1}.
#
#CALLING   DbQuery(where); or DbQuery(where, sort);
#
#ARGUS     where @ string
#          sort  @ string
#
#XREF
#
#PACKAGE   db.kash
#
#KASH      We will give a number of examples here, starting with simple
#          ones: All queries, except the first expect that the database
#          is already opened.
#          First we are interested in arbitrary cubic fields: @
#          DbOpen();
#          DbQuery("degree=3");
#          DbNextOrder();
#          DbNextOrder();
#          DbClose(); # <!IGNORE>
#          @ Now we will restrict the signature to totally real fields: @
#          DbOpen(); # <!IGNORE>
#          DbQuery("degree=3 and [number of real zeroes]=3");
#          o := DbNextOrder();
#          OrderSig(o);
#          DbClose(); # <!IGNORE>
#          @ Quintics with non-trivial class group: @
#          DbOpen(); # <!IGNORE>
#          DbQuery("degree=5 and [class number]>1");
#          DbNextOrder();
#          DbClose(); # <!IGNORE>
#          @ Degree $>2$ and 3 divides the class number:@
#          DbOpen(); # <!IGNORE>
#          DbQuery("degree >2 and [class number] % 3 =0");
#          DbNextOrder();
#          DbClose(); # <!IGNORE>
#          @Fields with prescribed discriminant: @
#          DbOpen(); # <!IGNORE>
#          DbQuery("disc > 1000 and disc < 1040");
#          DbNextOrder();
#          DbClose(); # <!IGNORE>
#          @Fields with prescribed Galois group: @
#          DbOpen(); # <!IGNORE>
#          DbQuery("Galgrp='D4'");
#          DbClose(); # <!IGNORE>
#          @Abelian fields with non-trivial class group:@
#          DbOpen(); # <!IGNORE>
#          DbQuery("is_abelian and [class number]>1");
#          DbClose(); # <!IGNORE>
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbQuery := function(arg)
  local qs, ins, sort, i, add_galinfo, where;

  ins := arg[1];
  if not IsString(ins) then
    Error("usage: DbQuery( \"select\" [, \"sort\" ] );");
  fi;

  if Length(arg)=2 then
    sort := Concatenation(" order by ", DbChangeNames(arg[2]));
    if LocateSubString(sort, "*GAL1*", 1)<>false then
      return Error("sorry cannot do this");
    fi;
  else
    sort := "";
  fi;
  if not _Db_startQuery() then return false; fi;
  ins := _Db_create_where(ins);

  qs := "declare cf cursor for select * from KNFT0 ";

  qs := Concatenation(qs, ins, sort);

  Info(qs, "\n");
  if not _DbGenericRetrievalQuery(qs, 2) then return false; fi;
  _Db_in_query := 2;
  return true;
end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbCountQuery
#
#SHORTDOC  Simply returns the number of matching orders.
#
#LONGDOC
#
#CALLING   n := DbCountQuery(where);
#
#ARGUS     n @ int
#          where @ string
#
#XREF      DbQuery
#
#PACKAGE
#
#KASH      All {\tt where} string from {\tt Query} are valid, note that
#          the actual numbers may vary from the number given here,
#          if we extend the database. @
#          DbOpen();
#          DbCountQuery("degree=4 and [number of real zeroes]=2");
#          DbClose(); # <!IGNORE>
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbCountQuery := function(ins)
  local qs, ins, sort, i;

  ins := _Db_create_where(ins);

  qs := "select count(degree) from KNFT0 ";

  qs := Concatenation(qs, ins);

  Info(qs, "\n");
  if not _DbGenericRetrievalQuery(qs, 3) then return false; fi;
  if not _DbNextRecordFromChannel(3) then return false; fi;
  return _DbGetColumnValue(0,3);
end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbExplainQuery
#
#SHORTDOC  Returns internal database information.
#
#LONGDOC   Firstly this function will convert the strings using
#          {\tt DbChangeNames}. Afterwards the function will
#          return the information. The Syntax is the same as in
#          {\tt DbQuery}.
#
#          As {\tt sort} string you can use e.g. "degree", "disc", 
#          "regulator", "sig_im", "sig_real" or "[class number]".
#
#          For further information about design and implementation
#          see \cite{DaWe1}.
#
#CALLING   DbExplainQuery(where); or DbExplainQuery(where, sort);
#
#ARGUS     where @ string
#          sort  @ string
#
#XREF      DbQuery
#
#PACKAGE   db.kash
#
#KASH      We will give a number of examples here, starting with simple
#          ones: All queries, except the first expect that the database
#          is already opened.
#          First we are interested in arbitrary cubic fields: @
#          DbOpen();
#          DbExplainQuery("degree=3");
#          DbClose(); # <!IGNORE>
#          @Now we are interested in Abelian fields with non-trivial 
#          class group:@
#          DbOpen(); # <!IGNORE>
#          DbExplainQuery("is_abelian and [class number]>1");
#          DbClose(); # <!IGNORE>
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbExplainQuery := function(arg)
  local qs, ins, sort, i, add_galinfo, where;

  ins := arg[1];

  if Length(arg)=2 then
    sort := Concatenation(" order by ", DbChangeNames(arg[2]));
    if LocateSubString(sort, "*GAL1*", 1)<>false then
      return Error("sorry cannot do this");
    fi;
  else
    sort := "";
  fi;
  ins := _Db_create_where(ins);

  qs := "explain select * from KNFT0 ";

  qs := Concatenation(qs, ins, sort);

  Info(qs, "\n");
  if not _DbGenericRetrievalQuery(qs, 3) then return false; fi;

  return true;
end;

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbNextOrder
#
#SHORTDOC  Gets the next order corresponding to the previous query, 
#          performed by {\tt DbQuery}.
#
#LONGDOC   Gets the next order corresponding to the previous query, 
#          performed by {\tt DbQuery}.
#          When present, the orders will be returned in the order
#          described by {\tt sort} of {\tt DbQuery}, if not the database will
#          return them in arbitrary order.
#
#          If there is no further order then {\tt DbNextOrder} returns
#          {\tt false}.
#          If you are interested in the number of orders matching your
#          query, use {\tt DbCountQuery} with the same {\tt where} string.

#CALLING   DbNextOrer();
#
#ARGUS     
#
#XREF      DbQuery, DbCountQuery
#
#PACKAGE   db.kash
#
#KASH      We will give a number of examples here, starting with simple
#          ones: All queries, except the first expect that the database
#          is already opened.
#          First we are interested in arbitrary quadratic fields: @
#          DbOpen();
#          DbQuery("degree=2");
#          @ Now we get the first order: @
#          DbNextOrder();
#          @ Then the second order: @
#          DbNextOrder();
#          DbClose(); # <!IGNORE>
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbNextOrder := function()
  if not _Db_in_query=2 then return Error("no previous Query"); fi;

  if not _DbGenericRetrievalQuery("fetch forward 1 in cf", 2) then
    return false;
  fi;
  return _DbNextOrderFromQuery(2);
end;

DbFindSamePoly := function(o)
  local s;

  if not _Db_startQuery() then return false; fi;

  s := _DbPolyHashOfOrder(o);
  s := Concatenation("declare cf cursor for select * from KNFT0 ",
         "where hashpoly='", s, "' ");

  if not _DbGenericRetrievalQuery(s, 2) then return false; fi;
  _Db_in_query := 2;
  return true;
end;

DbGenericQuery := function(s)
  s := DbChangeNames(s);
  Info(s, "\n");
  return _DbGenericRetrievalQuery(s);
end;

DbGenericNext := function()
  local i, n, l;
  if not _DbNextRecordFromChannel() then return false; fi;
  n := _DbNumberOfColumns()-1;
  l := [];
  for i in [0..n] do
    Add(l, _DbGetColumnValue(i));
  od;
  return l;
end;

DbGenericNames := function()
  local n, i, l;
  n := _DbNumberOfColumns();
  if n=false then return false; fi;
  l := [];
  for i in [0..n-1] do
    Add(l, _DbGetColumnName(i));
  od;
  return l;
end;

_Db_has_opened := false;
_Db_is_open := false;
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbClose
#
#SHORTDOC  Closes the database.
#
#LONGDOC
#
#CALLING   DbClose();
#
#ARGUS     
#
#XREF      DbOpen
#
#PACKAGE
#
#KASH      Open the database: @
#          DbOpen();
#          @ Now we can use the other database functions.
#          Then close the database: @
#          DbClose(); 
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbClose := function()
  if _Db_in_query=2 then
    Info("close cf", "\n");
    if not _DbGenericRetrievalQuery("close cf", 2) then return false; fi;
    _Db_in_query := 1;
  fi;
  if _Db_in_query =1 then
    Info("end", "\n");
    if not _DbGenericRetrievalQuery("end", 2) then return false; fi;
    _Db_in_query := 0;
  fi;
  _Db_has_opened := false;
#  _DbClose();
end;


#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#
#FXN       DbOpen
#
#SHORTDOC  Opens the database.
#
#LONGDOC
#
#CALLING   DbOpen();
#
#ARGUS     
#
#XREF      DbClose
#
#PACKAGE
#
#KASH      Open the database: @
#          DbOpen();
#          @ Now we can use the other database functions.
#          Then close the database: @
#          DbClose();
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
DbOpen := function(arg)
  local rv;
  if _Db_is_open = true then
    _Db_has_opened := true;
    return true;
  fi;
  if Length(arg)>1 or (Length(arg)=1 and not IsString(arg[1])) then
    Error("usage: DbOpen( [ \"database\" ] );");
  fi;
  if Length(arg) = 0 then
    rv := _DbOpen();
  else
    rv := _DbOpen(arg[1]);
  fi;

  if rv and not _Db_has_opened then
    Info("select * from INFO", "\n");
    if not _DbGenericRetrievalQuery("select * from INFO") then  return true; fi;
    while _DbNextRecordFromChannel() do
      Print(_DbGetColumnValue(0), "\n");
    od;
    _Db_has_opened := true;
    _Db_is_open := true;
    return true;
  fi;
  if not rv then
    if _DbOpen("primus.math.tu-berlin.de:kantnf_info") then
      Info("select * from INFO", "\n");
      if _DbGenericRetrievalQuery("select * from INFO") then
        while _DbNextRecordFromChannel() do
          Print(_DbGetColumnValue(0), "\n");
        od;
        DbClose();
        return false;
      fi;
    fi;
  else
    return true;
  fi;
  Error("I'm not able to connect to Berlin");
  return false;
end;

DbDeleteOrder := function(o)
  local hash, q;

  hash := _DbPolyHashOfOrder(o);
  q := Concatenation("delete from KNFT0 where hashpoly = '",
                      hash, "'");
  Info(q, "\n");
  return _DbGenericSQLQuery(q);
end;

DbUpdateOrderGalois := function(o)
  local hash, q;

  hash := _DbPolyHashOfOrder(o);
  q := Concatenation("update KNFT0 set galois_grno = ",
                      String(GALOIS(o)),
                      " where hashpoly = '",
                      hash, "'");
  Info(q, "\n");
  return _DbGenericSQLQuery(q);
end;

DbUpdateOrderDisc := function(o)
  local hash, q;

  hash := _DbPolyHashOfOrder(o);
  q := Concatenation("update KNFT0 set discriminant = ",
                      String(OrderDisc(o)),
                      " where hashpoly = '",
                      hash, "'");
  Info(q, "\n");
  return _DbGenericSQLQuery(q);
end;

