ZIdealCreate := Ignore;
#############################################################################
##
#A  kant.g                    KANT library              Mario Daberkow
##
#A  @(#)$Id: kant.g 1.00 1995/02/09 Daberkow $
##
#Y  Copyright 1995,  Prof. Dr. M. Pohst, TU - Berlin,  Germany
##
##  This file contains the general KANT functions .
##
#H  $Log: kant.g $
#H  Revision 1.03  1999/01/22  Mendez
#H  added 'Den Num'
#H
##
#H  $Log: kant.g $
#H  Revision 1.03  1998/09/10  Mendez
#H  added 'CharPoly MinPoly' 'Index'
#H
##
#H  $Log: kant.g $
#H  Revision 1.02  1996/08/17  Hoppe
#H  added 'Basis'
#H
#H  $Log: kant.g $
#H  Revision 1.01  1996/04/17  Hoppe
#H  added 'Min'
#H
#H  $Log: kant.g $
#H  Revision 1.00  1995/02/09  Daberkow
#H  added 'Norm'
#H
##
#H  $Log: kant.g $
#H  Revision 1.00  1995/02/09  Daberkow
#H  added 'Disc'
#H
##
#H  $Log: kant.g $
#H  Revision 1.00  1995/02/09  Daberkow
#H  added 'IsPrime'
#H
##

#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF   
#
#FXN       IsRecType
#
#SHORTDOC  Returns {\tt true} iff the argument is an record with an {\bf Type} enry containing the correct string.
#
#LONGDOC   Useful in User-supplied functions, see the Abelian group package
#          as an example.
#
#CALLING   flag := IsRecType(a, type);
#
#ARGUS     flag @ boolean
#          a @ anything
#          type @ string
#
#CSRC      
#
#DEBUG    
#
#XREF      
#
#PACKAGE   
#
#KASH      @
#          a := rec(Type := "Test", data := 1);
#          IsRecType(a, "Test");
#          IsRecType(a, "Paul");
#          IsRecType(2, "two");
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
IsRecType := function(r, t)
  if IsRec(r) and IsBound(r.Type) and r.Type=t then
    return true;
  else
    return false;
  fi;
end;

IsAlffElt := function(a)
  return TYPE(a) = "KANT function field order elt";
end;

###########################################################################
IsPolyAlg := function (a)
   
   if TYPE( a ) = "KANT polynomial algebra"  then
      return true;
   else
      return false;
   fi;
    
end;

#############################################################################
##
#F  IsPrime ( <object> ) . . . Checks if obj is a prime element.
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#FXN       IsPrime
#
#SHORTDOC  Returns {\tt true} iff the argument is either a
#          prime number or a prime ideal.
#
#CALLING   b := IsPrime(x);
#
#ARGUS     b @ boolean
#          x
#
#CSRC      IdealIsPrime, IntIsPrime
#
#PACKAGE   lib
#
#KASH      First some rational integers @
#          IsPrime(1997);
#          IsPrime(4);
#          @
#          Now some ideals in $\Q(\sqrt3)$ @
#          o := Order(Z, 2, 3);;
#          OrderMaximal(o);;
#          IsPrime(5*o);
#          IsPrime(6*o);
#          IsPrime(Factor(6*o)[1][1]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

IsPrime := function ( a )

  if  (IsIdeal (a)) then
    return IdealIsPrime (a);
  fi;

  if (IsInt (a)) then
       return IntIsPrime (a);
  fi;

  return false;
  
end;


#############################################################################
##
#F  IsThue ( <object> ) . . . Checks if obj is a Thue object.
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#FXN       IsThue
#
#SHORTDOC  Returns {\tt true} iff the argument is a {\sf Thue} object.
#
#CALLING   b := IsThue(x);
#
#ARGUS     b @ boolean
#          x
#
#PACKAGE   lib
#
#KASH      @
#          t := Thue([1,2,2,2,2]);
#          IsThue(t);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

IsThue := function ( a )


  if (TYPE (a) = "KANT Thue equation") then
    return true;
  else
    return false;
  fi;

end;

#############################################################################
##
#F  Solve ( <object>, <object> ) . . . Solves a problem
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#FXN       Solve
#
#SHORTDOC  Solves an equation or computes the roots of a polynomial.
#
#LONGDOC   At the moment two kinds of equations
#          can be solved using {\tt Solve}. Furthermore it is possible to
#          compute the roots of a polynomial.
#
#BROADDOC  \begin{itemize}
#          \item[{\tt Solve(t,A)}] Solves a Thue equation. The
#            command {\tt Solve(t,A)} is tantamount
#            to the calling sequence {\tt ThueSolve(t,A)}.
#          \item[{\tt Solve(o,a)}] Solves a norm equation.
#            The command {\tt Solve(o,a)} is tantamount
#            to the calling sequence {\tt OrderNormEquation(o,a)}.
#          \item[{\tt Solve(f)}] Computes the roots of a polynomial.
#          \end{itemize}
#
#CALLING   L := Solve(t, A);\\
#          L := Solve(o, a);\\
#          L := Solve(f);
#
#ARGUS     L     @ list
#          t     @ Thue object
#          A     @ int
#          o     @ order
#          a     @ int | algebraic element
#          f     @ polynomial
#
#CSRC      Norm\_equation, thue\_solve
#
#PACKAGE   lib
#
#XREF      OrderNormEquation, ThueSolve, PolyZeros
#
#KASH      Compute all $x,y \in \Z$ with
#          $x^3 + x^2 y - 6 x y^2 + 2 y^3 = 2$:@
#          t := Thue( [1, 1, -6, 2] );
#          Solve(t, 2);
#
#KASH      Compute an element $\alpha \in \Z[\sqrt[3]{5}]$ with
#          norm $\pm 2$. @
#          o := Order(Z, 3, 5);
#          a := Solve(o, 2);
#          Norm(a[1]);
#
#KASH      Computes the roots of a polynomial:@
#          f := Poly(Zx, [1,0,-4,0,1]);
#          Solve(f);

#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

Solve := function ( arg )

  if ( IsThue (arg[1]) ) then
    if ( IsInt (arg[2]) ) then
      return ThueSolve (arg[1],arg[2]);
    else
      Print("Second argument has to be an int.\n");
    fi;

  elif ( IsOrder (arg[1]) ) then
    if ( IsElt (arg[2]) or IsInt(arg[2]) ) then
       return OrderNormEquation(arg[1],arg[2]);
    else
      Print("Second argument has wrong type.\n");
    fi;

  elif ( IsPoly (arg[1]) ) then
    return PolyZeros (arg[1] );

  else
    Print ("Type of arguments is not supported\n");
  fi;

end;

#############################################################################
##
#F  Norm ( <obj> ) . . . Computes the norm of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Norm
#CALLING   n := Norm(a);\\
#          n := Norm(b, o);
#SHORTDOC  Computes the norm of an algebraic number, an ideal, a polynomial
#          or an alff element or divisor.
#LONGDOC   Depending on the type of {\tt a} or {\tt b} the norm of an algebraic
#          element, the norm of an ideal, polynomial or an alff element or
#          divisor
#          is computed.
#ARGUS     n @ int $|$ rational $|$ algebraic number $|$ polynomial
#          a @ algebraic element $|$ ideal $|$ polynomial $|$ alff element $|$ ff element
#          b @ algebraic element $|$ ff element
#          o @ order
#CSRC      IdealNorm, EltNorm, PolyNorm, AlffEltNorm, FFEltNorm
#
#PACKAGE   lib
#
#KASH      This examples take place in ${\Bbb Q}(\sqrt5)$: @
#          o := Order(Z, 2, 5);;
#          Norm(Elt(o, [1, 2]));
#          u := OrderUnitsFund(o);
#          Norm(u[1]);
#          Norm(2*o);
#          Norm(Ideal(2, Elt(o,[2,3 ])));
#KASH      An example for a polynomial: @
#          ox := PolyAlg(o);;
#          a := Poly(ox, [1, 0, Elt(o, [1, -1]), 3]);
#          Norm(a);
#KASH      An example for an element of a global function field:@
#          AlffInit(FF(7, 2));
#          F:= Alff(y^3+T^2+3);
#          o:= AlffOrderEqFinite(F);
#          a:= AlffElt(o, [T-2, 1, T]);
#          Norm(a);
#         
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

Norm := function ( arg )
   local argc, a;
   
   
   argc:=  Length(arg);
   
   if argc <1 or argc>2 then
      
      return Error("usage: Norm( <order element> | <ff element> [, <ring> ] | <ideal> | <polynomial> | <alff element>");
   fi;
   
   a:= arg[1];
   
   if argc = 1 then
      
      if  (IsIdeal (a)) then
         return IdealNorm (a);
      fi;

      if (IsPoly (a)) then
       return PolyNorm (a);
      fi;

      if IsElt(a) or IsRat(a) then
         return EltNorm (a);
      fi;
      
      if (IsAlffElt(a)) then
         return AlffEltNorm (a);
      fi;

      if (IsAlffIdeal(a)) then
         return AlffIdealNorm (a);
      fi;
      
      if (IsAlffDivisor(a)) then
         return AlffDivisorNorm (a);
      fi;

      if IsFFElt(a) then
         return FFEltNorm(a);
      fi;
      
      if IsRecType(a, "ZIdeal") then
        return a.Generator;
      fi;
      
   else
      
      if IsElt (a) or IsRat(a) then
         return EltNorm(a, arg[2]);
      fi;
      
      if IsFFElt(a) then
         return FFEltNorm(a, arg[2]);
      fi;
      
   fi;      
  
  return Error("usage: Norm( <order element> $|$ <ff element> [, <ring> ] | <ideal> | <polynomial> | <alff element>");
end;



#############################################################################
##
#F  Trace ( <obj> ) . . . Computes the Trace of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Trace
#CALLING   n := Trace(a);\\
#          n := Trace(b, o);
#SHORTDOC  Computes the trace of an algebraic number or alff element or 
#          a matrix 
#LONGDOC   Depending on the type of {\tt a} the trace of an algebraic element,
#          alff element or matrix  is computed.
#ARGUS     a @ algebraic element $|$ alff elt $|$ matrix $|$ ff element
#          b @ algebraic element $|$ ff element
#          o @ order
#
#PACKAGE   lib
#
#CSRC      AlffEltTrace, EltTrace, MatTrace, FFEltTrace
#KASH      All examples take place in ${\Bbb Q}(\sqrt5)$: @
#          o := Order(Z, 2, 5);
#          Trace(Elt(o, [1, 2]));
#KASH      Another example for global function fields: @
#          AlffInit(FF(673, 1));
#          F := Alff(y^3+T^2+1);
#          o := AlffOrderEqFinite(F);
#          a := AlffElt(o, [T-2, 1, T]);
#          Trace(a);
#        
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

  
Trace := function(arg)
   local argc, a, b;
   
   argc := Length(arg);
   
   if argc < 1 or argc > 2 then
      return Error("usage: Trace( <order element> | <ff element> [ , <ring> ] | <matrix> | <alff element>");
   fi;
   
   a := arg[1];
   
   if argc = 1 then
   
      if IsMat(a) then
         return MatTrace(a);
      fi;
   
      if IsElt(a) or IsRat(a) then
         return EltTrace(a);
      fi;
      
      if IsAlffElt(a) then
         return AlffEltTrace(a);
      fi;
      
      if IsFFElt(a) then
         return FFEltTrace(a);
      fi;
      
      if IsRat(a) then
         return a;
      fi;
      
   else
      
      if IsElt(a) or IsRat(a) then
         return EltTrace(a, arg[2]);
      fi;
      
      if IsFFElt(a) then
         return FFEltTrace(a, arg[2]);
      fi;
      
   fi;
      
   return Error("usage: Trace( <order element> | <ff element> [, <ring> ] | <matrix> | <alff element>");

end;


#############################################################################
##
#F  Min ( <obj> ) . . . Computes the minimum of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Min
#CALLING   n := Min(a);
#SHORTDOC  Computes the minimum of an ideal.
#LONGDOC
#ARGUS     n @ int $|$ rational $|$ ideal
#          a @ ideal
#
#PACKAGE   lib
#
#CSRC      IdealMin
#KASH      All examples take place in ${\Bbb Q}(\sqrt5)$: @
#          O := OrderMaximal(Order(x^3 - 10*x^2 - 3*x - 2));;
#          o := Order(O, 3, 3);;
#          IdealMin(Ideal(5, Elt(O, [1, 1, 0])) / 4);
#          IdealMin(Ideal(Elt(O, [3, 1, 0]), Elt(O, [1, 1, 0])));
#          IdealMin(5*o + Elt(o, [3, 1, 0])*o);
#
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

Min := function ( a )

  if  (IsIdeal (a)) then
    return IdealMin (a);
  elif IsRecType(a, "ZIdeal") then
    return a.Generator;
  fi;

  return Error("usage: Min( <ideal> );\n");

end;


Basis := function ( a )

  if  (IsIdeal (a)) then
    return IdealBasis (a);
  fi;
  if (IsOrder(a)) then
    return OrderBasis(a);
  fi;

  return Error("usage: Basis( <order> | <ideal> )");

end;


_EltDisc := function(elt)
  local i, j, n, mat, o, O, tmp, s, nq;

  O := EltOrder(elt);
  o := OrderCoefOrder(O);
  if o=Z then o:=Q; fi;
  n := OrderDeg(O);

  mat := MatToColList(MatId(o, n));

  tmp := Elt(O, 1);
  nq := 2*n-2;
  for s in [0..nq] do
    for i in [Maximum([s-n+1,0])..Minimum([s,n-1])] do
      mat[i+1][s-i+1] := tmp;
    od;
    tmp := tmp*elt;
  od;
  Apply(mat, i->List(i, EltTrace));
  mat := Mat(o, mat);
  if o=Q then
    return MatDet(mat);
  else
    return MatDet(mat);
  fi;
end;


#############################################################################
##
#F  Disc ( <obj> ) . . . Computes the discriminant of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Disc
#CALLING   d := Disc(f);\\
#          d := Disc(o);\\
#          d := Disc(l);
#SHORTDOC  Computes the discriminant of a polynomial, an order, an algebraic 
#          element, an alff order or a lattice.
#ARGUS     d @ ring element
#          f @ polynomial
#          o @ order alff order
#          l @ lattice
#
#PACKAGE   lib
#
#CSRC      PolyDisc, EltDisc OrderDisc AlffOrderDisc LatDisc
#KASH      We will compute some indices: @
#          o := OrderMaximal(Order(x^6 - 9*x^4 - 4*x^3 + 27*x^2 - 36*x - 23));;
#          O := OrderShort(o);;
#          Disc(OrderPoly(Zx, o)) / Disc(o);
#          Disc(OrderPoly(Zx, O)) / Disc(O);
#
#KASH      An example for global function fields: @
#          AlffInit(FF(11, 1));
#          AlffOrders(y^2-y-T^3+T^2);
#          Disc(o);
#         
#KASH      An example for lattices:  @
#          lambda := Lat(7*MatId(Z,8));
#          Disc(lambda); 
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#  98-09-16 extended jm
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##


Disc := function ( a )

  if  (IsPoly (a)) then
    return PolyDisc (a);
  fi;

  if (IsOrder (a)) then
       return OrderDisc (a);
  fi;

  if (IsElt (a)) then
       return _EltDisc (a);
  fi;

  if (IsAlffOrder (a)) then
     return AlffOrderDisc (a);
  fi;
  if (IsLat (a)) then
     return LatDisc (a);
  fi;
  return Error("usage: Disc( <order> | <order elt> | <polynomial> | <alff order>| <lattice> )");

end;


#############################################################################
##
#F  Factor ( <object> ) or Factor ( <object> , <object> ) . . . Returns the factorization.
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#
#USERDEF
#
#FXN       Factor
#
#SHORTDOC  Returns the factorization of the given argument.
#
#BROADDOC   At the moment 4 different objects can be factored:
#          \begin{itemize}
#          \item[{\tt Factor(d)}] Returns the factorization of the integer
#            $d\in{\Bbb N}$. $d$ must be greater $1$.
#          \item[{\tt Factor(f)}] Returns the factorization of the polynomial
#            $f\in{\Bbb Z}[x], {\Bbb Q}[x], o[x], FF[x]$.
#          \item[{\tt Factor(f, p)}] Returns the factorization of the polynomial
#          $f\in{\Bbb Z}[x] \bmod p{\Bbb Z}[x]$.
#          \item[{\tt Factor(a)}]  Let $\O$ be the maximal order where $\a$
#          has been defined within.
#          The function returns the factorization of $\a$ in $\O$.
#          \item[{\tt Factor(alpha)}] Let $\O$ be the maximal order where
#          $\alpha$
#          has been defined within. The function returns the factorization of
#            the principal ideal $\alpha \O$ in $\O$.
#          \item[{\tt Factor(O,d)}] Returns the factorization of the principal
#            ideal $d\O$ in the maximal order $\O$.
#          \item[{\tt Factor(O,alpha)}] Returns the factorization of the
#          principal ideal $\alpha \O$ in the maximal order $\O$.  $\alpha$ must
#            be defined within $\O$. Same as {\tt Factor(alpha)}.
#          \item[{\tt Factor(O,a)}] Returns the factorization of the
#            ideal $\a$ in the maximal order $\O$. $\a$ must be defined
#          within $\O$. Same as {\tt Factor(a)}.
#          \end{itemize}
#          The factorization of the argument $A=u \prod|{i=1}^{k} \pi|i^{e|i}$ with
#          $A\in\{d, f, {\goth{a}}, \alpha \}$, a unit $u$, $k\in{\Bbb N}|0$ and prime elements
#          $\pi|i, 1\leq i \leq k$ is
#          returned in a list of the following form: $[[\pi|1,e|1], \ldots, [\pi|k,e|k]]$.
#
#CALLING   F := Factor(d);\\
#          F := Factor(f);\\
#          F := Factor(f, p);\\
#          F := Factor(a);\\
#          F := Factor(alpha);\\
#          F := Factor(O, d);\\
#          F := Factor(O, a);\\
#          F := Factor(O, alpha);
#
#ARGUS     F @ list
#          p @ prime
#          d @ integer
#          f @ polynomial
#          a @ ideal
#          alpha @ algebraic element
#          O @ order
#
#CSRC      integer\_factorise, vpoly\_fact\_integer, order\_ideal\_factorize,
#          order\_elt\_factorize
#
#XREF      IntFactor, EltFactor, PolyFactor, IdealFactor
#
#PACKAGE   lib
#
#KASH      Factorization of $8274626472648264826427648723648276$:@
#          Factor(8274626472648264826427648723648276);
#
#KASH      Factorization of the polynomial $f(x)=4x^7+6x^6+12x^5+14x^4+27x^3+24x^2+30x+9\in{\Bbb Z}[x]$:@
#          f := Poly(Zx, [4,6,12,14,27,24,30,9]);
#          Factor(f);
#
#KASH      Factorization of the principal ideal $(3)$ in the maximal order of ${\Bbb Q}(\sqrt{-110})$:@
#          o := Order(Z, 2, -110);
#          O := OrderMaximal(o);
#          Factor(O, 3);
#          a := Ideal(O, 3);
#          Factor(a);
#          Factor(O, a);
#
#KASH      Factorization of an algebraic element:@
#          o := Order(Z, 2, -110);
#          O := OrderMaximal(o);
#          alpha := Elt(O, [0,1]);
#          Factor(alpha);
#          Factor(O, alpha);
#
#**HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/

Factor := function(arg)
   local Fp,Fpo,g,s;

   s := "usage:\n Factor( [<order>,] <integer> )\n Factor( [<order>,] <ideal> )\n Factor( [<order,>] <algebraic element> )\n Factor( <polynomial> [, <prime>] )\n ";

   if Length(arg) = 1 then
      if IsInt(arg[1]) then
         return IntFactor(arg[1]);
      elif IsPoly(arg[1]) then
         return PolyFactor(arg[1]);
      elif IsElt(arg[1]) then
         return EltFactor(arg[1]);
      elif IsIdeal(arg[1]) then
         return IdealFactor(arg[1]);
      elif IsAlffIdeal(arg[1]) then
         return AlffIdealFactor(arg[1]);
      elif IsRecType(arg[1], "ZIdeal") then
         return List(Factor(arg[1].Generator), s->[ZIdealCreate(s[1]), s[2]]);
      else
         return Error(s);
      fi;
   elif Length(arg) = 2 then

      if IsElt(arg[2]) and EltOrder(arg[2]) = arg[1] then
         return EltFactor(arg[2]);
      elif IsIdeal(arg[2]) and IdealOrder(arg[2]) = arg[1] then
         return IdealFactor(arg[2]);
      elif IsInt(arg[2]) and IsOrder(arg[1]) then
         return EltFactor(Elt(arg[1],arg[2]));
      elif IsPoly(arg[1]) and IsPrime(arg[2]) then
         return PolyFactor(arg[1], arg[2]);
      else
         return Error(s);
      fi;
   else
      return Error(s);
   fi;
end;

if not IsBound(SETPRINTLEVEL) then
  SETPRINTLEVEL := function(name,level)
    return PRINTLEVEL(name,level);
  end;
fi;


#############################################################################
##
#F  Gcd ( <object> , <object> ) or Gcd ([ <object> ])  Returns the
#   greatest common divisor
#/*TEX***TE|***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Gcd
#
#SHORTDOC  Returns greatest common divisor of list of arguments.
#
#LONGDOC   Returns the greatest common divisor of list of arguments or simply the
#          greatest common divisor of two arguments. Supported are integers,
#          ideals and polynomials over $ S$ where $S$ is a field or $\Z$  .
#
#CALLING   gcd := Gcd(L);\\
#          gcd := Gcd(a, b);
#
#ARGUS     gcd @ integer or polynomial or ideal
#          L @ list of integers of polynomials or ideals
#          a @ integer or polynomial or ideal
#          b @ integer or polynomial or ideal
#
#PACKAGE   lib
#
#XREF      IntGcd, PolyGcd
#
#KASH      Gcd of integers:@
#          Gcd([18, 12, 4, 6]);
#          Gcd(18, 12);
#          Gcd([18, 12]);
#
#KASH      Gcd of polynomials:@
#          x := Poly(Zx, [1,0]);
#          f1 := x^2-1;
#          f2 := x+1;
#          f3 := 12*x^3 + 12*x^2 + 13*x + 13;
#          Gcd([f1, f2, f3, x-1]);
#          Gcd(f1, f2);
#          Gcd([f1, f2]);
#
#KASH      Gcd of ideals:@
#          o := OrderMaximal(Order(Z, 2, 3));
#          i1 := Ideal(o, 2);
#          i2 := Ideal(o, 3);
#          i3 := Ideal(o, 4);
#          i4 := Ideal(o, 8);
#          Gcd([i1, i3, i4]);
#          a := Gcd(i1, i2);
#          a = Ideal(o, 1);
#          Gcd([i1, i2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##


Gcd := function(arg)
   local GcdList,s;

   GcdList := function(l)
     return Iterated(l, Gcd);
   end;

   s := "usage:\n Gcd( [<integer>] )\n Gcd( [<polynomial>] )\n Gcd( [<ideal>] )\n Gcd( <integer>, <integer> )\n Gcd( <polynomial>, <polynomial> )\n Gcd( <ideal>, <ideal> ) ";

   if Length(arg) = 1 and IsList(arg[1]) then
      if Length(arg[1]) > 0 then
         if IsInt(arg[1][1]) then
            return GcdList(arg[1]);
         elif IsPoly(arg[1][1]) then
            return GcdList(arg[1]);
         elif IsIdeal(arg[1][1]) then
            return GcdList(arg[1]);
         else
            return Error(s);
         fi;
      else
         return Error(s);
      fi;
   elif Length(arg) = 2 then
      if IsInt(arg[1]) and IsInt(arg[2]) then
         return IntGcd(arg[1],arg[2]);
      elif IsPoly(arg[1]) and IsPoly(arg[2]) then
         return PolyGcd(arg[1],arg[2]);
      elif IsIdeal(arg[1]) and IsIdeal(arg[2]) then
         return (arg[1]+arg[2]);
      else
         return Error(s);
      fi;
   else
      return Error(s);
   fi;
end;

PolyLcm := function(a,b)
  return a*b/PolyGcd(a,b);
end;

#############################################################################
##
#F  Lcm ( <object> , <object> ) or Lcm ([ <object> ])  Returns the
#   least commom multiple
#/*TEX***TE|***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Lcm
#
#SHORTDOC  Returns least common multiple of list of arguments.
#
#LONGDOC   Returns the least common multiple of list of arguments or simply the
#          least common multiple of two arguments. Supported are integers,
#          ideals and polynomials over $S$ where $S$ is a field or $\Z$ .
#
#CALLING   lcm := Lcm(L);\\
#          lcm := Lcm(a, b);
#
#ARGUS     lcm @ integer or polynomial or ideal
#          L @ list of integers of polynomials or ideals
#          a @ integer or polynomial or ideal
#          b @ integer or polynomial or ideal
#
#XREF      IntLcm, IdealLcm;
#
#PACKAGE   lib
#
#
#KASH      Lcm of integers:@
#          Lcm([18, 12 ,4]);
#          Lcm(18, 12);
#          Lcm([18, 12]);
#
#KASH      Lcm of polynomials:@
#          x := Poly(Zx, [1,0]);
#          f1 := x^2-1;
#          f2 := x+1;
#          f3 := 12*x^3 + 12*x^2 + 13*x + 13;
#          Lcm([f1, f2, f3]);
#          Lcm(f1, f2);
#          Lcm([f1, f2]);
#
#KASH      Lcm of ideals:@
#          o := OrderMaximal(Order(Z, 2, 3));
#          i1 := Ideal(o, 2);
#          i2 := Ideal(o, 3);
#          i3 := Ideal(o, 4);
#          i4 := Ideal(o, 8);
#          Lcm([i1, i2, i3, i4]);
#          Lcm(i1, i2);
#          Lcm([i1, i2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##


Lcm := function(arg)
   local LcmList,s;

   LcmList := function(l)
      return Iterated(l, Lcm);
   end;

   s := "usage:\n Lcm( [<integer>] )\n Lcm( [<polynomial>] )\n Lcm( [<ideal>] )\n Lcm( <integer>,<integer>\n Lcm( <polynomial>,<polynomial> )\n Lcm( <ideal>,<ideal> ) ) ";

   if Length(arg) = 1 and IsList(arg[1]) then
      if Length(arg[1]) > 0 then
         if IsInt(arg[1][1]) then
            return LcmList(arg[1]);
         elif IsPoly(arg[1][1]) then
            return LcmList(arg[1]);
         elif IsIdeal(arg[1][1]) then
            return LcmList(arg[1]);
         else
            return Error(s);
         fi;
      else
         return Error(s);
      fi;
   elif Length(arg) = 2 then
      if IsInt(arg[1]) and IsInt(arg[2]) then
         return IntLcm(arg[1],arg[2]);
      elif IsPoly(arg[1]) and IsPoly(arg[2]) then
         return PolyLcm(arg[1],arg[2]);
      elif IsIdeal(arg[1]) and IsIdeal(arg[2]) then
         return IdealLcm(arg[1],arg[2]);
      else
         return Error(s);
      fi;
   else
      return Error(s);
   fi;
end;



_D2H := D2H;
Unbind(D2H);
D2H := function(a)
  local s, i, m;

  if not IsInt(a) or a<0 then
    Error("usage: D2H( int>=0 )");
  fi;

  m := 2^24;
  s := "";
  repeat
    i := _D2H(a mod m);
    a := IntQuo(a, m);
    if a>0 then
      i := Concatenation("000000", i){[Length(i)+1..Length(i)+6]};
    fi;
    s := Concatenation(i, s);
  until a=0;

  return s;
end;


_H2D := H2D;
Unbind(H2D);
H2D := function(s)
  local a, m, i, j, t;
  m := 2^24;
  i := Length(s);
  if i mod 6 <> 0 then
    a := "000000";
    s := Concatenation(a{[1..6-(i mod 6)]}, s);
    i := Length(s);
  fi;

  a := 0;
  j := 1;
  while j<i do
    t := s{[j..j+5]}; IsString(t);
    a := a*m + _H2D(t);
    j := j+6;
  od;
  return a;
end;



#############################################################################
##
#F  Move ( <object> , <structure> )
#
#/*TEX***TE|***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Move
#
#SHORTDOC  Moves an object to another structure (order, ring)
#
#LONGDOC   Moves algebraic numbers (Elt), ideals (Ideal),
#          pseudomatrices (Module), or lists of those to the order {\tt O}.\\
#          Moves a matrix (Mat) to a matrix over another ring R.\\
#          Tries to compute a representation of a polynomial in a different
#          polynomial algebra. 
#CALLING   a := Move(L, O);\\
#          N := Move(M, r);\\
#          a := Move(f, r)
#ARGUS     a @ integer or polynomial or ideal
#          L @ list of integers of polynomials or ideals
#          O @ Order
#          N, M @ matrix
#          f @ polynomial
#          r @ ring
#PACKAGE   lib
#
#XREF      EltMove, IdealMove, ModuleMove, MatMove, PolyMove, FFEltMove
#
#KASH      This is the same example as IdealMove apart from using the
#          function Move.@
#          O := OrderMaximal(Order(x^6 - 9*x^4 - 4*x^3 + 27*x^2 - 36*x - 23));;
#          o := OrderShort(O);;
#          Move(2*o, O);
#          Move(Factor(2*o), O);
#KASH      An example for a matrix:@
#          M:=Mat(Z,[[1236,125,5],[65460,1997,3585],[2045,546550,1178]]);
#          k:=FF(5,3);
#          N:= Move(M,k);
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#      20. 1.98   AH   written
#      99-02-21   jm   minor changes and extended
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##
Move := function(arg)
  local s, O;
  
  s := "usage:\n Move( <elt>,<order> ) \n Move( <ideal>,<order> ) \n Move( <module>,<order> ) \n Move( <list>,<order> ) \n Move( <matrix>,<ring> )\n Move( <polynomial> , <polynomial algebra>)\n Move( <polynomial> , <Ring>) ";

  if Length(arg)<>2 then Error(s); fi;
  if (TYPE(arg[1])="KANT polynomial") then
     if(TYPE(arg[2])="KANT polynomial algebra") then
        return PolyMove(arg[1],arg[2]);
        elif(TYPE(arg[2])= "KANT ring") then
           return PolyMove(arg[1],arg[2]);
        elif(TYPE(arg[2])="KANT order") then
           return PolyMove(arg[1],arg[2]);
        elif(TYPE(arg[2])="KANT finite field") then
           return PolyMove(arg[1],arg[2]);
        else
           Print("Second argument has to be a ring. \n");
       fi;
  elif (TYPE(arg[1])="KANT matrix") then
     if(TYPE(arg[2])="KANT ring") then
        return MatMove(arg[1],arg[2]);
     elif(TYPE(arg[2])="KANT order") then
        return MatMove(arg[1],arg[2]);
     elif(TYPE(arg[2])="KANT finite field") then
        return MatMove(arg[1],arg[2]);
     elif(TYPE(arg[2])="KANT polynomial algebra") then
        return MatMove(arg[1],arg[2]);
     elif(TYPE(arg[2])="KANT quotient field") then
        return MatMove(arg[1],arg[2]);
     else
        Print("Second argument has to be a ring. \n");
     fi;
  elif TYPE(arg[1])="KANT algebraic element" then
     if(TYPE(arg[2])="KANT lattice") then
        return EltMove(arg[1],arg[2]);
     fi;
  elif (TYPE(arg[1])="KANT lattice") then
     if(TYPE(arg[2])="KANT lattice")then
        return EltMove(arg[1],arg[2]);
     fi;
  else

  O := arg[2];
  if TYPE(O)<>"KANT order" then Error("Second argument must be an order\n",s); fi;
  
  if TYPE(arg[1])="KANT algebraic element" then return EltMove(arg[1],O); fi;
  if TYPE(arg[1])="KANT lattice elt" then return EltMove(arg[1],O); fi;
  if TYPE(arg[1])="KANT ideal" then return IdealMove(arg[1],O); fi;
  if TYPE(arg[1])="KANT module" then return ModuleMove(arg[1],O); fi;
  if TYPE(arg[1])="list" then return IdealMove(arg[1],O); fi;
  
  Error("illegal first argument!\n",s);
  fi;
end;

###############################################################################
#F  CharPoly ( <obj> ) . . . Computes the characteristic polynomial of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       CharPoly
#
#SHORTDOC  Computes the characteristic polynomial of an algebraic element, a 
#          matrix or an alff element
#LONGDOC   For an algebraic element $a$ in an Order $O$ and a polynomial 
#          algebra $PA$ over another order $o$ or this order,
#          this function computes the characteristic polynomial of $a$ over $o$
#          i.e. the returned polynomial is contained in $PA$.\\
#          Given a matrix this function returns its characteristic polynomial\\
#          For an alff element $a$ in an order $o$ in the alff $F$ this 
#          function computes the characteristic polynomial of $a$ over $o$
#CALLING   p := CharPoly (a [,PA]);\\
#          p := CharPoly (a [,O]);\\
#          p := CharPoly( M );
#ARGUS     p @ polynomial
#          PA @ polynomial algebra
#          O @ suborder
#          a @ algebraic element alff order element
#          M @ matrix
#    
#PACKAGE   lib
#
#XREF      EltCharPoly AlffEltCharPoly MatCharPoly
#       
#KASH       @
#          o := Order(Z,2,2);
#          a := Elt(o,[0,1]);
#          p := CharPoly(a);
#          M := Mat(o,[[1,a],[1,-a]]);
#          p := CharPoly(M);
#KASH      An example for global function fields: @
#          AlffInit(FF(11, 1));
#          AlffOrders(y^2-y-T^3+T^2);
#          a := AlffElt(o, [1,1]);
#          p := CharPoly(a);
#         
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#  98-09-08 jm written
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

CharPoly := function ( arg )
     local argc, a;
     
      argc:=  Length(arg);
   
      if argc <1 or argc>2 then
       return Error("usage: CharPoly( <order element> [, <polynomial algebra> ] | <order element> [, <order> | <matrix> | <alff element>");
      fi;
    
    a:= arg[1];
    if argc = 1 then
       
      if ( IsElt (a) ) then
         return  EltCharPoly(a);
      fi;

      if (IsMat (a)) then
         return MatCharPoly (a);
      fi;

      if (IsAlffElt (a)) then
        return AlffEltCharPoly(a);
      fi;
   
    else 
      if (IsElt (a)) then
         return EltCharPoly( a, arg[2]);
      fi;
   fi;
   return Error("usage: CharPoly( <order element> [, <polynomial algebra> ] | <order element> [, <order> | <matrix> | <alff element>");
end;
   
###############################################################################
#
#F  MinPoly ( <obj> ) . . . Computes the minimal polynomial of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       MinPoly
#
#SHORTDOC  Computes the minimal polynomial of an algebraic element, a matrix,  
#          or an alff element
#LONGDOC   For an algebraic element $a$ in an Order $O$ and a polynomial 
#          algebra $PA$ over another order $o$ or this order,
#          this function computes the minimal polynomial of $a$ over $o$
#          i.e. the returned polynomial is contained in $PA$.\\
#          Given a matrix this function returns its minimal polynomial.\\
#          Given a alff element this function computes the minimal polynomial
#CALLING   p := MinPoly (a [,PA]);\\
#          p := MinPoly (a [,O]);\\
#          p := MinPoly( M );\\
#          p := MinPoly(a);
#ARGUS     p @ polynomial
#          PA @ polynomial algebra
#          O @ suborder
#          a @ algebraic element  alff order element
#          M @ matrix
#PACKAGE   lib
#
#XREF      EltMinPoly MatMinPoly  AlffEltMinPoly 
# 
#KASH       @
#          o := Order(Z,2,2);
#          a := Elt(o,[0,1]);
#          p := MinPoly(a);
#          M := Mat(o,[[1,a],[1,-a]]);
#          p := MinPoly(M);
#
#KASH      An example for global function fields: @
#          AlffInit(FF(3, 1));
#          AlffOrders(y^3+T^3*y+T);
#          a := AlffElt(o, [1,1,1]);
#          p := MinPoly(a);
#         
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#  1998-09-10 jm written
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

MinPoly := function ( arg )
     local argc, a;
     
      argc:=  Length(arg);
   
      if argc <1 or argc>2 then
       return Error("usage: MinPoly( <order element> [, <polynomial algebra> ]                      | <order element> [, <order> | <matrix> | <alff element>");
      fi;
    
    a:= arg[1];
    if argc = 1 then
      
      if ( IsElt (a) ) then
         return  EltMinPoly(a);
      fi;

      if (IsMat (a)) then
         return MatMinPoly (a);
      fi;

      if (IsAlffElt (a)) then
        return AlffEltMinPoly(a);
      fi;
   
    else 
      if (IsElt (a)) then
         return EltMinPoly( a, arg[2]);
      fi;
    fi;

   return Error("usage: MinPoly( <order element> [, <polynomial algebra> ] | <order element> [, <order> | <matrix> | <alff element>");
end;

###############################################################################
#
#F   Index ( <obj> ) . . . Computes the index of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Index
#
#SHORTDOC  Computes the index of an algebraic element, an order
#          or an alff order
#LONGDOC   For an algebraic element $a$ in an arbitrary order $o$ over $\Z$  
#          this function returns the modul index $(o : \Z[\alpha] $
#          Given an order this function returns the index of the suborder 
#          $o_1$ of $o$ in $o$.
#          Given an alff order this function returns the index of the 
#          equation order in the given order.
#CALLING   I := Index (a);\\
#          I := Index (o);
#ARGUS     I @ coefficient ring element
#          a @ algebraic element 
#          o @ order alff order
#
#PACKAGE   lib
#
#XREF      EltIndex OrderIndex AlffOrderIndex
#       
#KASH       @
#          o := Order(Z,5,3);
#          a := Elt(o, [0,0,1,1,1]);
#          I := Index(a);
#KASH       @
#          o := Order(Poly(Zx,[1,0,111,0,177,0,1999]));
#          O := OrderMaximal(o);
#          I := Index(O);
#KASH      An example for global function fields: @
#          AlffInit(FF(11, 1));
#          AlffOrders(y^2-y-T^3+T^2);
#          o := AlffOrderMaxFinite(F);
#          I := Index(o);
#         
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#    1998-09-16 jm written
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

Index := function ( a )

      
      if ( IsElt (a) ) then
         return  EltIndex(a);
      fi;
      
      if ( IsOrder(a) ) then
         return OrderIndex(a);
      fi;
      
      if (IsAlffOrder (a) ) then
        return AlffOrderIndex(a);
     fi;
     
   return Error("usage: Index( <order element> | <order> | <alff order>");
   end;



#############################################################################
##
#F  OrderPoly ( <obj> ) . . . Returns the defining polynomial of obj.
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       OrderPoly
#   
#
#SHORTDOC  Returns the defining polynomial of the associated  equation order   
#LONGDOC   Returns the polynomial defining the equation order of $\o$
#          represented in the polynomial algebra $P$.\\
#          The polynomial algebra $P$ has to be defined over the coefficient
#          ring of $\o$. If the polynomial algebra $P$ is omitted,
#          the polynomial $f$ is represented over the coefficient ring
#          of the order $\o.$\\
#          For an order of an algebraic function field, the polynomial $f$ is 
#          an element of $k[T,y]$ or ${\cal O}_\infty[y]$.
#             
#CALLING   f := OrderPoly(P,o); \\   
#          f := OrderPoly(o);   
#ARGUS     f @ polynomial
#          P @ polynomial algebra
#          o @ order $|$ algebraic function field order
#
#XREF      OrderEquationOrder, AlffOrderDeg, Alff.
#PACKAGE   lib
#
#
#KASH       @
#          o := Order (Poly (Zx,[1,0,-186,0,13097,0,-412704,0,4946176]));
#          f := OrderPoly(Zx,o);
#          o := OrderMaximal (o);
#          f := OrderPoly(Zx,o);
#          O := Order (o,2,99);
#          ox := PolyAlg (o);
#          f := OrderPoly (ox,O);
#KASH      An example for algebraic function fields: @
#          AlffInit(FF(2,5));
#          AlffOrders(y^2+y+T^3+T);
#          OrderPoly(o);   
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#  1999-01-13 jm written
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

OrderPoly := function ( arg )
   local argc, a;
   
   argc:=  Length(arg);
   if argc <1 or argc>2 then
       return Error("usage: OrderPoly([<polynomial algebra>,] <order> |  <alff                      order>");
    fi;
    a:= arg[1];
    if argc = 1 then
       
      if (IsOrder( a )) then
        return _OrderPoly( a );
      fi;
     
      if (IsAlffOrder( a )) then
        return AlffOrderPoly( a );
     fi;
  else
     if IsPolyAlg ( a ) then
        return _OrderPoly( a , arg[2]);
     fi;   
   fi;
    return Error("usage: OrderPoly( [<polynomial algebra>,] <order> | <alff order> )");
end;

#############################################################################
##
#F Den ( <object> ) or Den( <object>, <object> )  Returns the denominator
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#FXN       Den
#
#SHORTDOC  Returns the denominator of an object.
#  
#LONGDOC   For a rational number or a rational function the function 
#          returns  the denominator of the rational argument.\\
#          The denominator of the algebraic number $a$ is the smallest
#          natural number $d$ such that $d\cdot a$ is an algebraic
#          integer. If you use {\tt Den(a, "rep")} the  
#          denominator in the representation of $a$ is returned.\\
#          For an algebraic function field order element 
#          see {\tt AlffEltDen}.\\
#          The denominator of a fractional ideal $\a$ is the smallest
#          positive integer $d$ that $d\cdot \a$ is an integral ideal.
#          
#
#CALLING   d := Den( a );\\
#          d := Den(a, "rep");\\
#          d := Den( q  );\\
#          I := Den( I );
#
#ARGUS     d  @  integer 
#          d  @  quotient field element or polynomial
#          q  @  rational
#          a  @  algebraic element 
#          a  @  algebraic function field order element
#          q  @  quotient field element or polynomial
#          I  @  ideal
#
#
#KASH      An absolute example for an algebraic element: @
#
#          o := Order (x^3 - 23*x^2 + 146*x - 244);;
#          a := Elt (o, [1/2,3/5,45/5657567]);
#          Den ( a );
#          Den ( a, "rep" );
#KASH      An example for algebraic function fields: @
#          AlffInit(Q);;
#          AlffOrders( y^3+T^4+1 );;
#          a := AlffElt(o, [1,T,1]);
#          b := 1/a;
#          d := Den( b );
#KASH      An example for an ideal:@
#          o := Order( x^7 + x + 1);;
#          a := Elt(o,[0,0,0,1,1,0,1]);;
#          I := a * o;
#          b :=I^-2;
#          d := Den( b );
#
#          
#**HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
#
Den := function(arg)
local argc, a;

 argc :=  Length(arg);

 if argc < 1 or argc > 2 then
    return Error("usage: Den( <algebraic element> [,rep ] | <ideal> | <rational> | <alff element>");
 fi;

   a := arg[1];
   if argc = 1 then

       if IsIdeal ( a ) then
         return IdealDen ( a );
       fi;
       if IsPoly(a) or IsQfe(a) then
          return QfeDen(a);
       fi;   
       if IsAlffElt(a) then
         return AlffEltDen ( a );
       fi;
       if  IsRat(a) then
         return Denominator ( a );
       fi;
       if IsElt( a ) then
          return EltDen( a );
       fi;
       if IsInt(a) then
          return  1;
       fi;
    else 
       if IsElt ( a ) then
          return EltDen (a, arg[2]);
       fi; 
    fi;
    return Error("usage: Den( <algebraic element> [,rep ] | <ideal> | <rational> | <alff element> | < quotient field element or polynomial > ");
end;

#############################################################################
##
#F Num ( <object> ) or Num( <object>, <object> )  Returns the numerator
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#FXN       Num
#
#SHORTDOC  Returns the numerator of an object.
#
#LONGDOC   For a rational number the function returns the numerator of the
#          rational argument.\\
#          For an algebraic number $a$ or for a  fractional ideal $\a$ the
#          function returns $d\cdot a$ or $d\cdot \a$. $d$ is the smallest
#          integer  that $d\cdot a$ is an algebraic integral
#          or an integral ideal\\
#          For an algebraic function field order element
#          see {\tt AlffEltDen}.
#CALLING   d := Num( q );\\
#          d := Num( a );\\
#          d := Num(I);
#ARGUS     d  @  integer 
#          d  @  quotient field element or polynomial
#          q  @  rational
#          a  @  algebraic function field order element
#          a  @  algebraic element
#          q  @  quotient field element or polynomial
#          I  @  ideal
#
#KASH      An absolute example for an algebraic element: @
#          o := Order (x^3 - 23*x^2 + 146*x - 244);;
#          a := Elt (o, [1/2,3/5,45/5657567]);
#          Num ( a );
#KASH      An example for an ideal:@
#          o := Order( x^7 + x + 1);;
#          a := Elt(o,[0,0,0,1,1,0,1]);;
#          I := a * o;
#          b :=I^-2;
#          Num( b );
#KASH      An example for a quotient field element: @
#          k := FF(2,5);
#          kx := PolyAlg(k);
#          f:= Poly(kx,[1,0]);
#          g :=Poly(kx,[1,1]);
#          h := f/g;
#          Num( h );
#KASH      An example for algebraic function fields: @
#          AlffInit(Q);;
#          AlffOrders( y^3+T^4+1 );;
#          a := AlffElt(o, [1,T,1]);
#          b := 1/a;
#          d := Num( b );
#**HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
#
Num := function(a)
        if  IsIdeal ( a ) then
         return IdealDen ( a )*a;
        fi;
        if IsElt( a ) then
          return EltDen( a )*a;
        fi;
        if IsPoly(a) or IsQfe(a) then
           return QfeNum( a );
        fi;
        if IsAlffElt(a) then
         return AlffEltNum ( a );
        fi;
        if IsRat(a) then
         return Numerator ( a );
        fi;

    return Error("usage: Num( <algebraic element> | <ideal> | <rational> | <alff element> | <quotient field element or polynomial>");
end;


PrDate := function(d)
  Print(d.mday,".",d.mon,".",d.year," ",d.hour);
  if (d.min<10) then
    Print(":0",d.min);
  else
    Print(":",d.min);
  fi;
  if (d.sec<10) then
    Print(":0",d.sec);
  else
    Print(":",d.sec);
  fi;
  return true;
end;

SubDate := function(d1,d2)
local d, u;

#  d:=Date();
#  if (d1.sec>d2.sec)
#    d.sec=d1.sec-d2.sec;
#    u=0;
#  else
#    d.sec=d1.sec-d2.sec+60;
#    u=1;
#  fi;  
#  if (d1.min>d2.min)
#    d.min=d1.min-d2.min;
#    u=0;
#  else
#    d.min=d1.min-d2.min+60-u;
#    u=1;
#  fi;  
#  if (d1.hour>d2.min)
#    d.hour=d1.hour-d2.hour;
#    u=0;
#  else
#    d.hour=d1.hour-d2.hour+24-u;
#    u=1;
#  fi;  
#  if (d1.mday>d2.mday)
#    d.mday=d1.mday-d2.mday;
#    u=0;
#  else
#    d.mday=d1.mday-d2.mday+24-u;
#    u=1;
#  fi;  
#
#  d.mday:=d1.mday-d2.mday;
#  d.mon:=d1.mon-d2.mon;
#  d.year:=d1.year-d2.year;
#
#  d.wday=false;
#
  return (false);
end;

#############################################################################
##
#F Date( )  Returns the current date and time in a record
#/*TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX*
#
#USERDEF
#FXN       Date
#
#SHORTDOC  Returns the current date and time in a record
#
#CALLING   t := Date(  );
#
#ARGUS     t  @  record 
#
#KASH      An example:@
#          t := Date();
#
#**HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
#
Date := function()
local  d;
  d:=_Date();
#  d.operations:=rec(Print:=PrDate,\-:=SubDate);
  d.operations.Print:=PrDate;
  return d;
end;

