#############################################################################
##
#A  kantpoly.g                    KANT library              Mario Daberkow
##
#A  @(#)$Id: kantpoly.g 1.00 1995/02/09 Daberkow $
##
#Y  Copyright 1995,  Prof. Dr. M. Pohst, TU - Berlin,  Germany
##
##  This file contains the Polynomial KANT functions .
##
#H  $Log: kantpoly.g $
#H  Revision 1.00  1995/02/09  Daberkow
#H  added 'IsPoly'
#H  added  PolyMove CF
#H  added  Eval     CF
#H  removed  PolyMove MD
##


#############################################################################
##
#F  IsPoly( <object> )  . . . . . . . . .  checks if argument is a polynomial
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       IsPoly
#CALLING   b := IsPoly(f);
#SHORTDOC  Returns true iff the argument is a polynomial.
#ARGUS     b @ boolean
#          f
#
#PACKAGE   lib
#
#KASH      @
#          f := Poly(Zx, [1, 2, 1, 2]);
#          IsPoly(f);
#          IsPoly(3);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##


IsPoly := function ( pol )

  if (TYPE (pol) = "KANT polynomial") then
    return true;
  else
    return false;
  fi;
end;


#############################################################################
##
#F  Eval( poly, at )  . . . . evaluates a polynomial at a certain value
#/*TEX**TEX|***TEX***TEX***TEX***TEX**TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       Eval
#CALLING   y := Eval(f, s);
#SHORTDOC  Evaluates a polynomial at a value.
#LONGDOC   This function assumes that the arithmetic with elements of
#          the coefficient ring of the polynomial algebra of {\tt f} and {\tt s}
#          is defined.
#ARGUS     y
#          f @ polynomial
#          s
#
#PACKAGE   lib
#
#KASH      First we create a polynomial (over $\Z$): @
#          f := x^2+2*x+1;
#          @
#          Now we want to evalute it at $2$ and $\pi$: @
#          Eval(f, 2);
#          Eval(f, pi);@
#          Finally, we want to substitute x by x+1: @
#          Eval(f, x+1);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

Eval := function(poly, at)
  local i, value, coefs;

  if not IsPoly(poly) then
    return Error("usage: Eval( <poly>, <elt> );");
  fi;

  coefs := PolyToList(poly);

  if IsMat(at) then
    if MatCols(at)<> MatRows(at) then
      Error("<elt> must be a square matrix");
    fi;
    Apply(coefs, i->i*MatId(Z, MatCols(at)));
  fi;
  if Length(coefs) = 0 then
    return 0*at;
  fi;

  value := coefs[1];
  for i in [2 .. Length(coefs)] do
    value := at * value + coefs[i];
  od;

  return  value;
end;



