#############################################################################
#                                                                            #
#                        HB 18.09.98 written                                 #
#                                                                            #
##############################################################################

_RRRFinitePrimePower        := Ignore;
_SecondFactor               := Ignore;
_SecondFactorDiscreteLog    := Ignore;
_PrimePowerDiscreteLog      := Ignore;
_DirectProductDiscreteLog   := Ignore;
_RRRFinite                  := Ignore;
_RRR                        := Ignore;
_RCG                        := Ignore;
RayResidueRingCyclicFactors := Ignore;
RayClassGroupCyclicFactors  := Ignore;
_FindZRay                   := Ignore;
_gAssureZRCG                := Ignore;
_gAssureZRRR                := Ignore;
RayClassGroup               := Ignore;
RayResidueRing              := Ignore;
RayConductor                := Ignore;
RayConductorTest            := Ignore;
EltRayResidueRingRep        := Ignore;
RayResidueRingRepToElt      := Ignore;
IdealRayClassRep            := Ignore;
ZIdealCreate                := Ignore;


if not IsBound(_ZRayInfo) then
  _ZRayInfo := [];
fi;

_ZRayInfoClear := function()
  _ZRayInfo := [];
  GASMAN("g");
end;

_RCG := function(arg)
local rrr, sub, g, s, gens, h, m, iso, iso_inv, G;

  if Length(arg) = 1 then
    arg[2] := [];
  fi;


  if not((Length(arg) = 1 and IsRecType(arg[1], "ZIdeal")) or
         (Length(arg) = 2 and IsRecType(arg[1], "ZIdeal")
         and IsList(arg[2]))) then
    Error("use RayClassGroup(m0, [, minf])");
  fi;

  rrr := _gAssureZRRR(arg[1], arg[2]);

  sub := AbelianSubGroup(rrr, AbelianGroupDiscreteLog(rrr, -1));
  h := AbelianQuotientGroup(rrr, sub);    

  # the ray class group g is isomorphic to h
  g := AbelianGroupCreate(h.Relations);
  
  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  if not IsBound(g.Information.RCG) then
    g.Information.RCG := rec();
  fi;
  g.Information.RCG.m0 := arg[1];
  g.Information.RCG.minf := arg[2];

  m := MatId(Z, AbelianGroupNumberGenerators(g));
  iso := AbelianGroupHomCreate(g, h, m, m);
  iso_inv := iso^-1;
  g.Isomorphic := [iso];
  if not IsBound(h.Isomorphic) then
    h.Isomorphic := [iso_inv];
  else
    Add(h.Isomorphic, iso_inv);
  fi;

  G := AbelianGroupSmithCreate(g, true);
  iso := iso*G.Isomorphic[1];
  iso_inv := (G.Isomorphic[1])^-1*iso_inv;
  G.Information.LogExpHom := [iso, iso_inv];

  G.DiscreteLog := function(g, a)
    local hom, hom_inv, h, elt;

    if not IsRecType(a, "ZIdeal") then
      Error("use DiscreteLog(ideal)");
    fi;

    hom := g.Information.LogExpHom[2];
    h := hom.g1;

    a := a.Generator;
    elt := AbelianGroupDiscreteLog(h, a);
    elt := hom*elt;
    return elt;
  end;

  G.DiscreteExp := function(a)
  local hom, g, gen;

    g := a.Group;
    hom := g.Information.LogExpHom[1];

    a := hom*a;
    gen := AbelianGroupDiscreteExp(a);
    a := ZIdealCreate(gen);
    return a;
  end;
  return G;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
# 
#FXN       IdealRayClassRep
#SHORTDOC  Returns a representive of the class of an ideal in the 
#          ray class group.
#LONGDOC   Decomposes an ideal, or more exactly, the class of this ideal
#          (that need to be coprime to ${\goth{m}}_0$)
#          into a power product of the 
#          classes generating the whole ray class group.
#         
#          Based on the solutions for the discrete logarithm problem for
#          class groups (\cite{He1}) and for ray residue rings
#          (see {\tt EltRayResidueRingRep}), an exponent vector relative to
#          the generators is computed. 
#         
#CALLING   r := IdealRayClassRep(I,m0,minf);
#ARGUS     r @ matrix
#          I @ ideal @ 
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes@
#CSRC      order\_ideal\_to\_rcg
#XREF      EltCon, OrderClassGroup, RayClassGroup, RayClassGroupCyclicFactors.
#KASH      @
#          O := OrderMaximal(Order(x^2-2*x-5));  
#          OrderClassGroup(O,500,"euler","fast");
#          m0 := 27*O;; 
#          minf := [2];;
#          L := RayClassGroupCyclicFactors(m0,minf);
#          I := L[1][1];
#          IdealRayClassRep(I,m0,minf);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

IdealRayClassRep := function(arg)
local g, arg2, a, elt, expvec;

  if Length(arg) > 3 or Length(arg) < 2 then
    Error("use IdealRayClassRep(ideal, m0 [,minf])");
  fi;

  if Length(arg) = 2 then
    arg[3] := [];
  fi;

  if TYPE(arg[2]) = "KANT ideal" then
    return OrderIdealRayClassRep(arg[1], arg[2], arg[3]);
  fi;  

  a := arg[1];
  if Length(arg) = 2 then
    g := _gAssureZRCG(arg[2]);
  elif Length(arg) = 3 then
    g := _gAssureZRCG(arg[2], arg[3]);
  else 
    Error("use IdealRayClassRep(elt, m0, minf)");
  fi;

  elt := AbelianGroupDiscreteLog(g, a);
  expvec := elt.expvec;
  
  return expvec;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       RayClassGroupCyclicFactors
#
#SHORTDOC  Returns the generators and the orders of the generators of the ray 
#          class group modulo a congruence module as computed using 
#          {\tt RayClassGroup}.
#
#CALLING   L := RayClassGroupCyclicFactors(m0 [,minf]);
#ARGUS     L @ list
#          m0 @ ideal @
#          minf @ list @of integers/infinite primes @
#       
#PACKAGE   
#
#CSRC      order\_ideal\_rcr\_mg\_assure
#
#XREF      EltCon, IdealRayClassRep, OrderClassGroup, RayClassGroup.
#KASH      @
#          O := OrderMaximal(Order(Z,2,10));;
#          OrderClassGroup(O,500,"euler");
#          m0 := 7^3*O;;
#          L := RayClassGroupCyclicFactors(m0);
#          IdealRayClassRep(L[1][1],m0);
#          IdealRayClassRep(L[2][1],m0);
#          IdealRayClassRep(L[3][1],m0);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayClassGroupCyclicFactors := function(arg)
local g, list;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use RayClassGroupCyclicFactors(m0 [,minf])");
  fi;

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if TYPE(arg[1]) = "KANT ideal" then
    return OrderRayClassGroupCyclicFactors(arg[1], arg[2]);
  fi;  

  g := _gAssureZRCG(arg[1], arg[2]);

  list := Copy(AbelianGroupCyclicFactors(g));

  return list;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       RayClassGroup
#
#SHORTDOC  Returns the ray class number and the orders of the generators of 
#          the ray class group modulo a congruence module.
#LONGDOC   Computes the ray class group modulo a congruence module 
#          ${\goth{m}} = {\goth{m}}_0{\goth{m}}_\infty$, where ${\goth{m}}_0$ 
#          is an integral ideal of a maximal 
#          order $\O$ and ${\goth{m}}_\infty$ is a formal product of primes at 
#          infinity --- here a subset of the real embeddings of $\O$.
#          ${\goth{m}}_\infty$ is represented by a list of integers, where 
#          each entry represents a real embedding (as given by {\tt EltCon}).
#          {\tt RayClassGroup} returns a list consisting of the ray class
#          number and a
#          list of the orders of the cyclic factors of the ray class group.
#
#          It might be necessary to use a higher precision (set by 
#          {\tt OrderPrec}) in order to get correct results when working
#          with primes at infinity.
#          Computing the class group and the units of the maximal order
#          first can speed up the computation of the ray class group, as 
#          it is possible to influence the calculation of the class group,
#          for instance by setting a lower bound for the regulator.
#          For a description of the algorithm see \cite{Pau1,Pau2}.           
#CALLING   L := RayClassGroup(m0 [,minf]);\\
#          L := RayClassGroup(o [,minf]);
#ARGUS     L @ list
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes@
#          o @ order @
#         
#PACKAGE   
#
#CSRC      order\_ideal\_rcr\_mg\_assure
#
#XREF      EltCon, IdealRayClassRep, OrderClassGroup, OrderPrec,
#          RayClassGroupCyclicFactors, RayConductor.
#KASH      @
#          O := OrderMaximal(Order(Z,2,10));;
#          OrderClassGroup(O,500,"euler","fast");
#          m0 := 27*O;;
#          minf := [1,2];;
#          L := RayClassGroup(m0,minf);
#          o:=OrderMaximal(Poly(Zx,[1,1,1]));
#          L:=RayClassGroup(o);
#          @ compute the ray class group of an ideal in $\Z$ @
#          a := ZIdealCreate(9);
#          RayClassGroup(a);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#   17.3.98 jm argument added
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayClassGroup := function(arg)
local g, list;
 
  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use RayClassGroup(m0 [,minf])");
  fi;

  if Length(arg) > 2 then
    Error("use RayClassGroup(m0, [, minf])"); 
  fi; 

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if IsOrder(arg[1]) then 
    arg[1] := 1*arg[1];
  fi;

  if TYPE(arg[1]) = "KANT ideal" then
    return OrderRayClassGroup(arg[1], arg[2]);
  fi;  

  if IsRec(arg[1]) and IsBound(arg[1].Type) and arg[1].Type = "ZIdeal" then
    if arg[1].Generator = 0 then
      Error("m0 must not be zero");
    fi;
  else
    Error("use RayResidueRing(m0, [, minf])");
  fi;

  g := _gAssureZRCG(arg[1], arg[2]);

#  list := AbelianGroupBasis(g);
#  Apply(list, x-> AbelianGroupEltOrder(x));

  list := Copy(AbelianGroupSmithCreate(g).Information.Orders);

  return [AbelianGroupOrder(g), list];
end;

_RRR := function(arg)
local g, g_finite, g_infinite, minf_list;

  if Length(arg) = 1 then
    if IsRec(arg[1]) and IsBound(arg[1].Type) and arg[1].Type = "ZIdeal" then
      g := _RRRFinite(arg[1].Generator);
    else
      Error("use RayResidueRing(m0, [, minf])");
    fi;
  elif Length(arg) = 2 then
    if IsRec(arg[1]) and IsBound(arg[1].Type) and arg[1].Type = "ZIdeal" 
       and IsList(arg[2]) then
      g_finite := _RRRFinite(arg[1].Generator);
      minf_list := Set(arg[2]);
      if Length(minf_list) = 0 then
        g := g_finite;
      elif Length(minf_list) > 1 then
        Error("use RayResidueRing(m0, [, minf])");
      elif not 1 in minf_list then
        Error("use RayResidueRing(m0, [, minf])");
      else
        g_infinite := AbelianGroupCreate([[2]], "*");
        g_infinite.DiscreteLog := function(g, a)
          local expvec, elt;
            expvec := [0];
            if a < 0 then
              expvec := [1];
            fi;
            elt := AbelianGroupEltCreate(g, expvec);
            return elt;          
          end;
        
        g := g_finite * g_infinite;

        g.DiscreteExp := function(elt)
          local i, expvec, elt, elt2, a;
            g := elt.Group;
            expvec := [0];
            for i in [1..Length(elt.expvec[1])-1] do
              expvec[i] := elt.expvec[1][i];
            od;
            elt2 := AbelianGroupEltCreate(g.Information.DiscLog[1], expvec);
            a := AbelianGroupDiscreteExp(elt2); # Es muss gelten 0 < a < modul
            if ((elt.expvec[1][Length(elt.expvec[1])] mod 2) = 1) then
              a := a - g.Information.DiscLog[1].Information.ExpIntModul; 
              # Es muss -modul < a < 0 sein
            fi;
            return a;
          end;
      fi;
    else
      Error("use RayResidueRing(m0, [, minf])");
    fi;
  fi;

  if not IsBound(g.Information) then
    g.Information := rec();
  fi;
  if not IsBound(g.Information.RRR) then
    g.Information.RRR := rec();
  fi;
  g.Information.RRR.m0 := arg[1];
  if Length(arg) >1 then
    g.Information.RRR.minf := arg[2];
  else
    g.Information.RRR.minf := [];
  fi;

  g := AbelianGroupSmithCreate(g, true);
  return g;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#
#FXN       EltRayResidueRingRep
#SHORTDOC  Returns a representive of the class of an algebraic element in the 
#          multiplicative group of the ray residue ring. 
#LONGDOC   Calculates a representative of an algebraic element ${\tt elt}$
#          which is represented by a matrix $[a_0,\ldots,a_k]$. If
#          $\eta_0,\ldots,\eta_k$ is a basis of the residue ring as given by
#          {\tt RayResidueRingCyclicFactors}, ${\tt elt} \equiv
#          \eta_0^{a_0}\cdots\eta_k^{a_k} \text{mod}^* 
#          {({\goth{m}}_0 {\goth{m}}_\infty)}$ holds.
#
#          If the residue ring is not known yet, it will be computed. 
#          Infinite primes are represented in the same way as for
#          RayResidueRing. 
#          
#          This function uses a mixture of the discrete logarithm algorithms
#          developed in \cite{Pau1,Pau2} and \cite{CDOants,CDO1}.
#
#CALLING   r := EltRayResidueRingRep(elt,m0,minf);
#ARGUS     r @ matrix
#          elt @ algebraic element @ 
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes
#     
#CSRC      order\_ideal\_to\_ommm
#XREF      EltCon, RayResidueRing, RayResidueRingRepToElt, 
#          RayResidueRingCyclicFactors.
#KASH      @
#          O := OrderMaximal(Order(x^2-2*x-5));  
#          m0 := 27*O;;
#          minf := [2];;
#          L := RayResidueRingCyclicFactors(m0,minf);
#          elt := L[1][1];
#          EltRayResidueRingRep(elt,m0,minf);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

EltRayResidueRingRep := function(arg)
local g, arg2, a, elt, expvec;

  if Length(arg) > 3 or Length(arg) < 2 then
    Error("use EltRayResidueRingRep(elt, m0 [,minf])");
  fi;

  if Length(arg) = 2 then
    arg[3] := [];
  fi;

  if TYPE(arg[2]) = "KANT ideal" then
    return OrderEltRayResidueRingRep(arg[1], arg[2], arg[3]);
  fi;  

  a := arg[1];
  if Length(arg) = 2 then
    g := _gAssureZRRR(arg[2]);
  elif Length(arg) = 3 then
    g := _gAssureZRRR(arg[2], arg[3]);
  else 
    Error("use EltRayResidueRingRep(elt, m0, minf)");
  fi;

  elt := AbelianGroupDiscreteLog(g, a);
  expvec := elt.expvec;
  
  return expvec;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#
#FXN       RayResidueRingRepToElt
#SHORTDOC  Returns a canonical representative of an element of the 
#          multiplicative group of the residue class ring of a congruence
#          module.
#LONGDOC   This is the inverse to {\tt EltRayResidueRingRep}. The function will
#          compute the following product:
#          $$\prod|{i=1}^l a|i^{{\tt r}|i}$$
#          with $a|i$ the multiplicative basis for $(\o/{\goth{m}}_0)^*$ as
#          returned by {\tt RayResidueRingCyclicFactors}.
#CALLING   b := RayResidueRingRepToElt( r, m0, minf);
#ARGUS     b @ algebraic element
#          r @ matrix @ 
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes
#      
#CSRC      ommm\_to\_order\_elt
#PACKAGE     
#
#DEBUG     
#XREF      RayResidueRing, RayResidueRingCyclicFactors, 
#          EltRayResidueRingRep.
#KASH      @
#          O := OrderMaximal(Order(Poly(Zx,[1,3,-4,7])));;
#          m0 := 15*O;;
#          a := Elt(O,[7,1,8]);;
#          r := EltRayResidueRingRep(a,m0);
#          b := RayResidueRingRepToElt(r,m0);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayResidueRingRepToElt := function(arg)
local g, arg2, a, elt;

  if Length(arg) > 3 or Length(arg) < 2 then
    Error("use RayResidueRingRepToElt(matrix, m0 [,minf])");
  fi;

  if Length(arg) = 2 then
    arg[3] := [];
  fi;

  if TYPE(arg[2]) = "KANT ideal" then
    return OrderRayResidueRingRepToElt(arg[1], arg[2], arg[3]);
  fi;  

  elt := arg[1];
  if Length(arg) = 2 then
    g := _gAssureZRRR(arg[2]);
  elif Length(arg) = 3 then
    g := _gAssureZRRR(arg[2], arg[3]);
  else 
    Error("use RayResidueRingRepToElt(r, m0, minf)");
  fi;

  elt := AbelianGroupEltCreate(g, elt);
  a := AbelianGroupDiscreteExp(elt);
  
  return a;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#USERDEF   
#FXN       RayResidueRingCyclicFactors
#
#SHORTDOC  Returns generators (and their orders) for the multiplicative
#          group of the residue class ring of an ideal or a congruence module,
#          as computed by {\tt RayResidueRing}.
#       
#LONGDOC   A list containing generators and their orders is computed.
#          This structure can be used to solve multiplicative
#          congruences which may involve infinite places. 
#CALLING   L := RayResidueRingCyclicFactors(m0 [,minf]);
#ARGUS     L @ list
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes
#    
#PACKAGE   
#
#CSRC      order_ideal_rcr_mg_assure
#
#XREF      EltCon, IdealRayClassRep, OrderClassGroup, RayClassGroup.
#KASH      @
#          O := OrderMaximal(Order(Z,2,10));;
#          m0 := 7^3*O;; minf := [1,2];;
#          L := RayResidueRingCyclicFactors(m0,minf);
#          EltRayResidueRingRep(L[1][1],m0,minf);
#          EltCon(L[1][1]);
#          EltRayResidueRingRep(L[2][1]^2*L[3][1],m0,minf);
#          EltRayResidueRingRep(L[4][1],m0,minf);
#          L[4][1] mod m0;
#          EltRayResidueRingRep(L[2][1]^2*L[3][1]*L[4][1],m0,minf);          
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.8.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayResidueRingCyclicFactors := function(arg)
local g, list;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use RayResidueRingCyclicFactors(m0 [,minf])");
  fi;
    
  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if TYPE(arg[1]) = "KANT ideal" then
    return OrderRayResidueRingCyclicFactors(arg[1], arg[2]);
  fi;  

  g := _gAssureZRRR(arg[1], arg[2]);

  list := Copy(AbelianGroupCyclicFactors(g));

  return list;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#FXN       RayResidueRing
#
#SHORTDOC  This function computes the multiplicative group of the residue
#          class ring.
#
#LONGDOC   This function computes the multiplicative group of the residue
#          class ring of
#          an ideal or a congruence module
#          ${\goth{m}} = {\goth{m}}_0{\goth{m}}_\infty$, where
#          ${\goth{m}}_0$
#          is an integral ideal of a maximal                                
#          order $\O$ and ${\goth{m}}_\infty$ is a formal product of primes
#          at infinity.
#          ${\goth{m}}_\infty$ is represented by a list of integers, where 
#          each entry represents a real embedding (as given by {\tt 
#          EltCon}).
#          {\tt RayResidueRing} returns a list of the order of the
#          multiplicative group and a list
#          of the orders of the cyclic factors.
#
#          This structure can be used to solve multiplicative
#          congruences which may involve infinite places. This algorithm
#          bases an the presentation of the Einseinheiten in \cite{Hablau}
#          and is described in \cite{Pau1}.
#
#CALLING   L := RayResidueRing(m0 [,minf]);
#ARGUS     L @ list
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes
#       
#PACKAGE   
#
#CSRC      order\_ray\_ommm\_assure
#
#XREF      EltCon, EltRayResidueRingRep, RayResidueRingRepToElt 
#          RayResidueRingCyclicFactors.
#KASH      @
#          O := OrderMaximal(Order(Z,2,10));
#          P3 := Factor(3*O)[1][1];;
#          IdealDegree(P3);  
#          IdealRamIndex(P3);
#          m0 := P3^2;; L := RayResidueRing(m0);  
#          m0 := P3^4;; L := RayResidueRing(m0);
#          P5 := Factor(5*O)[1][1];;
#          IdealDegree(P5);
#          IdealRamIndex(P5);
#          m0 := P5^2;; L := RayResidueRing(m0);       
#          m0 := P5^4;; L := RayResidueRing(m0);
#          P7 := 7*O;; 
#          IdealDegree(P7);
#          IdealRamIndex(P7);      
#          m0 := P7^2;; RayResidueRing(m0);
#          m0 := P7^4;; RayResidueRing(m0);
#          m0 := P5^4*P7^2;; L := RayResidueRing(m0);
#          minf := [2];; L := RayResidueRing(1*O,minf);
#          L := RayResidueRing(m0,minf);
#          @ compute the ray residue ring of an ideal in $\Z$ @
#          a := ZIdealCreate(9);
#          RayResidueRing(a);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    8.6.97 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayResidueRing := function(arg)
local g, list;

  if Length(arg) > 2 or Length(arg) < 1 then
    Error("use RayResidueRing(m0 [,minf])");
  fi;

  if Length(arg) > 2 then
    Error("use RayResidueRing(m0, [, minf])"); 
  fi; 

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if IsOrder(arg[1]) then 
    arg[1] := 1*arg[1];
  fi;

  if TYPE(arg[1]) = "KANT ideal" then
    return OrderRayResidueRing(arg[1], arg[2]);
  fi;  
  
  if IsRec(arg[1]) and IsBound(arg[1].Type) and arg[1].Type = "ZIdeal" then
    if arg[1].Generator = 0 then
      Error("m0 must not be zero");
    fi;
  else
    Error("use RayResidueRing(m0, [, minf])");
  fi;

  g := _gAssureZRRR(arg[1], arg[2]);

#  list := AbelianGroupBasis(g);
#  Apply(list, x-> AbelianGroupEltOrder(x));

  list := Copy(AbelianGroupSmithCreate(g).Information.Orders);
  return [AbelianGroupOrder(g), list];
end;

_RRRFinite := function(m)
local i, fac_list, group_list, a, b, u, v, list, gen, g, group, x, g_new;

  if m = 1 then 
    g := AbelianGroupCreate([[1]], "*");
    g.Generators := [1];
    g.DiscreteLog := function(g, a)
      return AbelianGroupEltCreate(g, [0]);
    end;
    if not IsBound(g.Information) then
      g.Information := rec();
    fi;
    g.Information.ExpIntModul := m;

    return g;
  fi;

  fac_list := Factor(m);
  group_list := List(fac_list, x->_RRRFinitePrimePower(x[1], x[2]));
  
  # erzeuger aba"ndern f"ur DiscreteExp
  for i in [1..Length(fac_list)] do     
    group := group_list[i];
    a := fac_list[i][1]^fac_list[i][2];
    b := m/a;
    list := IntXGcd(a, b)[2];
    u := list[1];
    v := list[2];  # ua+vb = 1
    
    Apply(group.Generators, x->(u*a*1 + v*b*x) mod m);
  od;

  g := group_list[1];
  for i in [2..Length(group_list)] do
    g_new := g*group_list[i];
    g := g_new;
  od;

  if not IsBound(g.Information) then
    g.Information := rec();
  fi;

#  g := AbelianGroupSmithCreate(g, true);
#  Apply(g.Generators, x-> (x mod m));

  g.Information.ExpIntModul := m;
  return g;
end;

_RRRFinitePrimePower := function(p, v)
local ff, p_elt, g1, g2, g, order;

  ff := FF(p);
  p_elt := FFPrimitiveElt(ff);
  p_elt := MatMove(Mat(ff, [[p_elt]]), Z)[1][1];
  g1 := AbelianGroupCreate([p-1], "*");
  g1.Generators := [p_elt];

  g1.Information.DiscLog := ff;
  g1.DiscreteLog := function(g, a)
    local ff;

      ff := g1.Information.DiscLog;
      a := FFElt(ff, a);
      a := FFEltLog(a);
      a := AbelianGroupEltCreate(g, [a]);
      return a;
    end;

  if v=1 then return g1; fi;

  g2 := _SecondFactor(p, v);

  order := AbelianGroupExponent(g2);           # ??? exponent
  g1.Generators := [_IntFastPower(p_elt, order, p^v)];
  g := g1*g2;

  g.Information.DiscLog := [g1, g2, p^v]; 
  g.DiscreteLog := _PrimePowerDiscreteLog;

#  g := AbelianGroupSmithCreate(g, true);
#  Apply(g.Generators, x-> (x mod p^v));

  return g;
end;

_SecondFactor := function(p, v)
local g, h, A, B, g, M, g1, k, s, l, N, g1, g2;

  g := [];
  h := [];

  A := p;
  B := p^2;
  g := 1+A;
  M := B/A;

  g1 := AbelianGroupCreate([M], "*");
  g1.Generators := [g];
  if not IsBound(g1.Information) then
    g1.Information := rec();
  fi;
  g1.Information.DiscLog := [p, 2, v];
  g1.DiscreteLog := _SecondFactorDiscreteLog;

  k := 2;
  s := 2;

  while (k<>v) do
    l := Minimum(2^s, v);
    
    A := p^k;
    B := p^l;
    h := 1+A;
    N := B/A;
  
    g2 := AbelianGroupCreate([N], "*");
    g2.Generators := [h];

    g2.DiscreteLog := function(g, a)
      local expvec;

      expvec := [(a-1)/g.Information.DiscLog];
      return AbelianGroupEltCreate(g, expvec);
    end;
    g2.Information.DiscLog := p^k; # k ???
    g2.Information.Extension := p^l; 

    g := _AbelianGroupExtension(g2, g1); # passt Extension zu DiscLog ???
  
    Unbind(g1);
    Unbind(g2);

    g1 := g;
    if not IsBound(g1.Information) then
      g1.Information := rec();
    fi;
    g1.Information.DiscLog := [p, l, v];
    g1.DiscreteLog := _SecondFactorDiscreteLog;

    k := l;
    s := s+1;
  od;

#  g1 := AbelianGroupSmithCreate(g1);
#  Apply(g1.Generators, x-> (x mod p^v));

  g1.Information.ExpIntModul := p^v;
  return g1;
end;

_SecondFactorDiscreteLog := function(g, a)
local p, m, v, A, alpha, beta, gamma, l, elt, factor, list, k, s;

  alpha := [];

  list := g.Information.DiscLog;
  p := list[1];
  m := list[2];
  v := list[3];

  a := a mod p^v;

  A := p;
  beta := a-1;
  alpha := [beta/A];
  a := a * _IntFastPower(1+A, -alpha[1], p^v) mod p^v;  

  k := 2;
  s := 2;

  while (k<>m) do
    l := Minimum(2^s, m);
    
    A := p^k;
    beta := a-1;
    gamma := beta/A; 
  
    factor := 1;
    a := a * _IntFastPower(1+A, -gamma, p^v) mod p^v; 
        
    Add(alpha, gamma);
    k := l;
    s := s+1;
  od;

  elt := AbelianGroupEltCreate(g, alpha);
  return AbelianGroupEltReduce(elt);
end;

_DirectProductDiscreteLog := function(g, a)
local elt1, elt2, elt, expvec;

  elt1 := AbelianGroupDiscreteLog(g.Information.DiscLog[1], a);
  elt2 := AbelianGroupDiscreteLog(g.Information.DiscLog[2], a);
  expvec := MatToRowList(elt1.expvec)[1];
  Append(expvec, MatToRowList(elt2.expvec)[1]);
  elt := AbelianGroupEltCreate(g, expvec);

  return AbelianGroupEltReduce(elt);
end;

_PrimePowerDiscreteLog := function(g, a)
local g1, g2, elt1, elt2, elt, expvec;

  g1 := g.Information.DiscLog[1];
  g2 := g.Information.DiscLog[2];

  elt1 := AbelianGroupDiscreteLog(g1, a);
  a := a * _IntFastPower(g1.Generators[1], -elt1.expvec[1][1], 
                        g.Information.DiscLog[3]) mod g.Information.DiscLog[3];

  elt2 := AbelianGroupDiscreteLog(g2, a);
  expvec := MatToRowList(elt1.expvec)[1];
  Append(expvec, MatToRowList(elt2.expvec)[1]);
  elt := AbelianGroupEltCreate(g, expvec);

  return AbelianGroupEltReduce(elt);
end;

RConductor := function(g)
local i, j, fac, sub, m0, minf, sub_new, weiter, n, h, k, m, gens, gens_new;

  if IsBound(g.Information) and IsBound(g.Information.RCG) then
    m0 := Min(g.Information.RCG.m0);
    minf := g.Information.RCG.minf;
  else
    Error("group has to be ray class group or quotient of a ray class group");
  fi;

  if IsBound(g.Conductor) then
    return g.Conductor;
  fi;

  if IsBound(g.Information.RCG.Sub) then
    sub := g.Information.RCG.Sub;
    gens := AbelianGroupGenerators(sub, true);                      # langsam
  fi;

  n := AbelianGroupOrder(g);
  fac := Factor(m0);

  for i in [1..Length(fac)] do
    weiter := true;
    j := 1;
    while weiter and j <= fac[i][2] do
      m0 := m0/fac[i][1];
      h := _RCG(ZIdealCreate(m0), minf);
      if IsBound(sub) then # g is quotient of a ray class group
        gens_new := List(gens, x-> AbelianGroupDiscreteLog(h, x));  # langsam
        sub_new := AbelianSubGroup(gens_new);              
        h := AbelianQuotientGroup(h, sub_new);
      fi;
      k := AbelianGroupOrder(h);
      GASMAN("g");
      weiter := (k = n);
      j := j+1;
    od;
    if not weiter then
      m0 := m0 * fac[i][1];
    fi;  
  od;
  
  m0 := ZIdealCreate(m0);

  if Length(minf) > 0 then
    h := _RCG(m0);
    k := AbelianGroupOrder(h);
    if k = n then 
      minf := [];
    fi;
  fi;

  m := [m0, minf];
  g.Conductor := m;
  return Copy(m);
end;

_gAssureZRCG := function(arg)
local rcg, pos;

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if IsRec(arg[1]) and IsBound(arg[1].Type) and 
      arg[1].Type = "ZIdeal" then
    if TYPE(arg[2])  = "list" then
      arg[2] := Set(arg[2]);
      if not arg[2] in [[1], []] then
        Error("use RayClassGroup(m0 [, minf])");
      fi;
    fi;
  else 
    Error("use RayClassGroup(m0 [, minf])");
  fi;

  _gAssureZRRR(arg[1], arg[2]);

  pos := _FindZRay(arg[1], arg[2]);
  if not IsBound(_ZRayInfo[pos].rcg) then
    rcg := _RCG(arg[1], arg[2]); 
    _ZRayInfo[pos].rcg := rcg; 
  else
    rcg := _ZRayInfo[pos].rcg;
  fi;

  return rcg;
end;

_gAssureZRRR := function(arg)
local rrr, pos, entry;

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if IsRec(arg[1]) and IsBound(arg[1].Type) and  
      arg[1].Type = "ZIdeal" then
    if TYPE(arg[2]) = "list" then
      arg[2] := Set(arg[2]);
      if not arg[2] in [[1], []] then
        Error("use RayResidueRing(m0 [, minf])");
      fi;
    fi;
  else 
    Error("use RayResidueRing(m0 [,minf])");
  fi;

  pos := _FindZRay(arg[1], arg[2]);
  if pos = false then
    rrr := _RRR(arg[1], arg[2]);
    entry := rec(
      m0   := arg[1],
      minf := arg[2],
      rrr  := rrr
    );
    Add(_ZRayInfo, entry);
  else 
    rrr := _ZRayInfo[pos].rrr;
  fi;

  return rrr;
end;

###### save information in _ZRayInfo ######

# _ZRayInfo is List of records:
# rec(
#   m0,           finite place
#   minf,         infinite place
#   rrr,          ray residue ring
#   rcg           ray class group
#  );
#
# m0, minf, rrr are always set
# if the conductor is known, you can find it in rcg.Conductor


#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#USERDEF
#FXN       RayConductor
#
#SHORTDOC  Calculates the conductor of the ray class group modulo a congruence 
#          module.
#LONGDOC   Calculates the minimal ideal and the minimal set of infinite primes
#          with the same ray class group as the given congruence module. The 
#          conductor is represented by a list of an ideal and a list of primes
#          at infinity --- the numbers correspond to the real embeddings of the 
#          order.
#
#          This is done by a careful analysis of the structure of the ray class
#          group as described in \cite{Pau1,Pau2}.
#CALLING   L := RayConductor(m0 [, minf] [, rels]);
#ARGUS     L @ list
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes 
#          rels @ matrix @ relation matrix over $\Z$           
#       
#PACKAGE
#
#CSRC      order_ray_conductor
#
#XREF      EltCon, EltRayResidueRingRep, RayResidueRing, RayClassGroup.
#KASH      @
#          O := OrderMaximal(Order(x^3+9*x^2-8*x-9));;
#          m0 := 94*O;;
#          minf:= [1,2];;
#          L := RayConductor(m0,minf);
#          OrderClassGroup(O,500,euler,fast);
#          RayClassGroup(m0,minf);
#          RayClassGroup(L[1],L[2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    13.9.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayConductor := function(arg)
local g, G, rG, L, U, m, s;

  if Length(arg) > 3 or Length(arg) < 1 then
    Error("use RayConductor(m0 [,minf] [, rels])");
  fi;

  if Length(arg) > 3 then
    Error("use RayConductorTest(m0 [, minf] [, rels])");
  fi;

  if Length(arg) = 1 and not(IsRec(arg[1])) then
    arg[2] := [];
  fi;

  if Length(arg)=1 and IsRecType(arg[1], "AbelianGroup") then
      G:=arg[1];  
    if IsBound(G.Conductor) then 
       return G.Conductor; 
    elif IsBound(G.Information.RCG) and IsIdeal(G.Information.RCG.m0) then
       m:=OrderRayConductor(G.Information.RCG.m0,G.Information.RCG.minf, 
                                                               G.Relations); 
       G.Conductor:=m;  
       return m;
    elif IsBound(G.Information.RCG) and not IsIdeal(G.Information.RCG.m0) then
       return RConductor(G);
    else 
      U:=Copy(G); 

      if IsBound(G.QuotientsFrom) then U:=Copy(U.subgroup); fi;
            
      while IsBound(U.parentgroup) do U:=Copy(U.parentgroup); od;
      if IsBound(U.Information.RCG) then
        L:=List(AbelianGroupGenerators(G),z->AbelianGroupDiscreteExp(z));
        rG:=RayClassGroupToAbelianGroup(U.Information.RCG.m0,
                                        U.Information.RCG.minf);
        L:=List(L,z->AbelianGroupDiscreteLog(rG,z));
        L:=List(L,z->MatToRowList(z.expvec)[1]);
        L:=Mat(Z,L);
             
        m:=OrderRayConductor(U.Information.RCG.m0,U.Information.RCG.minf,L);
        G.Conductor:=m;
        return m;
      else 
        return Error("<usage: subgroup or quotientgroup of a abstract ray class group>");
      fi;  
        
    fi;
    

  
  fi; 


  if TYPE(arg[1]) = "KANT ideal" then
    if Length(arg)=3 then
      return OrderRayConductor(arg[1], arg[2], arg[3]);
    else
      return OrderRayConductor(arg[1], arg[2]);
    fi;
  fi;  

  if IsRecType(arg[1], "ZIdeal") and arg[1].Generator <> 0 then
    if Length(arg)>1 and IsList(arg[2]) then
      arg[2] := Set(arg[2]);
      if not arg[2] in [[1], []] then
        Error("use RayConductor(m0 [, minf] [, rels])");
      fi;
    fi;
  else 
    Error("use RayConductor(m0 [, minf] [, rels])");
  fi;
  
  if Length(arg)>1 and IsList(arg[2]) then
    g := _gAssureZRCG(arg[1], arg[2]);
  else
    g := _gAssureZRCG(arg[1], []);
  fi;
  if Length(arg) = 3 then
    s := AbelianSubGroup(g, arg[3]);
    g := AbelianQuotientGroup(g, s);
  elif Length(arg)>1 and TYPE(arg[2]) = "KANT matrix" then
    s := AbelianSubGroup(g, arg[2]);
    g := AbelianQuotientGroup(g, s);
  fi;
  m := RConductor(g);
  
  return Copy(m);
end;

_FindZRay := function(arg)
local i, weiter, entry, h, m0, minf;

  m0 := arg[1];
  minf := arg[2];

  weiter := true;
  i := 1;
  while weiter and i <= Length(_ZRayInfo) do
    h := _ZRayInfo[i];
    if h.m0 = m0 and h.minf = minf then
      entry := h;
      weiter := false;
    fi;
    i := i+1;
  od;

  if weiter then
    i := false;
  else
    i := i-1;
  fi;
  
  return i;
end;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
#USERDEF
#FXN       RayConductorTest
#
#SHORTDOC  Tests iff the given module is the true conductor.
#CALLING   b := RayConductorTest(m0 [, minf] [, rels]);
#ARGUS     b @ boolean
#          m0 @ ideal @
#          minf @ list @ of integers/infinite primes
#          rels @ matrix @ relation matrix over $\Z$
#   
#PACKAGE
#
#CSRC      order_ray_conductor_test
#
#XREF      EltCon, EltRayResidueRingRep, RayResidueRing, RayClassGroup.
#KASH      @
#          O := OrderMaximal(Order(x^3+9*x^2-8*x-9));;
#          m0 := 94*O;;
#          minf := [1,2];;
#          RayConductorTest(m0,minf);
#          L := RayConductor(m0,minf);
#          OrderClassGroup(O,500,euler,fast);
#          RayClassGroup(m0,minf);
#          RayClassGroup(L[1],L[2]);
#
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    13.9.96 SP written
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

RayConductorTest := function(arg)
local L1, L2, mat;
# rels have to be given corresponding to the representation
# of the ray class group of the congruence module m0, minf !!!

  if Length(arg) > 3 or Length(arg) < 1 then
    Error("use RayConductorTest(m0 [, minf] [, rels])");
  fi;

  if Length(arg) = 1 then
    arg[2] := [];
  fi;

  if TYPE(arg[1]) = "KANT ideal" then
    if Length(arg)=3 then
      return OrderRayConductorTest(arg[1], arg[2], arg[3]);
    else
      return OrderRayConductorTest(arg[1], arg[2]);
    fi;
  fi;  

  if IsRec(arg[1]) and IsBound(arg[1].Type) and 
      arg[1].Type = "ZIdeal" and arg[1].Generator <> 0 then
    if IsList(arg[2]) then
      arg[2] := Set(arg[2]);
      if not arg[2] in [[1], []] then
        Error("use RayConductorTest(m0 [, minf] [, rels])");
      fi;
    fi;
  else 
    Error("use RayConductorTest(m0 [, minf] [, rels])");
  fi;

  if Length(arg) = 3 then  
    L1 := RayConductor(arg[1], arg[2], arg[3]);
  else
    L1 := RayConductor(arg[1], arg[2]);
  fi;

  L2 := [];
  L2[1] := arg[1];
  if IsList(arg[2]) then
    L2[2] := arg[2];
  else
    L2[2] := [];
  fi;

return L1 = L2;
end;


###### structure ZIdeal #####
_ZIdealPrint      := Ignore;
_ZIdealMult       := Ignore;
_ZIdealDiv        := Ignore;
_ZIdealPower      := Ignore;
_ZIdealAdd        := Ignore;
_ZIdealOperations := Ignore;

#/*TEX**TEX|**TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX***TEX**
# 
#FXN       ZIdealCreate
#SHORTDOC  Returns the ideal in $\Z$ generated by the given integer.
#
#CALLING   I := IdealRayClassRep(a);
#ARGUS     I @ ideal in $\Z$
#          a @ integer 
#CSRC      
#XREF      Ideal
#KASH      @
#          I := ZIdealCreate(6);
#          
#**HISTORY*|*HISTORY***HISTORY***HISTORY***HISTORY***HISTORY***HISTORY**HISTORY
#
#    HB 18.09.98 written 
#
#**KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH***KASH*/
##

ZIdealCreate := function(arg)
local id, a;

  a := arg[1];
  if Length(arg) <> 1 or not IsInt(arg[1]) then
    Error("use ZIdealCreate(int)");
  fi;

  id := rec(
            Type       := "ZIdeal",
            operations := _ZIdealOperations,
            Generator  := Abs(a)
           );
  return id;
end;

_ZIdealPrint := function(id)
  Print("<");
  Print(id.Generator);
  Print(">");
end;

_ZIdealMult := function(id1, id2)
local gen;

  if IsRecType(id2, "ZIdeal") then
    gen := id1;
    id1 := id2;
    id2 := gen;
  fi;
  if IsRecType(id1, "ZIdeal") then
    if IsInt(id2) then
      return ZIdealCreate(id1.Generator*id2);
    elif IsRecType(id2, "ZIdeal") then
      return ZIdealCreate(id1.Generator*id2.Generator);
    elif IsOrder(id2) then
      return id2*id1.Generator;
    elif IsIdeal(id2) then
     return id2*id1.Generator; 
    else
      Error("use _ZIdealMult(ideal, ideal) or _ZIdealMult(ideal, int)");
    fi;    
  fi; 
end;

_ZIdealDiv := function(id1, id2)
local gen;

  if not IsRecType(id1, "ZIdeal") or
     not IsRecType(id2, "ZIdeal") then
    Error("use _ZIdealDiv(ideal, ideal)");
  fi; 

  gen := id1.Generator / id2.Generator;
  if not IsInt(gen) then
    Error("not defined");
  fi;
  return ZIdealCreate(gen);
end;

_ZIdealAdd := function(id1, id2)
local gen;

  if not IsRec(id1) or
     not IsBound(id1.Type) or 
     not id1.Type = "ZIdeal" or 
     not IsRec(id2) or
     not IsBound(id2.Type) or  
     not id2.Type = "ZIdeal" then 
    Error("use _ZIdealAdd(ideal, ideal)");
  fi; 

  gen := IntGcd(id1.Generator, id2.Generator);
  return ZIdealCreate(gen);
end;

_ZIdealPower := function(id, a)
local gen;

  if not IsRec(id) or
     not IsBound(id.Type) or 
     not id.Type = "ZIdeal" or 
     not IsInt(a) then
    Error("use _ZIdealPower(ideal, int)");
  fi; 

  gen := id.Generator^a;
  return ZIdealCreate(gen);
end;

_ZIdealCompare := function(id1, id2)
  if not IsRecType(id1, "ZIdeal") or
     not IsRecType(id2, "ZIdeal") then
    return false;
  fi; 

  return id1.Generator = id2.Generator;
end;
_ZIdealLess := function(id1, id2)
  if not IsRecType(id1, "ZIdeal") then
    return false;
  elif not IsRecType(id2, "ZIdeal") then
    return true;
  fi; 

  return id1.Generator < id2.Generator;
end;


# muss am ende stehen
_ZIdealOperations := rec( 
                         Print := _ZIdealPrint,
                         \*    := _ZIdealMult,
                         \^    := _ZIdealPower,
                         \+    := _ZIdealAdd,
                         \-    := _ZIdealAdd,
                         \/    := _ZIdealDiv,
                         \=    := _ZIdealCompare,
                         \<    := _ZIdealLess
                        );










