String := function(arg)
  return "";
end;
PrintRec := function(arg)
  return "";
end;
#############################################################################
##
#A  string.g                    GAP library                      Frank Celler
##
#A  @(#)$Id: string.g,v 3.22.1.1 1995/03/15 10:13:08 sam Rel $
##
#Y  Copyright 1990-1992,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
##
##  This file contains the string functions.
##
#############################################################################
##
#F  StringInt( <n> )  . . . . . . . . . . . convert integer <n> into a string
##
StringInt := function ( n )
    local  str,  num,  digits, pr;

    pr := Prec();
    Prec(Trunc(Log(Abs(n)+1)/Log(10))+4);
    # construct the string without sign
    num := Abs( n );
    digits := [ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' ];
    str := "";
    repeat
        Add( str, digits[num mod 10 + 1] );
        num := Trunc(num/10);
    until num = 0;

    # add the sign and return
    if n < 0  then
        Add( str, '-' );
    fi;
    str := str{[Length(str),Length(str)-1 .. 1]};
    IsString(str);
    Prec(pr);
    return str;
end;


#############################################################################
##
#F  StringRat( <rat> )  . . . . . . . . . .  convert a rational into a string
##
StringRat := function ( rat )
    local   str;

    str := StringInt( Numerator(rat) );
    if Denominator(rat) <> 1  then
        str := Concatenation( str, "/", StringInt(Denominator(rat)) );
    fi;

    return str;
end;


#############################################################################
##
#F  StringBool( <bool> )  . . . . . . . . . . . convert boolean into a string
##
StringBool := function( bool )

    if bool  then
        return "true";
    else
        return "false";
    fi;

end;


#############################################################################
##
#F  StringList( <list> )  . . . . . . . . . . . .  convert list into a string
##
StringList := function( list )
    local   str, i;

    str := "[ ";
    for i in [ 1 .. Length( list ) ]  do
        if IsBound( list[ i ] )  then
            str := Concatenation( str, String( list[ i ] ) );
        fi;
        if i <> Length( list )  then
            str := Concatenation( str, ", " );
        fi;
    od;
    return Concatenation( str, " ]" );

end;


#############################################################################
##
#F  StringRec( <rec> )  . . . . . . . . . . . .  convert record into a string
##
StringRec := function( record )
    local   str,  nam,  com;

    if IsBound(record.operations) and IsBound(record.operations.String)  then
        str := record.operations.String( record );
    else
        str := "rec( ";
        com := false;
        for nam  in RecFields( record )  do
            if com  then
                str := Concatenation( str, ", " );
            else
                com := true;
            fi;
            str := Concatenation( str, nam, " := ",
                                        String( record.(nam) ) );
        od;
        str := Concatenation( str, " )" );
    fi;
    return str;

end;


#############################################################################
##
#F  String( <obj>, <width> ) . . . . . . . . . . convert object into a string
##
##  Main dispatcher for 'StringType( <obj> )', where <Type> is  the  type  of
##  <obj>.
##
String := function( arg )
    local   str, blanks, obj, fill;

    # Check the arguments.
    if Length( arg ) < 1
       or Length( arg ) > 2
       or ( Length( arg ) = 2 and not IsInt( arg[ 2 ] ) )
    then
        Error( "usage: String( <obj> ) or String( <obj>, <width> )" );
    fi;
    obj := arg[ 1 ];

    # Dispatcher.
    if   IsInt( obj )     then str := StringInt   ( obj );
    elif IsRat( obj )     then str := StringRat   ( obj );
    elif IsBool( obj )    then str := StringBool  ( obj );
    elif IsString( obj )  then str := obj;
    elif IsList( obj )    then str := StringList  ( obj );
    elif IsRec( obj )     then str := StringRec   ( obj );
    else Error( "String: cannot convert ", obj, " into a string" );
    fi;

    # If no <width> is given or it is to small, return.
    if Length( arg ) = 1 or String( str ) >= Abs( arg[ 2 ] )  then
        return str;
    fi;

    # If <width> is positive, blanks are filled in from the left.
    blanks := "                                                 ";
    if arg[ 2 ] > 0  then
        fill := arg[ 2 ] - String( str );
        while fill > 0  do
            if fill >= String( blanks )  then
                str := Concatenation( blanks, str );
            else
                str := Concatenation( blanks{[1..fill]},str );
            fi;
            fill := arg[ 2 ] - Length( str );
        od;
    else
        fill :=  - arg[ 2 ] - Length( str );
        while fill > 0  do
            if fill >= Length( blanks )  then
                str := Concatenation( str, blanks );
            else
                str := Concatenation( str, blanks{[1..fill]} );
            fi;
            fill :=  - arg[ 2 ] - String( str );
        od;
    fi;
    IsString(str);
    return str;

end;


#############################################################################
##
#F  PrintArray( <array> ) . . . . . . . . . . . . . . . . pretty print matrix
##
PrintArray := function( array )
    local   arr,  max,  l,  k;

    if array = [[]]  then
        Print( "[ [ ] ]\n" );
    elif array = []  then
        Print( "[ ]\n" );
    elif not ForAll( array, IsList )  then
        arr := List( array, x -> String( x ) );
        max := Maximum( List( arr, String ) );
        Print( "[ ", String( arr[ 1 ], max + 1 ) );
        for l  in [ 2 .. Length( arr ) ]  do
            Print( ", ", String( arr[ l ], max + 1 ) );
        od;
        Print( " ]\n" );
    else
        arr := List( array, x -> List( x, String ) );
        max := Maximum( List( arr, x -> Maximum( List(x,String) ) ) );
        Print( "[ " );
        for l  in [ 1 .. Length( arr ) ]  do
            if l > 1  then
                Print( "  " );
            fi;
            Print( "[ " );
            for k  in [ 1 .. Length( arr[ l ] ) ]  do
                Print( String( arr[ l ][ k ], max + 1 ) );
                if k = Length( arr[ l ] )  then
                    Print( " ]" );
                else
                    Print( ", " );
                fi;
            od;
            if l = Length( arr )  then
                Print( " ]\n" );
            else
                Print( ",\n" );
            fi;
        od;
    fi;

end;


#############################################################################
##
#F  PrintRec(<record>)  . . . . . . . . . . . . . . . . . . .  print a record
##
##  'PrintRec' must  call 'Print'  so that 'Print'   assigns  the record   to
##  '~' and  prints for  example 'rec( a := ~  )'  in this  form and does not
##  go into an  infinite loop 'rec( a  := rec(  a := ...'.   To make  'Print'
##  do the right   thing, we  assign to '<record>.operation'  the  operations
##  record 'RecordOps', which contains the appropriate 'Print' function.
##
PrintRecIgnore := [ "operations", "parent" ];

PrintRecIndent := "  ";

RecordOps := rec();

RecordOps.Print := function ( record )
    local  len, i, nam, lst, printRecIndent;
    len  := 0;
    for nam in RecFields( record )  do
        if len < Length( nam )  then
            len := Length( nam );
        fi;
        lst := nam;
    od;
    Print( "rec(\n" );
    for nam  in RecFields( record )  do
        if not nam in PrintRecIgnore then
            if not IsRec( record.(nam) )  then
                Print( PrintRecIndent, nam );
                for i  in [Length(nam)..len]  do
                    Print( " " );
                od;
                Print( ":= ", record.(nam) );
                if nam <> lst  then  Print( ",\n" );  fi;
            else
                Print( PrintRecIndent, nam );
                for i  in [Length(nam)..len]  do
                    Print( " " );
                od;
                Print( ":= " );
                printRecIndent := PrintRecIndent;
                PrintRecIndent := Concatenation(PrintRecIndent,"  ");
                PrintRec( record.(nam) );
                if nam <> lst  then Print( ",\n" );  fi;
                PrintRecIndent := printRecIndent;
            fi;
        else
            Print( PrintRecIndent, nam );
            for i  in [Length(nam)..len]  do
                Print( " " );
            od;
            Print( ":= ..." );
            if nam <> lst  then Print( ",\n" );  fi;
        fi;
    od;
    Print( " )" );
end;

PrintRec := function ( record )
    local   operations;
    if IsBound( record.operations )  then
        operations := record.operations;
    fi;
    record.operations := RecordOps;
    Print( record );
    if IsBound( operations )  then
        record.operations := operations;
    else
        Unbind( record.operations );
    fi;
end;

