
ZerosToPolList:=function(L)
   local LL, i, Fqx, len;
   len:=Length(L);
   Fqx:=PolyAlg(FFEltFF(L[1]));
   LL:=[];
   for i in [1..len] do
      Add(LL, Poly(Fqx,[1,-L[i]]));
   od;
   
   return LL;
end;


FrobeniusSort:=function(l, p)
   local L, Fq, len, i, a;

   Fq:=FFEltFF(l[1]);
   len:=Length(l);
   if Length(l)<= 2 then return l; fi;
   a:=l[1];
   L:=[a];
   for i in [2..len] do
      a:=a^p;
      if a in l then 
         Add(L,a);
      else 
         Error("Frobenius sort: element not found\n");
      fi;
   od;
   return L;
end;

CalcZeros:=function(arg)
   local  o, p, k, f, CalcDeg, F, Fp, Fx, Fpx, fact_f, l, i;

   CalcDeg:=function(f, p)
      local   F, Fx, L, k;

      F:=FFCreate(o, p,1);
      Fx:=PolyAlg(F);
      L:=Factor(f, p);
      L:=List(L,x->PolyDeg(x[1]));
      k:=Iterated(L,IntLcm);
      return k;
   end;

   o:=arg[1];
   f:=arg[2];
   p:=arg[3];
   if Length(arg)=4 then
      k:=arg[4];
   else
      k:=CalcDeg(f,p);
   fi;

   if k=1 then
      F:=FFCreate(o, p,1);
      Fp:=F;
   else
      Fp:=FFCreate(o, p,1);
      F:=FFCreate(o, p,k);
   fi;
   Fpx:=PolyAlg(Fp);
   Fx:=PolyAlg(F);
   f:=PolyMove(f,Fpx);
   fact_f := Factor(f);
   if k>1 then
      fact_f:=List(fact_f,x->Factor(PolyMove(x[1],Fx)));
      l := [];
      for i in [1..Length(fact_f)] do
         fact_f[i]:=List(fact_f[i],x->x[1]);
         fact_f[i]:=List(fact_f[i],x->-PolyToList(x)[2]);
         Append(l, FrobeniusSort(fact_f[i],p));
      od;
      fact_f:=l;
   fi;
   return fact_f;
end;

BlockSystem := function(arg)
   local  CalcDeg,O, o, p, k,pos, i, block, f, g, h, Fx, fact_f, fact_g, eq_O, eq_o, len, l;


   O:=arg[1];
   o:=arg[2];
   p:=arg[3];
   if Length(arg) = 4 then
      k:=arg[4];
   fi;
   if IsBound(k) and IsList(k) then
      fact_f:=k;
   else
      f:=OrderPoly(O);
      if IsBound(k) then
         fact_f := CalcZeros(O,f,p,k);
      else
         fact_f := CalcZeros(O,f,p);
      fi; 
   fi;
   Fx:=PolyAlg(FFEltFF(fact_f[1]));
   g:=PolyMove(OrderPoly(Zx,o),Fx);
   fact_g := Factor(g);
   Apply(fact_g,x->x[1]);
   Apply(fact_g,x->-PolyToList(x)[2]);

   h:=PolyMove(Poly(Zx, Reversed(EltToList(EltMove(OrderBasis(o)[2],O) mod p))),Fx);
   len :=Length(fact_g);
   block:=List([1..len],x->[]);
   len:=Length(fact_f);
   for i in [1..len] do
      pos:=Position(fact_g, Eval(h, fact_f[i]));
      if pos = false then return false; fi;
      Add(block[pos],i);
   od;
   return block;
end;

BlockIsSubfield := function (O, o1, o2)
   local PositionSubList, FindPrime, p, L1, L2, f, g, h, l1, a, l2, i;
 
   PositionSubList:=function(a, L)
      local i , len;
      len:=Length(L);
      for i in [1..len] do
         if a in L[i] then 
            return i;
         fi;
      od;
   end;


   FindPrime:=function(o1, o2, o3)
      local disc1, disc2, disc3, p;

      p:=2;
      disc1:=OrderDisc(o1);
      disc2:=OrderDisc(o2);
      disc3:=OrderDisc(o3);

      while true do
         p:=NextPrime(p);
         if (disc1 mod p <> 0) and (disc2 mod p <>0) and (disc3 mod p <>0) then 
            return p;
         fi;
      od;
   end;
   
   if (OrderDeg(o1)<=OrderDeg(o2)) or (OrderDeg(o1) mod OrderDeg(o2) <> 0) then
      return false;
   fi;


   p:=FindPrime(O,o1,o2);
   L1:=BlockSystem(O,o1,p);
   L2:=BlockSystem(O,o2,p);
   l1:=L1[1];
   i:=PositionSubList(l1[1],L2);
   l2:=L2[i];
   for a in l1 do
      if not a in l2 then
         return false;
      fi;
   od;
   return true;
end;

ReduceSubfieldList:=function(o,L)
   local i, len, deg, j, M, test;

   M:=[];
   len:=Length(L);
   for i in L do
      if not ForAny(L, x->BlockIsSubfield(o,x,i)) then
         Add(M, i);
      fi;
   od;
   return M;
end;


RelativePoly := function(O,o)
   local  f, h, a, b, eq_o, eq_ox;

   eq_o := OrderEquationOrder(o);
   eq_ox:=PolyAlg(eq_o);
   a:=OrderBasis(eq_o)[2];
   b:=EltMove(a,O);
   h:=Reversed(EltToList(b));
   h:=Poly(eq_ox,h)-a;
   f:=PolyMove(OrderPoly(Zx,O),eq_ox);

   return PolyGcd(f,h);
end;


ShortFieldList:=function(L)
   local  LL, a, b;

   LL:=[];
   for a in L do
      repeat
         b:=OrderShort(a);
         if b <> false then
            a:=b;
         fi;
      until b=false;
      Add(LL, a);
   od;
   return LL;
end;



PolyDecompose:=function(arg)

   local    short, recurse, pol, o, sub, a, g, h, L, l, LL, b;

   pol :=arg[1];
   if Length(arg)=1 or arg[2]=false then
      recurse:=false;
   else
      recurse:=true;
   fi;

   if Length(arg)<3 or arg[3]=false then
      short:=false;
   else
      short:=true;
   fi;


   if IsPrime(PolyDeg(pol)) then return []; fi;
   o:=Order(pol);
   sub:=OrderSubfield(o);
   sub:=Reversed(sub);
   if recurse then
      sub:=ReduceSubfieldList(o,sub);
   fi;
   L:=[];

   if short then
      sub:=ShortFieldList(sub);
   fi;

   for a in sub do
      g:=OrderPoly(Zx,a);
      h:=RelativePoly(o,a);
      if recurse then
         LL:=PolyDecompose(g, recurse, short);
      else
         LL:=[];
      fi;
      for b in LL do
         l:=Flat([b,h]);
         Add(L,l);
      od;
      if Length(LL) =0 then       
         l:=[g,h];
         Add(L,l);
      fi;
   od;
   return L;
end;

PolyRootsPQ := function(f)
   local p, q, L, p_2;

   if PolyDeg(f)<>2 then return false;fi;

   L:=PolyToList(f);
   p:=L[2];
   q:=L[3];
   p_2:=p/2;
   Print("Roots: ",-p_2," +- Sqrt(",p_2^2-q,")\n");
   return true;
end;


MakeRelative:=function(L)

   local  i, o, a, LL, l, b, len;

   LL:=[];
   for a in L do
      if Length(a) <= 2 then
         Add(LL,a);
      else
         l:=[];
         for b in a do
            Add(l, PolyAlgCoef(PolyAlg(b)));
         od;
         len:=Length(l);
         l[1]:=a[1];
         l[2]:=a[2];
         for i in [3..len] do
            o:=Order(l[i-1]);
            OrderInstallHom(l[i],o,OrderBasis(o)[2]);
            l[i]:=PolyMove(a[i],o);
         od;
         Add(LL,l);
      fi;
   od;
   return LL;
end;



