OrderCoefIdealList := function(O)
local L; 
 
  if OrderCoefOrder(O) = Z then
    L := [1..OrderDeg(O)];
    Apply (L, x-> 1);
  else
    L := OrderCoefIdeals(O);
  fi;
  return L;
end;

OrderIsAbs := function(O)
  return OrderCoefOrder(O) = Z;
end;
#######################################

OrderTraceMat := function(O)
local L,x;

  L := List(OrderBasis(O,O));
  L := Mat(O, [L]);
  L := MatTrans(L)*L;
  L := MatToRowList(L);
  Apply(L, x->List(x, EltTrace));


  return Mat(OrderCoefOrder(O), L);
end;
#######################################


MatAppendRight := function (CR, M1, M2)
local L1, L2;

  L1 := MatToColList(M1);
  L2 := MatToColList(M2);
  Append(L1, L2);

  IsMat(L1);
  return MatTrans(L1);

end;
#######################################


MatAppendLower := function (CR, M1, M2)

  IsMat(M1); IsMat(M2);
  return MatTrans(MatAppendRight(CR, MatTrans(M1), MatTrans(M2)));
end;

MatRightPart := function (CR, M, cols)
local L1;

  L1 := MatToColList(M);
  L1 := L1{[Length(L1)-cols+1..Length(L1)]};
  IsMat(L1);

  return MatTrans(L1);
end;
#######################################


MatLowerPart := function(CR, M, rows)

return MatTrans(MatRightPart(CR, MatTrans(M), rows));
end;
#######################################


ArithRadical := function (O)
local disc, IdealList, M, Mod, Rad, deg;

  disc := OrderDisc(O);
  deg := OrderDeg(O);

  IdealList := OrderCoefIdealList(O);

  M := OrderTraceMat(O);

  if OrderIsAbs(O) then
    M := MatAppendRight(OrderCoefOrder(O), M,
                        disc*MatId(OrderCoefOrder(O), OrderDeg(O)));
  else
    M := MatAppendRight(OrderCoefOrder(O), M,
                        MatId(OrderCoefOrder(O), OrderDeg(O)));
  fi;

  if OrderIsAbs(O) then
    M := MatHermiteColUpper(M);
    M := MatRightPart(OrderCoefOrder(O), M, deg);

    M := MatTrans(M);
    M := MatInv(M);

    Rad := Ideal(O, M, disc);
  fi;


  return Rad;
end;
#######################################

MatZDen := function(M)
local L, den;
  L:=Flat(MatToRowList(M));
  L:=List(L,Denominator);
  den := Iterated(L,IntLcm);

  return den;
end;
      
#######################################

RingOfMultiplikators := function (I)
local L, M, O, S, i, LBasis, deg, den;

  O := IdealOrder(I);
  deg := OrderDeg(O);
  S := IdealBasis(I)[2];
  LBasis := MatToColList(S);
  Apply(LBasis, x-> Elt(O, x));
  S := MatInv(S);
  S := IdealBasis(I)[1]*S;
  #Print("S= ",S,"\n\n");

  den := MatZDen(S);
  S := S*den;
  S := MatMove(S, Z);
  #Print("S= ",S,"\n\n");

  
  M := S*EltRepMat(LBasis[1])[2];
  
  for i in LBasis{[2..Length(LBasis)]} do
    if EltRepMat(i)[1] <> 1 then
      Print("Error, RepMat of ",i,"th Basis Elt is not integral!!\n");
    fi;
    M := MatAppendLower (OrderCoefOrder(O), M, S*EltRepMat(i)[2]);
  od;

  M := MatTrans(M);
  
  #Print("M= ",M,"\n\n");
  M := MatHermiteColUpper(M);
  M := MatRightPart(OrderCoefOrder(O), M, deg);
  M := MatTrans(M);
  #Print("M= ",M,"\n\n");
  
  M := MatInv(M);
  #Print("M= ",M,"\n\n");
  M := M*den;
  #Print("M= ",M,"\n\n");
  den := MatZDen(M);
  M := den*M;
  M := MatMove(M, Z);
  #Print("M= ",M,"\n\n");

  O := Order(O, M, den);

  return OrderSimplify(O);
end;
#######################################

TraceRad := function (O, id)
local M, Rad, deg;

  deg := OrderDeg(O);

  M := OrderTraceMat(O);
  M := MatAppendRight(OrderCoefOrder(O), M,
                      id*MatId(OrderCoefOrder(O), OrderDeg(O)));
  
  M := MatHermiteColUpper(M);
  M := MatRightPart(OrderCoefOrder(O), M, deg);

  M := MatTrans(M);
  M := MatInv(M);
  M := id * M;

  M := MatToRowList(M);
  M := Mat(OrderCoefOrder(O), M);

  Rad := Ideal(O, M, 1);

  return Rad;
end;
#######################################

OrderMaximalDiscOneStep := function(O, id)
local tr;

  tr:=TraceRad(O, id);
  return RingOfMultiplikators(tr);
#  return RingOfMultiplikators(TraceRad(O, id));
end;
#######################################

OrderMaximalDiscSub := function(O, id)
local newO, oldO, index, ideal;

  newO := O;
  ideal:=id;
  repeat
    oldO := newO;
    newO := OrderMaximalDiscOneStep(oldO, ideal);
    index := OrderDisc(newO)/OrderDisc(oldO);
    ideal:= IntGcd(ideal, OrderDisc(newO));
  until (index = 1 or ideal = 1);

  return newO;
end;

####################################################
#
#
#
#
####################################################
OrderMaximalDisc:=function(o, disc)
# o is some order, disc is field discriminant or list of primes
#  dividing the field discriminant.
# Computes the maximal order
local  run,LL, L, deg, O, i, d, d_other, dp;

deg:=OrderDeg(o);
if IsList(disc) then 
  L:=disc;
else
  L:=Factor(disc);
  Apply(L,x->x[1]);
  L:=Filtered(L,x->x>deg);
fi;

run:=Runtime();
#Small primes are different
O:=o;
d:=OrderDisc(O);
for i in [1..deg] do
  if IsPrime(i) and Valuation(i, d) > 0 then
    O:=OrderPMaximal(O,i);
    d:=d/i^Valuation(i,d);
  fi;
od;
#Print("Small primes time: ",Runtime()-run,"\n");
run:=Runtime();
# no small prime divides d...
d_other:=Abs(d);
for i in L do
  d_other:=d_other/i^Valuation(i,d_other);
od;
if not IntIsSquare(d_other) then
  Error("Something went wrong...\n");
fi;
#root of d_other
LL:=Factor(Poly(Zx,[1,0,-d_other]))[1][1];
d_other:=Abs(PolyToList(LL)[2]);
#Print("d_other",d_other,"\n");
if d_other > 1 then
  O:=OrderMaximalDiscSub(O,d_other);
  d_other:=IntGcd(d_other,OrderDisc(O));
  if d_other > 1 then 
#    Print("Warning other primes\n");
    if IntIsSquare(d_other) then
       LL:=Factor(Poly(Zx,[1,0,-d_other]))[1][1];
       d_other:=Abs(PolyToList(LL)[2]);
       O:=OrderMaximalDiscSub(O,d_other);
       d_other:=IntGcd(d_other,OrderDisc(O));
    fi;
    if d_other > 1 then
      d_other:=Factor(d_other);
      Apply(d_other,x->x[1]);
      Append(L,d_other);
    fi;
  fi;
fi;
#Print("Other primes time: ",Runtime()-run,"\n");
run:=Runtime();
#ramified primes
dp:=Product(L);
O:=OrderMaximalDiscSub(O,dp);
#Print("Ramified primes time: ",Runtime()-run,"\n");
OrderSetMaximal(O);

return O;
end;
