
PolyOrthoIdem := function(fl)
#   
#  Compute orthogonal idempotent polynomials.  
#
   local f, L, g, E, i, l;
   fl := List(fl, x->x[1]^x[2]);
   f := Product(fl);
   L := List([1..Length(fl)], i->f/fl[i]);
   g := L[1];
   E := [ g ];
   for i in [2..Length(L)] do
      l := PolyXGcd(g, L[i]);
      g := l[1];
      E := l[2] * E;
      Add(E, l[3]*L[i]);
   od;
   return List(E, x->x mod f);
end;


MatNormalForm := function(M, k)
   
   local kx, x, n, U, S, c, Cycl, i, l, k, fl, j, D, Mx;
   
#
#  M operates from right: phi (v_1, ..., v_n) = (v_1, ..., v_n) M
#
   n := MatRows(M);
   kx := PolyAlg(k);
   x := Poly(kx, [1,0]);
   Mx := MatMove(M, kx);

   S := MatSmithTrans(Mx - x * MatId(kx, n));
   
#
#  Get generating vectors for cyclic subspaces
#
   U := 1/S[4];
   S := S[2];
   Cycl := [];
   for i in [1..n] do
      if PolyDeg(S[i][i]) > 0 then
         Add(Cycl, [ S[i][i], Sum([1..n], j->Eval(U[i][j], M)[j]) ] );
      fi;
   od;
#
#  Decompose cyclic subspaces canonically (algebra splitting).
#  Compute corresponding basis vectors.
#
   D := [];
   for c in Cycl do
      fl := Factor(c[1]);
      l := PolyOrthoIdem(fl);
      for k in [1..Length(fl)] do
            Add(D, [ c[2]*Eval(l[k], M), fl[k][1], fl[k][2] ]);
      od;
   od;
   
   return D;
end;

